/****************************************************************************\
*   BIEWUTIL - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   biewutil.c                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain useful proimitives of BIEW project.                     *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stddef.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <stdarg.h>
/***************************************************************************/
/************ Misceleanous stuff: qsort & qfind  ***************************/
/***************************************************************************/

#include "pmalloc.h"
#include "bmfile.h"
#include "biewutil.h"
#include "bconsole.h"
#include "tstrings.h"
#include "biewlib.h"
#if __BITNESS__ == 16
#include <alloc.h>
#include <mem.h>
#endif

int DumpMode = 0;
int EditMode = 0;

int GetBool(int bool) { return bool ? 'x' : ' '; }

void FFreeArr(void ** arr,unsigned n)
{
  unsigned i;
  for(i = 0;i < n;i++) PFREE(arr[i]);
}

int Summ(unsigned char *str,unsigned size)
{
  unsigned res,i;
  res = 0;
  for(i = 0;i < size;i++) res += str[i];
  return res;
}

char * GetBinary(char val)
{
  static char bstr[9];
  int i;
  bstr[8] = 0;
  for(i = 0;i < 8;i++) bstr[7-i] = ((val >> i) & 1) + '0';
  return bstr;
}

static char str[4];

char * Get2Digit(unsigned char val)
{
  str[0] = legalchars[((val >> 4) & 0x0F)+2];
  str[1] = legalchars[(val & 0x0F)+2];
  str[2] = 0;
  return str;
}

char * Get2SignDig(signed char val)
{
  str[0] = val > 0 ? '+' : '-';
  if(val < 0) val = abs(val);
  str[1] = legalchars[((val >> 4) & 0x0F)+2];
  str[2] = legalchars[(val & 0x0F)+2];
  str[3] = 0;
  return str;
}

char * Get4Digit(unsigned short val)
{
  static char rstr[5];
  strcpy(rstr,Get2Digit(val>>8));
  strcat(rstr,Get2Digit(val));
  return rstr;
}

char * Get4SignDig(short val)
{
  static char rstr[6];
  rstr[0] = val > 0 ? '+' : '-';
  if(val < 0) val = abs(val);
  strcpy(&rstr[1],Get2Digit(val>>8));
  strcat(rstr,Get2Digit(val));
  return rstr;
}

char * Get8Digit(unsigned long val)
{
  static char rstr[9];
  strcpy(rstr,Get4Digit(val>>16));
  strcat(rstr,Get4Digit(val));
  return rstr;
}

char * Get8SignDig(long val)
{
  static char rstr[10];
  rstr[0] = val > 0 ? '+' : '-';
  if(val < 0) val = labs(val);
  strcpy(&rstr[1],Get4Digit(val>>16));
  strcat(rstr,Get4Digit(val));
  return rstr;
}

static char __NEAR__ GetHexAnalog(char val)
{
   if(val >= '0' && val <= '9') return val-'0';
   else return ((toupper(val)-'A'+10)) & 0x0F;
}

void CompressHex(unsigned char * dest,const char * src,int sizedest,int usespace)
{
  int i,j;
  for(i = j = 0;j < sizedest;j++)
  {
      dest[j] = (GetHexAnalog(src[i++]) << 4) | GetHexAnalog(src[i++]);
      if(usespace) i++;
  }
}

memArray * ma_Build( int nitems, int interact )
{
  memArray * ret;
  ret = PMalloc(sizeof(memArray));
  if(ret)
  {
    memset(ret,0,sizeof(memArray));
    if(nitems)
    {
      ret->data = PMalloc(nitems*sizeof(char *));
      if(ret->data)
      {
        ret->nSize = nitems;
      }
    }
  }
  else
  {
    if(interact) MemOutBox("List creation");
  }
  return ret;
}

void  ma_Destroy(memArray *obj)
{
  unsigned i;
  for(i = 0;i < obj->nItems;i++)
  {
    PFREE(obj->data[i]);
  }
  PFREE(obj->data);
  PFREE(obj);
}

#define LST_STEP 16

int  ma_AddData(memArray *obj,void *udata,unsigned len,int interact)
{
  char *new_item;
  if(obj->nSize > UINT_MAX - (LST_STEP+1)) return 0;
  if(obj->nItems + 1 > obj->nSize)
  {
    void *ptr;
    if(!obj->data) ptr = PMalloc((obj->nSize+LST_STEP)*sizeof(char *));
    else           ptr = PRealloc(obj->data,sizeof(char *)*(obj->nSize+LST_STEP));
    if(ptr)
    {
      obj->nSize = obj->nSize+LST_STEP;
      obj->data = ptr;
    }
    else goto err;
  }
  new_item = PMalloc(len);
  if(new_item)
  {
    memcpy(new_item,udata,len);
    obj->data[obj->nItems++] = new_item;
    return 1;
  }
  else
  {
    err:
    if(interact)
    {
      MemOutBox("Building list");
    }
  }
  return 0;
}

int ma_AddString(memArray *obj,char *udata,int interact)
{
  return ma_AddData(obj,udata,strlen(udata)+1,interact);
}

int ma_Display(memArray *obj,char *title,int flg, unsigned defsel)
{
  return CommonListBox((char **)obj->data,obj->nItems,title,flg,defsel);
}

linearArray * la_Build( unsigned long nitems, unsigned size_of_item,
                        int interact )
{
  linearArray * ret;
  ret = PMalloc(sizeof(linearArray));
  if(ret)
  {
    memset(ret,0,sizeof(linearArray));
    ret->itemSize = size_of_item;
    if(nitems)
    {
      ret->data = PHMalloc(nitems*size_of_item);
      if(ret->data)
      {
        ret->nSize = nitems;
      }
    }
  }
  else
  {
    if(interact) MemOutBox("Linear array creation");
  }
  return ret;
}

void  la_Destroy(linearArray *obj)
{
  PHFREE(obj->data);
  PFREE(obj);
}

int  la_AddData(linearArray *obj,void *udata,int interact)
{
  if(obj->nSize > ULONG_MAX - (LST_STEP+1)) return 0;
  if(obj->nItems + 1 > obj->nSize)
  {
    void *ptr;
    if(!obj->data) ptr = PHMalloc((obj->nSize+LST_STEP)*obj->itemSize);
    else           ptr = PHRealloc(obj->data,obj->itemSize*(obj->nSize+LST_STEP));
    if(ptr)
    {
      obj->nSize = obj->nSize+LST_STEP;
      obj->data = ptr;
    }
    else
    {
      if(interact)
      {
        MemOutBox("Building list");
      }
      return 0;
    }
  }
  HMemCpy(((char __HUGE__ *)obj->data) + obj->nItems*obj->itemSize,udata,obj->itemSize);
  obj->nItems++;
  return 1;
}

void         la_Sort(linearArray *obj,int (* compare)(const void __HUGE__ *e1, const void __HUGE__ *e2))
{
  if(obj)
    if(obj->nItems)
      HQSort(obj->data,obj->nItems,obj->itemSize,compare);
}

void __HUGE__ *la_Find(linearArray * obj,const void *key,
                       int (* compare)(const void __HUGE__ *e1, const void __HUGE__ *e2))
{
  if(obj)
    if(obj->nItems)
     return HLFind(key,obj->data,obj->nItems,obj->itemSize,compare);
  return NULL;
}

unsigned long la_FindNearest(linearArray *obj,const void *key,
                              int (* compare)(const void __HUGE__ *e1, const void __HUGE__ *e2))
{
  if(obj)
    if(obj->nItems)
      return HLFindNearest(key,obj->data,obj->nItems,obj->itemSize,compare);
  return 0;
}


void Space2Tab(char *dest,const char * src)
{
  unsigned char buff[8],nspc;
  unsigned int i,len,j,limit;
  unsigned int dest_idx;
  len = strlen(src);
  i = 0;
  dest_idx = 0;
  while(1)
  {
     if(i + 8 < len)
     {
        for(j = 0;j < 8;j++) buff[j] = src[i++];
        // scan
        nspc = 0;
        for(j = 7;j >= 0;j--)
        {
          if(buff[j] != ' ') break;
          else nspc++;
        }
        limit = 8 - nspc;
        for(j = 0;j < limit;j++) dest[dest_idx++] = buff[j];
        if(nspc) dest[dest_idx++] = '\t';
     }
     else
     {
       limit = len - i;
       for(j = 0;j < limit;j++) dest[dest_idx++] = src[i++];
       break;
     }
  }
  dest[dest_idx] = '\0';
}

void TrimTrailingSpace(char *str)
{
  unsigned len;
  len = strlen(str);
  while(isspace(str[len-1])) str[--len] = '\0';
}
 