/****************************************************************************\
*   BMFILE - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   bmfile.c                                           *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file has developed as Buffered stream Manager and present front      *
*  end interface to bbio library.                                            *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stddef.h>
#include <string.h>
#include <errno.h>

#include "bbio.h"
#include "bmfile.h"
#include "bconsole.h"
#include "__os_dep.h"

static BGLOBAL file = NULL;
static char FName[80];

BGLOBAL       biewOpen(char *fname,int cache_size)
{
  BGLOBAL fret;
  fret = bioOpen(fname,FO_READWRITE | SO_DENYNONE,cache_size);
  if(fret == &bNull)
    fret = bioOpen(fname,FO_READWRITE | SO_COMPAT,cache_size);
    if(fret == &bNull)
      fret = bioOpen(fname,FO_READONLY | SO_DENYNONE,cache_size);
      if(fret == &bNull)
        fret = bioOpen(fname,FO_READONLY | SO_COMPAT,cache_size);
  return fret;
}

int           BMOpen(char * fname)
{
  strcpy(FName,fname);
  file = biewOpen(fname,BBIO_CACHE_SIZE);
  if(file == &bNull)
  {
    errnoMessageBox("Can not open file",NULL,errno);
    return -1;
  }
  bioSetOptimization(file,BIO_OPT_RANDOM);
  return 0;
}

int BMHandle( void )
{
  return bioHandle(file);
}

BGLOBAL BMbioHandle( void )
{
  return file;
}

void           BMClose( void )
{
  if(file != &bNull) bioClose(file);
}

void           BMSeek(long pos,int RELATION)
{
  bioSeek(file,pos,RELATION);
}

void           BMReRead( void )
{
  bioReRead(file);
}

unsigned char  BMReadByte( void )
{
  return bioReadByte(file);
}

unsigned short BMReadWord( void )
{
 return bioReadWord(file);
}

unsigned long  BMReadDWord( void )
{
  return bioReadDWord(file);
}

unsigned long  BMGetCurrFilePos( void )
{
  return bioTell(file);
}

unsigned long BMGetFLength( void )
{
  return bioFLength(file);
}

int            BMReadBuffer(void * buffer,unsigned len)
{
  return bioReadBuffer(file,buffer,len);
}

unsigned char  BMReadByteEx(long pos,int RELATION)
{
 bioSeek(file,pos,RELATION);
 return bioReadByte(file);
}

unsigned short BMReadWordEx(long pos,int RELATION)
{
 bioSeek(file,pos,RELATION);
 return bioReadWord(file);
}

unsigned long  BMReadDWordEx(long pos,int RELATION)
{
 bioSeek(file,pos,RELATION);
 return bioReadDWord(file);
}

int           BMReadBufferEx(void * buffer,unsigned len,long pos,int RELATION)
{
 bioSeek(file,pos,RELATION);
 return bioReadBuffer(file,buffer,len);
}

int           BMWriteByte(unsigned char byte)
{
  return bioWriteByte(file,byte);
}

int           BMWriteWord(unsigned int word)
{
  return bioWriteWord(file,word);
}

int           BMWriteDWord(unsigned long dword)
{
  return bioWriteDWord(file,dword);
}

int           BMWriteBuff(void * buff,unsigned len)
{
  int ret;
  ret = bioWriteBuffer(file,buff,len);
  bioFlush(file);
  return ret;
}

int            BMWriteByteEx(long pos,int RELATION,unsigned char byte)
{
  bioSeek(file,pos,RELATION);
  return bioWriteByte(file,byte);
}

int           BMWriteWordEx(long pos,int RELATION,unsigned int word)
{
  bioSeek(file,pos,RELATION);
  return bioWriteWord(file,word);
}

int           BMWriteDWordEx(long pos,int RELATION,unsigned long dword)
{
  bioSeek(file,pos,RELATION);
  return bioWriteDWord(file,dword);
}

int           BMWriteBuffEx(long pos,int RELATION,void * buff,unsigned len)
{
  bioSeek(file,pos,RELATION);
  return bioWriteBuffer(file,buff,len);
}

char *        BMName( void )
{
  return bioFileName(file);
}

