/****************************************************************************\
*   EDITORS - this source file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   editors.c                                          *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain low level editor implementation of BIEW project.        *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <errno.h>

#include "pmalloc.h"
#include "bmfile.h"
#include "kbd_code.h"
#include "tstrings.h"
#include "disasm.h"
#include "bconsole.h"
#include "biewutil.h"
#include "biewhelp.h"
#include "editor.h"
#include "__os_dep.h"

unsigned char *alen,*aolen;
long edit_cp = 0;
union tag_emem EditorMem,EditorSaveMem;

int edit_x,edit_y;
unsigned char edit_XX = 0;

void ExtHelp( void )
{
 TWindow * using = twUsedWin();
 hlpDisplay(2);
 twUseWin(using);
}

void PaintETitle( int shift,int use_shift )
{
  TWindow * using = twUsedWin();
  char byte,obyte;
  char ostr[__TVIO_MAXSCREENWIDTH];
  memset(ostr,' ',sizeof(ostr));
  twUseWin(TitleWnd);
  twGotoXY(1,1);
  memcpy(ostr,Get8Digit(edit_cp + shift),8);
  ostr[8] = ':';
  ostr[9] = ' ';
  twPrintF("%08lX: ",edit_cp + shift);
  if(!use_shift)
  {
    byte  = EditorMem.arr[edit_y*tvioWidth+edit_x];
    obyte = EditorSaveMem.arr[edit_y*tvioWidth+edit_x];
  }
  else
  {
     byte  = EditorMem.arr[shift];
     obyte = EditorSaveMem.arr[shift];
  }
  ostr[11] = byte;
  memcpy(&ostr[13],Get2Digit(byte),2);
  ostr[15] = 'H';
  memcpy(&ostr[17],Get2SignDig(byte),3);
  ostr[20] = 'H';
  memcpy(&ostr[22],GetBinary(byte),8);
  ostr[30] = 'B';
  if(byte != obyte)
  {
    memcpy(&ostr[33],"ORIGINAL :",10);
    ostr[45] = obyte;
    memcpy(&ostr[47],Get2Digit(obyte),2);
    ostr[49] = 'H';
    memcpy(&ostr[51],Get2SignDig(obyte),3);
    ostr[54] = 'H';
    memcpy(&ostr[56],GetBinary(obyte),8);
    ostr[64] = 'B';
  }
  twDirectWrite(1,1,ostr,Width);
  memset(ostr,0,sizeof(ostr));
  memcpy(&ostr[1],"MASK :",6);
  memcpy(&ostr[9],Get2Digit(edit_XX),2);
  ostr[11] = 'H';
  twDirectWrite(Width-14,1,ostr,80);
  twUseWin(using);
}

unsigned long editInitBuffs( int width, int *err_state )
{
 unsigned long flen,cfp;
 int i,ws;
 unsigned int nbyte;
 EditorMem.arr = PMalloc(tvioHeight*tvioWidth);
 EditorSaveMem.arr = PMalloc(tvioHeight*tvioWidth);
 alen = PMalloc(tvioHeight);
 aolen = PMalloc(tvioHeight);
 if((!EditorMem.arr) || (!EditorSaveMem.arr) || (!alen) || (!aolen))
 {
   if(EditorMem.arr) PFREE(EditorMem.arr);
   if(EditorSaveMem.arr) PFREE(EditorSaveMem.arr);
   if(alen) PFREE(alen);
   if(aolen) PFREE(aolen);
   MemOutBox("Editor initialization");
   *err_state = 1;
   return 0;
 }
 *err_state = 0;
 flen = BMGetFLength();
 edit_cp = cfp = BMGetCurrFilePos();
 ws = flen > width ? width : (int)flen;
 for(i = 0;i < Height;i++)
 {
   if(edit_cp < flen - ws) nbyte = width;
   else
     if(edit_cp < flen)  nbyte = (int)(flen - edit_cp);
     else                nbyte = 0;
   alen[i] = nbyte;
   if(nbyte)
   {
    BMReadBufferEx(&EditorMem.arr[i*tvioWidth],nbyte,edit_cp,BM_SEEK_SET);
    memcpy(&EditorSaveMem.arr[i*tvioWidth],&EditorMem.arr[i*tvioWidth],alen[i]);
   }
   edit_cp += width;
 }
 BMSeek(cfp,BM_SEEK_SET);
 return cfp;
}

void editDestroyBuffs( void )
{
   PFREE(EditorMem.arr);
   PFREE(EditorSaveMem.arr);
   PFREE(alen);
   PFREE(aolen);
}

void CheckBounds( void )
{
   if(edit_y < 0) edit_y = 0;
   if(edit_y > Height - 1) edit_y = Height - 1;
   if(!alen[edit_y]) edit_y--;
   if(edit_x >= alen[edit_y]) edit_x = alen[edit_y] - 1;
}

void CheckYBounds( void )
{
 if(edit_y < 0) edit_y = 0;
 if(edit_y > Height - 1) edit_y = Height - 1;
 while(!alen[edit_y]) edit_y--;
}

void CheckXYBounds( void )
{
   CheckYBounds();
   if(edit_x < 0) edit_x = alen[--edit_y]*2;
   if(edit_x >= alen[edit_y]*2) { edit_x = 0; edit_y++; }
   CheckYBounds();
}

void editSaveContest( void )
{
 int i,j;
  for(i = 0,j = 0;i < Height;i++)
  {
    if(alen[i])
    {
      memmove(&EditorMem.arr[j*tvioWidth],&EditorMem.arr[i*tvioWidth],alen[i]);
      j += alen[i];
    }
  }
  if(!BMWriteBuffEx(edit_cp,BM_SEEK_SET,(void *)EditorMem.arr,j))
  {
    errnoMessageBox("Can not write to file",NULL,errno);
  }
}

int edit_defaction(int lastbyte)
{
 int redraw;
  redraw = 1;
   switch(lastbyte)
   {
     case KE_ENTER    : redraw = 0; break;
     case KE_UPARROW  : edit_y--; redraw = 0; break;
     case KE_DOWNARROW: edit_y++; redraw = 0; break;
     case KE_LEFTARROW:
     case KE_RIGHTARROW: redraw = 0; break;
     case KE_F(3)     :
                      Get2DigitDlg(INIT_MASK,INPUT_MASK,&edit_XX);
                      redraw = 0;
                      break;
     default: redraw = 1; break;
   }
 return redraw;
}

int editDefAction(int lastbyte)
{
 int redraw = 1;
   switch(lastbyte)
   {
     case KE_F(4)     : EditorMem.arr[edit_y*tvioWidth+edit_x] = ~EditorMem.arr[edit_y*tvioWidth+edit_x]; break;
     case KE_F(5)     : EditorMem.arr[edit_y*tvioWidth+edit_x] |= edit_XX; break;
     case KE_F(6)     : EditorMem.arr[edit_y*tvioWidth+edit_x] &= edit_XX; break;
     case KE_F(7)     : EditorMem.arr[edit_y*tvioWidth+edit_x] ^= edit_XX; break;
     case KE_F(8)     : EditorMem.arr[edit_y*tvioWidth+edit_x]  = edit_XX; break;
     case KE_F(9)     : EditorMem.arr[edit_y*tvioWidth+edit_x] = EditorSaveMem.arr[edit_y*tvioWidth+edit_x]; break;
     default        : redraw = edit_defaction(lastbyte); edit_x--; break;
   }
   edit_x++;
   return redraw;
}

int FullEdit( int width ,TWindow * txtwnd,char ncol,char hcol)
{
 int i,j,mlen;
 unsigned int lastbyte;
 char attr = __ESS_HARDEDIT | __ESS_WANTRETURN;
 twTextColor(ncol);
 __MsSetState(0);
 twFreezeWin(twUsedWin());
 for(i = 0;i < Height;i++)
 {
   for(j = 0;j < alen[i];j++)
   {
     twTextColor(EditorMem.arr[i*tvioWidth+j] == EditorSaveMem.arr[i*tvioWidth+j] ? ncol : hcol );
     twDirectWrite(j + 1,i + 1,&EditorMem.arr[i*tvioWidth+j],1);
   }
   if(alen[i] + 1 < width)
   {
      twGotoXY(alen[i] + 1,i + 1); twClrEOL();
   }
 }
 twRefreshWin(twUsedWin());
 __MsSetState(1);
 PaintETitle(edit_y*width + edit_x,0);
 while(1)
 {
   mlen = alen[edit_y];
   lastbyte = eeditstring((char *)&EditorMem.arr[edit_y*tvioWidth],NULL,&mlen,edit_y + 1,
                          &edit_x,attr,(char *)&EditorSaveMem.arr[edit_y*tvioWidth],ncol,hcol);
   switch(lastbyte)
   {
     case KE_F(1)   : ExtHelp(); continue;
     case KE_F(2)   : editSaveContest();
     case KE_F(10)  :
     case KE_ESCAPE : return lastbyte;
     case KE_TAB : if(txtwnd) return lastbyte;
     default     : editDefAction(lastbyte); break;
   }
   CheckBounds();
   PaintETitle(edit_y*width + edit_x,0);
 }
 return lastbyte;
}
 