/****************************************************************************\
*   EVENTS - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   events.c                                           *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain console event handler of BIEW project.                  *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <limits.h>
#include <string.h>

#include "kbd_code.h"
#include "bconsole.h"
#include "__os_dep.h"
#include "_sys_dep.h"

static int KB_Buff[64];
static unsigned char KB_freq = 0;

int  IsKbdTerminate( void )
{
  int ch = 0;
  if(__kbdTestKey(0)) ch = __kbdGetKey(0);
  return ch == KE_ESCAPE;
}

static unsigned mx,my;

static int __NEAR__ getPromptKey(int flag)
{
 int ret;
 if(mx < 7) ret = KE_F(1);
 else
   if(mx < 15) ret = KE_F(2);
   else
     if(mx < 23) ret = KE_F(3);
     else
       if(mx < 31) ret = KE_F(4);
       else
         if(mx < 39) ret = KE_F(5);
         else
           if(mx < 47) ret = KE_F(6);
           else
             if(mx < 55) ret = KE_F(7);
             else
               if(mx < 63) ret = KE_F(8);
               else
                 if(mx < 71) ret = KE_F(9);
                 else        ret = KE_F(10);
 if(flag & KS_SHIFT) ret += KE_SHIFT_F(1) - KE_F(1);
 else
   if(flag & KS_ALT) ret += KE_ALT_F(1) - KE_F(1);
   else
     if(flag & KS_CTRL) ret += KE_CTL_F(1) - KE_F(1);
 return ret;
}

static int __NEAR__ __GetEvent( void (*prompt)(void) ,TWindow *win)
{
 static char oFlag = 0xFF;
 static void (* oprompt)(void) = 0;
 char Flag;
 int key;
 /* Unconditional repaint prompt */
 Flag = __kbdGetShiftsKey();
 if(Flag != oFlag || oprompt !=prompt)
 {
    oFlag = Flag; oprompt = prompt;
    if(prompt) (*prompt)();
 }
 while(1)
 {
     key = __kbdGetKey(KBD_NONSTOP_ON_MOUSE_PRESS);
     switch( key )
     {
       case KE_SHIFTKEYS:
             Flag = __kbdGetShiftsKey();
             if(Flag != oFlag || oprompt !=prompt)
             {
               oFlag = Flag; oprompt = prompt;
               if(prompt) (*prompt)();
             }
             break;
       case KE_MOUSE:
             __MsGetPos(&mx,&my);
             if((__MsGetBtns() & MS_LEFTPRESS) == MS_LEFTPRESS)
             {
                  if(my == tvioHeight - 1)
                  {
                    while(1)
                    {
                      if((__MsGetBtns() & MS_LEFTPRESS) != MS_LEFTPRESS) break;
                      __MsGetPos(&mx,&my);
                    }
                    if(my != tvioHeight - 1) continue;
                    return getPromptKey(__kbdGetShiftsKey());
                  }
                  else
                  {
                    int X1,Y1,X2,Y2;
                    unsigned wdh,hght;
                    if(win)
                    {
                      twGetWinPos(win,&X1,&Y1,&X2,&Y2);
                      X1--; Y1--; X2--; Y2--;
                    }
                    else
                    {
                      X1 = 0; X2 = tvioWidth; Y1 = 1; Y2 = tvioHeight - 1;
                    }
                    wdh = X2 - X1;
                    hght = Y2 - Y1;
                    if(win && (mx < X1 || mx > X2 || my < Y1 || my > Y2))
                    {
                       while(1)
                       {
                          if((__MsGetBtns() & MS_LEFTPRESS) != MS_LEFTPRESS) break;
                       }
                       return KE_ESCAPE;
                    }
                    if(my <= Y2 && my >= Y1 + 3*hght/4) return KE_DOWNARROW;
                    else
                      if(my >= Y1 && my <= Y1 + hght/4) return KE_UPARROW;
                      else
                        if(mx <= X2 && mx >= X1 + 3*wdh/4) return KE_RIGHTARROW;
                        else
                          if(mx >= X1 && mx <= X1 + wdh/4) return KE_LEFTARROW;
                  }
              }
        default: return key;
     }
 }
 return key;
}

static void __NEAR__ __GetEventQue(void (*prompt)(void), TWindow *win)
{
  int key;
  key = __GetEvent(prompt,win);
  if(KB_freq < sizeof(KB_Buff)/sizeof(int)) KB_Buff[KB_freq++] = key;
}

int GetEvent( void (*prompt)(void) ,TWindow * win)
{
  int key;
  while(!KB_freq) __GetEventQue(prompt,win);
  key = KB_Buff[0];
  --KB_freq;
  if(KB_freq) memmove(KB_Buff,&KB_Buff[1],KB_freq-1);
  return key;
}

void PostEvent(int code)
{
  if(KB_freq < sizeof(KB_Buff)/sizeof(int)) KB_Buff[KB_freq++] = code;
}
 