/****************************************************************************\
*   FILEUTIL - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   fileutil.c                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file conatin file utilities of BIEW project                          *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <limits.h>
#include <errno.h>
#include <sys/stat.h>
#include <ctype.h>

#include "pmalloc.h"
#include "bbio.h"
#include "twin.h"
#include "editor.h"
#include "bmfile.h"
#include "tstrings.h"
#include "kbd_code.h"
#include "disasm.h"
#include "biewutil.h"
#include "bconsole.h"
#include "reg_form.h"
#include "__os_dep.h"

int ChSize( void )
{
 long psize,tile = 0;
 if(Get8DigitDlg(" Change size of file ","Num. of bytes (+-dec):",3,(unsigned long *)&tile))
 {
  if(tile != 0)
  {
    psize = BMGetFLength();
    psize += tile;
    if(psize > 0) { bioChSize(BMbioHandle(),psize); return 1; }
    else ErrMessageBox("New length has invalid value",NULL);
  }
 }
 return 0;
}

static int __NEAR__ InsBlock(unsigned long start,long psize)
{
   char *buffer;
   unsigned long tile,oflen,flen,crpos,cwpos;
   unsigned numtowrite;
   oflen = BMGetFLength();
   flen = oflen + psize;
   tile = oflen - start;
   buffer = PMalloc(51200U);
   if(!buffer) return 0;
   if(!bioChSize(BMbioHandle(),oflen+psize))
   {
     ErrMessageBox("Can not expand file",NULL);
     PFREE(buffer);
     return 0;
   }
   crpos = oflen-min(tile,51200U);
   cwpos = flen-min(tile,51200U);
   numtowrite = (unsigned)min(tile,51200U);
   while(tile)
   {
     BMReadBufferEx(buffer,numtowrite,crpos,BM_SEEK_SET);
     BMWriteBuffEx(cwpos,BM_SEEK_SET,buffer,numtowrite);
     tile -= numtowrite;
     numtowrite = (unsigned)min(tile,51200U);
     crpos -= numtowrite;
     cwpos -= numtowrite;
   }
   tile = oflen - start;
   cwpos = start;
   memset(buffer,0,51200U);
   while(psize)
   {
     numtowrite = (unsigned)min(psize,51200U);
     BMWriteBuffEx(cwpos,BM_SEEK_SET,buffer,numtowrite);
     psize -= numtowrite;
     cwpos += numtowrite;
   }
   PFREE(buffer);
   return 1;
}

static int __NEAR__ DelBlock(unsigned long start,long psize)
{
   char *buffer;
   unsigned long tile,oflen,crpos,cwpos;
   unsigned numtowrite;
   oflen = BMGetFLength();
   tile = oflen - start;
   buffer = PMalloc(51200U);
   if(!buffer) return 0;
   crpos = start-psize; // psize is negative value
   cwpos = start;
   while(tile)
   {
     numtowrite = (unsigned)min(tile,51200U);
     BMReadBufferEx(buffer,numtowrite,crpos,BM_SEEK_SET);
     BMWriteBuffEx(cwpos,BM_SEEK_SET,buffer,numtowrite);
     tile -= numtowrite;
     crpos += numtowrite;
     cwpos += numtowrite;
   }
   PFREE(buffer);
   if(!bioChSize(BMbioHandle(),oflen+psize))
   {
     ErrMessageBox("Can not truncate file",NULL);
     PFREE(buffer);
   }
   return 1;
}

int InsDelBlock( void )
{
 unsigned long start;
 static long psize;
 int ret = 0;
 start = BMGetCurrFilePos();
 if(GetInsDelBlkDlg(" Insert or delete block to/from file ",&start,&psize))
 {
    unsigned long fpos;
    fpos = BMGetCurrFilePos();
    if(start > BMGetFLength()) { ErrMessageBox("Can not start outside the file",NULL); return 0; }
    if(!psize) return 0;
    if(psize < 0) if(start+labs(psize) > BMGetFLength()) { ErrMessageBox("Use change size operation instead of block deletion",NULL); return 0; }
    if(psize < 0) ret = DelBlock(start,psize);
    else          ret = InsBlock(start,psize);
    BMSeek(fpos,BM_SEEK_SET);
 }
 return ret;
}

char ff_fname[80] = "biew.$$$";
unsigned long ff_startpos = 0L,ff_len = 0L;

static void __NEAR__ printObject(FILE *fout,unsigned obj_num,char *oname,int oclass,int obitness,unsigned long size)
{
  char *name,*btn;
  char onumname[30];
  switch(obitness)
  {
    case DAB_USE16: btn = "USE16"; break;
    case DAB_USE32: btn = "USE32"; break;
    case DAB_USE64: btn = "USE64"; break;
    case DAB_USE128:btn = "USE128"; break;
    case DAB_USE256:btn = "USE256"; break;
    default: btn = "";
  }

  name = oname[0] ? oname : oclass == OC_DATA ? "DUMP_DATA" :
                            oclass == OC_CODE ? "DUMP_TEXT" :
                            "Unknown";
  if(!oname[0]) { sprintf(onumname,"%s%u",name,obj_num); name = onumname; }
  fprintf(fout,"\nSEGMENT %s BYTE PUBLIC %s '%s'\n; size: %lu bytes\n\n"
              ,name
              ,btn
              ,oclass == OC_DATA ? "DATA" : oclass == OC_CODE ? "CODE" : "NoObject"
              ,size);
}

static void __NEAR__ printHdr(FILE * fout,REGISTRY_BIN *fmt)
{
  char *cptr,*cptr1,*cptr2;
  time_t tim;
  cptr = cptr1 = ";"; cptr2 = "";
  time(&tim);
  fprintf(fout,"%s\n%sDissasembler dump of \'%s\'\n"
               "%sRange : %08lXH-%08lXH\n"
               "%sWritten by BIEW Version "BIEW_VERSION"\n"
               "%sDumped : %s\n"
               "%sFormat : %s\n"
               "%s\n\n"
              ,cptr1,cptr,BMName()
              ,cptr,ff_startpos,ff_startpos+ff_len
              ,cptr
              ,cptr,ctime(&tim)
              ,cptr,fmt->name
              ,cptr2);
}

static int __NEAR__ printHelpComment(char *buff,MBuffer codebuff,DisasmRet *dret)
{
  int len,j;
  len = 0;
   for(j = 0;j < dret->codelen;j++) { memcpy((char *)&buff[len],(char *)Get2Digit(codebuff[j]),2); len += 2; }
   buff[len] = 0;
   return len;
}

extern REGISTRY_MODE disMode;

#define GET_FUNC_CLASS(x) x == SC_LOCAL ? "private" : "public"

int FStore( void )
{
 unsigned long flags;
 char *tmp_buff;
 unsigned long endpos,cpos;
 tmp_buff = PMalloc(0x1000);
 if(!tmp_buff)
 {
       MemOutBox("temporary buffer initialization");
       return 0;
 }
 flags = FSDLG_USEMODES | FSDLG_BINMODE | FSDLG_COMMENT;
 DumpMode = -1;
 ff_startpos = BMGetCurrFilePos();
 if(!ff_len) ff_len = BMGetFLength() - ff_startpos;
 if(GetFStoreDlg(" Store information to File [F2 - Asm/Bin F3 - comment F4 - structs] ",ff_fname,&flags,&ff_startpos,&ff_len,FILE_PRMT))
 {
  endpos = ff_startpos + ff_len;
  endpos = endpos > BMGetFLength() ? BMGetFLength() : endpos;
  if(endpos > ff_startpos)
  {
   TWindow *progress_wnd;
   unsigned prcnt_counter,oprcnt_counter;
   cpos = BMGetCurrFilePos();
   progress_wnd = PercentWnd("Saving ..."," Save block to file ");
   if(!(flags & FSDLG_ASMMODE)) /* Write in binary mode */
   {
     BGLOBAL bioHandle;
     int handle;
     unsigned long wsize,crpos,pwsize,awsize;
     unsigned rem;
     wsize = endpos - ff_startpos;
     if(!__IsFileExists(ff_fname) != 0) handle = __OsCreate(ff_fname);
     else
     {
       handle = __OsOpen(ff_fname,FO_READWRITE | SO_DENYNONE);
       if(handle == -1)  handle = __OsOpen(ff_fname,FO_READWRITE | SO_COMPAT);
       if(handle == -1)
       {
          use_err:
          errnoMessageBox("Can use file",NULL,errno);
          goto Exit;
       }
       __OsTruncFile(handle,0L);
     }
     __OsClose(handle);
     bioHandle = bioOpen(ff_fname,FO_READWRITE | SO_DENYNONE,BBIO_CACHE_SIZE);
     if(bioHandle == &bNull)  bioHandle = bioOpen(ff_fname,FO_READWRITE | SO_COMPAT,BBIO_CACHE_SIZE);
     if(bioHandle == &bNull)  goto use_err;
     crpos = ff_startpos;
     bioSeek(bioHandle,0L,SEEKF_START);
     prcnt_counter = oprcnt_counter = 0;
     pwsize = 0;
     awsize = wsize;
     while(wsize)
     {
        rem = (unsigned)min(wsize,4096);
        if(!BMReadBufferEx(tmp_buff,rem,crpos,BM_SEEK_SET))
        {
          errnoMessageBox("Can not read file",NULL,errno);
          bioClose(bioHandle);
          goto Exit;
        }
        if(activeMode->convert_cp) activeMode->convert_cp((char *)tmp_buff,rem);
        if(!bioWriteBuffer(bioHandle,tmp_buff,rem))
        {
          errnoMessageBox("Can not write to file",NULL,errno);
          bioClose(bioHandle);
          goto Exit;
        }
        wsize -= rem;
        crpos += rem;
        pwsize += rem;
        prcnt_counter = (unsigned)((pwsize*100)/awsize);
        if(prcnt_counter != oprcnt_counter)
        {
          oprcnt_counter = prcnt_counter;
          if(!ShowPercentInWnd(progress_wnd,prcnt_counter)) break;
        }
     }
     bioClose(bioHandle);
   }
   else /* Write in disassembler mode */
   {
     FILE * fout;
     unsigned char *codebuff;
     char *file_cache,*tmp_buff2;
     int MaxInsnLen;
     char func_name[300],obj_name[300],data_dis[300];
     unsigned long func_pa,stop;
     unsigned func_class;
     unsigned long awsize,pwsize;
     int has_string;

     unsigned long obj_start,obj_end;
     int obj_class,obj_bitness;
     unsigned obj_num;

     if(activeMode != &disMode) disMode.init();
     if(flags & FSDLG_STRUCTS)
     {
       if(detectedFormat->set_state) detectedFormat->set_state(PS_ACTIVE);
       if(detectedFormat->prepare_structs)
                    detectedFormat->prepare_structs(ff_startpos,ff_startpos+ff_len);
     }
     MaxInsnLen = activeDisasm->max_insn_len();
     codebuff = PMalloc(MaxInsnLen);
     if(!codebuff)
     {
       MemOutBox("Disasm initialization");
       goto dis_exit;
     }
     tmp_buff2 = PMalloc(0x1000);
     file_cache = PMalloc(BBIO_SMALL_CACHE_SIZE);
     fout = fopen(ff_fname,"wt");
     if(fout == NULL)
     {
        errnoMessageBox("Can not write to file",NULL,errno);
        PFREE(codebuff);
        goto Exit;
     }
     if(file_cache) setvbuf(fout,file_cache,_IOFBF,BBIO_SMALL_CACHE_SIZE);
     if(flags & FSDLG_COMMENT)
     {
       printHdr(fout,detectedFormat);
     }
     if(flags & FSDLG_STRUCTS)
     {
       if(detectedFormat->GetObjAttr)
       {
         obj_num = detectedFormat->GetObjAttr(ff_startpos,obj_name,
                                              sizeof(obj_name),&obj_start,
                                              &obj_end,&obj_class,&obj_bitness);
         obj_name[sizeof(obj_name)-1] = 0;
       }
       else goto defobj;
     }
     else
     {
       defobj:
       obj_num = 0;
       obj_start = 0;
       obj_end = BMGetFLength();
       obj_name[0] = 0;
       obj_class = OC_CODE;
       obj_bitness = detectedFormat->query_bitness ? detectedFormat->query_bitness(ff_startpos) : DAB_USE16;
     }
     if(flags & FSDLG_STRUCTS) printObject(fout,obj_num,obj_name,obj_class,obj_bitness,obj_end - obj_start);
     func_pa = 0;
     if(flags & FSDLG_STRUCTS)
     {
       if(detectedFormat->GetPubSym)
       {
         func_pa = detectedFormat->GetPubSym(func_name,sizeof(func_name),
                                             &func_class,ff_startpos,1);
         func_name[sizeof(func_name)-1] = 0;
         if(func_pa)
         {
            fprintf(fout,"%s %s:\n"
                        ,GET_FUNC_CLASS(func_class)
                        ,func_name);
            if(func_pa < ff_startpos && flags & FSDLG_COMMENT)
            {
              fprintf(fout,"; ...\n");
            }
         }
         func_pa = detectedFormat->GetPubSym(func_name,sizeof(func_name),
                                             &func_class,ff_startpos,0);
         func_name[sizeof(func_name)-1] = 0;
       }
     }
     prcnt_counter = oprcnt_counter = 0;
     awsize = endpos - ff_startpos;
     pwsize = 0;
     has_string = 0;
     while(1)
     {
      DisasmRet dret;
      int len;
       if(flags & FSDLG_STRUCTS)
       {
          if(detectedFormat->GetObjAttr)
          {
            if(ff_startpos >= obj_end)
            {
               obj_num = detectedFormat->GetObjAttr(ff_startpos,obj_name,
                                                    sizeof(obj_name),&obj_start,
                                                    &obj_end,&obj_class,
                                                    &obj_bitness);
               obj_name[sizeof(obj_name)-1] = 0;
               printObject(fout,obj_num,obj_name,obj_class,obj_bitness,obj_end - obj_start);

            }
          }
          if(obj_class == OC_NOOBJECT)
          {
            unsigned long diff;
            fprintf(fout,"; L%08lXH-L%08lXH - is no object\n",obj_start,obj_end);
            dret.codelen = min(UCHAR_MAX,obj_end - ff_startpos);
            /* some functions can placed in virtual area of objects
               mean at end of true data, but before next object */
             while(func_pa && func_pa >= obj_start && func_pa < obj_end)
             {
                  diff = func_pa - ff_startpos;
                  if(diff) fprintf(fout,"resb %08lXH\n",diff);
                  fprintf(fout,"%s %s: ;at offset - %08lXH\n"
                              ,GET_FUNC_CLASS(func_class)
                              ,func_name
                              ,func_pa);
                  ff_startpos = func_pa;
                  func_pa = detectedFormat->GetPubSym(func_name,sizeof(func_name),
                                                    &func_class,ff_startpos,0);
                  func_name[sizeof(func_name)-1] = 0;
              }
              diff = obj_end - ff_startpos;
              if(diff) fprintf(fout,"resb %08lXH\n",diff);
              ff_startpos = obj_end;
              goto next_obj;
          }
          if(detectedFormat->GetPubSym && func_pa)
          {
            if(ff_startpos == func_pa)
            {
              fprintf(fout,"%s %s:\n"
                          ,GET_FUNC_CLASS(func_class)
                          ,func_name);
              func_pa = detectedFormat->GetPubSym(func_name,sizeof(func_name),
                                                  &func_class,ff_startpos,0);
              func_name[sizeof(func_name)-1] = 0;
            }
          }
       }
       memset(codebuff,0,sizeof(codebuff));
       BMReadBufferEx((void *)codebuff,MaxInsnLen,ff_startpos,BM_SEEK_SET);
       if(obj_class == OC_CODE)
         dret = Disassembler(ff_startpos,codebuff);
       else /* Data object */
       {
         unsigned dis_data_len,ifreq,data_len;
         char coll_str[80];
         int cstr_idx = 0;
         dis_data_len = min(sizeof(coll_str)-1,MaxInsnLen);
         for(cstr_idx = 0;cstr_idx < dis_data_len;cstr_idx++)
         {
           if(isprint(codebuff[cstr_idx]))
           {
             coll_str[cstr_idx] = codebuff[cstr_idx];
           }
           else break;
         }
         coll_str[cstr_idx] = 0;
         switch(obj_bitness)
         {
           case DAB_USE16: dis_data_len = 2; break;
           case DAB_USE32: dis_data_len = 4; break;
           case DAB_USE64: dis_data_len = 8; break;
           case DAB_USE128: dis_data_len = 16; break;
           case DAB_USE256: dis_data_len = 32; break;
           default:         dis_data_len = 1; break;
         }
         data_len = 0;
         sprintf(data_dis,"db  ");
         if(cstr_idx > 1)
         {
           sprintf(&data_dis[strlen(data_dis)],"'%s'",coll_str);
           dret.codelen = cstr_idx;
           has_string = 1;
         }
         else
         {
           for(ifreq = 0;ifreq < dis_data_len;ifreq++)
           {
              if(isprint(codebuff[ifreq]) && isprint(codebuff[ifreq+1]))
                                              break;
              if(isprint(codebuff[ifreq]) && has_string)
                  sprintf(&data_dis[strlen(data_dis)],"'%c',",codebuff[ifreq]);
              else
                sprintf(&data_dis[strlen(data_dis)],"%02Xh,",codebuff[ifreq]);
              data_len++;
              has_string = 0;
           }
           dret.codelen = data_len;
         }
         dret.str = data_dis;
         dret.pro_clone = 0;
       }
       if(func_pa) stop = min(func_pa,obj_end);
       else        stop = obj_end;
       if(flags & FSDLG_STRUCTS)
       {
         if(detectedFormat->GetPubSym && stop && stop > ff_startpos &&
            ff_startpos + dret.codelen > stop)
         {
           unsigned lim,ii;
           sprintf(tmp_buff,"L%s: db ",Get8Digit(ff_startpos));
           lim = (unsigned)(stop-ff_startpos);
           if(lim > MaxInsnLen) lim = MaxInsnLen;
           for(ii = 0;ii < lim;ii++)
             sprintf(&tmp_buff[strlen(tmp_buff)],"%s ",Get2Digit(codebuff[ii]));
           dret.codelen = lim;
         }
         else goto normline;
       }
       else { normline: sprintf(tmp_buff,"L%s: %s",Get8Digit(ff_startpos),dret.str); }
       len = strlen(tmp_buff);
       if(flags & FSDLG_COMMENT)
       {
         int i;
          for(i = len;i < 48;i++) strcat(tmp_buff," ");
          len = strlen(tmp_buff);
          ((char *)tmp_buff)[len++] = ';';
          ((char *)tmp_buff)[len++] = ' ';
          len += printHelpComment(&((char *)tmp_buff)[len],codebuff,&dret);
       }
       if(tmp_buff2)
       {
         Space2Tab(tmp_buff2,tmp_buff);
         TrimTrailingSpace(tmp_buff2);
       }
       strcat(tmp_buff2 ? tmp_buff2 : tmp_buff,"\n");
       if(fputs(tmp_buff2 ? tmp_buff2 : tmp_buff,fout) == EOF)
       {
         errnoMessageBox("Can not write to file",NULL,errno);
         goto dis_exit;
       }
       if(flags & FSDLG_STRUCTS)
       {
         if(detectedFormat->GetPubSym && stop && ff_startpos != stop &&
            ff_startpos + dret.codelen > stop)
                      dret.codelen = stop - ff_startpos;
       }
       if(!dret.codelen)
       {
         ErrMessageBox("Internal Fatal error"," Put structures ");
         goto dis_exit;
       }
       ff_startpos += dret.codelen;
       next_obj:
       if(ff_startpos >= endpos) break;
       pwsize += dret.codelen;
       prcnt_counter = (unsigned)((pwsize*100)/awsize);
       if(prcnt_counter != oprcnt_counter)
       {
         oprcnt_counter = prcnt_counter;
         if(!ShowPercentInWnd(progress_wnd,prcnt_counter)) break;
       }
     }
     dis_exit:
     PFREE(codebuff);
     fclose(fout);
     if(file_cache) PFREE(file_cache);
     if(tmp_buff2) PFREE(tmp_buff2);
     if(flags & FSDLG_STRUCTS)
     {
       if(detectedFormat->drop_structs) detectedFormat->drop_structs();
       if(detectedFormat->set_state) detectedFormat->set_state(PS_INACTIVE);
     }
     if(activeMode != &disMode) disMode.term();
   }
   Exit:
   CloseWnd(progress_wnd);
   BMSeek(cpos,BM_SEEK_SET);
  }
  else  ErrMessageBox("Start > end position!",NULL);
 }
 PFREE(tmp_buff);
 DumpMode = 0;
 return 0;
}

int FRestore( void )
{
 unsigned long endpos,cpos,flags;
 int ret;
 ret = 0;
 flags = FSDLG_NOMODES;
 if(GetFStoreDlg(" Restore information from File ",ff_fname,&flags,&ff_startpos,&ff_len,FILE_PRMT))
 {
   unsigned long flen,lval;
   int handle;
   endpos = ff_startpos + ff_len;
   handle = __OsOpen(ff_fname,FO_READONLY | SO_DENYNONE);
   if(handle == -1) handle = __OsOpen(ff_fname,FO_READONLY | SO_COMPAT);
   if(handle == -1) goto err;
   flen = __FileLength(handle);
   __OsClose(handle);
   lval = endpos - ff_startpos;
   endpos = lval > flen ? flen + ff_startpos : endpos;
   endpos = endpos > BMGetFLength() ? BMGetFLength() : endpos;
   if(endpos > ff_startpos)
   {
     unsigned long wsize,cwpos;
     unsigned short remaind;
     void *tmp_buff;
     handle = __OsOpen(ff_fname,FO_READONLY | SO_DENYNONE);
     if(handle == -1) handle = __OsOpen(ff_fname,FO_READONLY | SO_COMPAT);
     if(handle == -1)
     {
        err:
        errnoMessageBox("Can not open file",NULL,errno);
        return 0;
     }
     cpos = BMGetCurrFilePos();
     wsize = endpos - ff_startpos;
     cwpos = ff_startpos;
     __OsSeek(handle,0L,SEEKF_START);
     tmp_buff = PMalloc(4096);
     if(!tmp_buff)
     {
       MemOutBox("temporary buffer initialization");
       return 0;
     }
     while(wsize)
     {
       remaind = (unsigned)min(wsize,4096);
       if(__OsRead(handle,tmp_buff,remaind) != remaind)
       {
         errnoMessageBox("Can not read from file",NULL,errno);
         __OsClose(handle);
         return 1;
       }
       BMWriteBuffEx(cwpos,BM_SEEK_SET,(void *)tmp_buff,remaind);
       wsize -= remaind;
       cwpos += remaind;
     }
     PFREE(tmp_buff);
     __OsClose(handle);
     BMSeek(cpos,BM_SEEK_SET);
     ret = 1;
   }
   else ErrMessageBox("Start > end position!",NULL);
 }
 return ret;
}

static void __NEAR__ CryptFunc(char * buff,int len,char *pass)
{
  char ch,cxor;
  int i,j;
  unsigned short bigkey_idx;
  int passlen;
  char big_key[UCHAR_MAX];
  cxor = 0;
  passlen = strlen(pass);
  memset(big_key,0,sizeof(big_key));
  for(j = 0;j < passlen;j++) cxor += pass[j]+j;
  cxor ^= passlen + len;
  for(j = i = 0;i < UCHAR_MAX;i++,j++)
  {
    if(j > passlen) j = 0;
    bigkey_idx = (pass[j] + i) * cxor;
    if(bigkey_idx > UCHAR_MAX) bigkey_idx = (bigkey_idx & 0xFF) ^ ((bigkey_idx >> 8) & 0xFF);
    big_key[i] = bigkey_idx;
  }
  for(bigkey_idx = j = i = 0;i < len;i++,bigkey_idx++,j++)
  {
   unsigned short xor;
   if(bigkey_idx > UCHAR_MAX)
   {
     // rotate of big key
     ch = big_key[0];
     memmove(big_key,&big_key[1],UCHAR_MAX-1);
     big_key[UCHAR_MAX-1] = ch;
     bigkey_idx = 0;
   }
   if(j > passlen) j = 0;
   xor = (big_key[bigkey_idx] + i)*cxor;
   if(xor > UCHAR_MAX) xor = (xor & 0xFF) ^ ((xor >> 8) & 0xFF);
   buff[i] = buff[i] ^ xor;
  }
  // rotate of pass
  ch = pass[0];
  memmove(pass,&pass[1],passlen-1);
  pass[passlen-1] = ch;
}

int CryptBlock( void )
{
 unsigned long endpos,cpos,flags;
 char pass[81];
 int ret;
 ret = 0;
 ff_startpos = BMGetCurrFilePos();
 if(!ff_len) ff_len = BMGetFLength() - ff_startpos;
 pass[0] = 0;
 flags = FSDLG_NOMODES;
 if(GetFStoreDlg(" (De)Crypt block of file ",pass,&flags,&ff_startpos,&ff_len,"Input password (WARNING! password will displayed):"))
 {
   unsigned long flen,lval;
   endpos = ff_startpos + ff_len;
   flen = BMGetFLength();
   lval = endpos - ff_startpos;
   endpos = lval > flen ? flen + ff_startpos : endpos;
   endpos = endpos > BMGetFLength() ? BMGetFLength() : endpos;
   if(!pass[0]) { ErrMessageBox("Password can not be empty",NULL); return 0; }
   if(endpos > ff_startpos)
   {
     unsigned long wsize,cwpos;
     unsigned short remaind;
     void *tmp_buff;
     cpos = BMGetCurrFilePos();
     wsize = endpos - ff_startpos;
     cwpos = ff_startpos;
     tmp_buff = PMalloc(4096);
     if(!tmp_buff)
     {
       MemOutBox("temporary buffer initialization");
       return 0;
     }
     BMSeek(ff_startpos,SEEK_SET);
     while(wsize)
     {
       remaind = (unsigned)min(wsize,4096);
       if(!BMReadBuffer(tmp_buff,remaind))
       {
         errnoMessageBox("Can not read file",NULL,errno);
         return 1;
       }
       CryptFunc(tmp_buff,remaind,pass);
       if(!(BMWriteBuffEx(cwpos,BM_SEEK_SET,(void *)tmp_buff,remaind)))
       {
         errnoMessageBox("Can not write to file",NULL,errno);
         return 1;
       }
       wsize -= remaind;
       cwpos += remaind;
     }
     PFREE(tmp_buff);
     BMSeek(cpos,BM_SEEK_SET);
     ret = 1;
   }
   else ErrMessageBox("Start > end position!",NULL);
 }
 return ret;
}

extern char shortname[];

int FileInfo( void )
{
  TWindow* wnd;
  struct stat statbuf;
  char attr[16];
  memset(&statbuf,0,sizeof(struct stat));
  fstat(BMHandle(),&statbuf);
  memset(attr,'-',15);
  attr[15] = 0;
#ifdef S_IXOTH /* Execute by other */
  if((statbuf.st_mode & S_IXOTH) == S_IXOTH) attr[15] = 'X';
#endif
#ifdef S_IWOTH /* Write by other */
  if((statbuf.st_mode & S_IWOTH) == S_IWOTH) attr[14] = 'W';
#endif
#ifdef S_IROTH /* Read by other */
  if((statbuf.st_mode & S_IROTH) == S_IROTH) attr[13] = 'R';
#endif
#ifdef S_IXGRP /* Execute by group */
  if((statbuf.st_mode & S_IXGRP) == S_IXGRP) attr[12] = 'X';
#endif
#ifdef S_IWGRP /* Write by group */
  if((statbuf.st_mode & S_IWGRP) == S_IWGRP) attr[11] = 'W';
#endif
#ifdef S_IRGRP /* Read by owner */
  if((statbuf.st_mode & S_IRGRP) == S_IRGRP) attr[10] = 'R';
#endif
#ifdef S_IEXEC /* Execute by owner */
  if((statbuf.st_mode & S_IEXEC) == S_IEXEC) attr[9] = 'X';
#endif
#ifdef S_IWRITE /* Write by owner */
  if((statbuf.st_mode & S_IWRITE) == S_IWRITE) attr[8] = 'W';
#endif
#ifdef S_IREAD /* Read by owner */
  if((statbuf.st_mode & S_IREAD) == S_IREAD) attr[7] = 'R';
#endif
#ifdef S_ISVTX /* Save swapped text after use (obsolete) */
  if((statbuf.st_mode & S_ISVTX) == S_ISVTX) attr[6] = 'V';
#endif
#ifdef S_ISGID /* Set GID on execution */
  if((statbuf.st_mode & S_ISGID) == S_ISGID) attr[5] = 'G';
#endif
#ifdef S_ISUID /* Set UID on execution */
  if((statbuf.st_mode & S_ISUID) == S_ISUID) attr[4] = 'U';
#endif
#ifdef S_ISFIFO /* it FIFO */
  if(S_ISFIFO(statbuf.st_mode)) attr[3] = 'F';
#endif
#ifdef S_IFCHR /* it character device */
  if((statbuf.st_mode & S_IFCHR) == S_IFCHR) attr[2] = 'C';
#endif
#ifdef S_IFDIR /* it directory */
  if((statbuf.st_mode & S_IFDIR) == S_IFDIR) attr[1] = 'D';
#endif
#ifdef S_IFBLK /* it block device */
  if((statbuf.st_mode & S_IFBLK) == S_IFBLK) attr[2] = 'B';
#endif
#ifdef S_IFREG /* it regular file (not dir or node) */
  if((statbuf.st_mode & S_IFREG) == S_IFREG) attr[0] = 'F';
#endif
#ifdef S_IFLNK /* it symbolic link */
  if((statbuf.st_mode & S_IFLNK) == S_IFLNK) attr[0] = 'L';
#endif
#ifdef S_IFSOCK /* it socket */
  if((statbuf.st_mode & S_IFSOCK) == S_IFSOCK) attr[0] = 'S';
#endif
  wnd = CrtHlpWnd(" File information: ",tvioWidth-18,12);
  twGotoXY(1,1);
  twPrintF("Name                          = %s\n"
           "Type                          = %s\n"
           "Length                        = %lu bytes\n"
           "Attributes                    = %s\n"
           "Creation time                 = %s"
           "Modification time             = %s"
           "Last access time              = %s"
           "Device containing file        = %u\n"
           "File serial number            = %u\n"
           "Number of hard links to file  = %u\n"
           "User ID of the file owner     = %u\n"
           "Group ID of the file owner    = %u"
           ,shortname
           ,detectedFormat->name ? detectedFormat->name : FType == OLD_EXE ? "MZ" : FType == DOS_SYS ? "DOSdriver" : "Not detected"
           ,BMGetFLength()
           ,attr
           ,ctime(&statbuf.st_ctime)
           ,ctime(&statbuf.st_mtime)
           ,ctime(&statbuf.st_atime)
           ,statbuf.st_dev
           ,statbuf.st_ino
           ,statbuf.st_nlink
           ,statbuf.st_uid
           ,statbuf.st_gid);
  while(GetEvent(drawEmptyPrompt,wnd) != KE_ESCAPE);
  CloseWnd(wnd);
  return 0;
}

static char * fu_names[] =
{
  "~File information...",
  "C~hange size of file",
  "~Save block as...",
  "~Restore block from...",
  "~Insert/delete block...",
  "~Crypt/decrypt block..."
};

typedef int (*FileFunc)( void );

static FileFunc fu_funcs[] =
{
  FileInfo,
  ChSize,
  FStore,
  FRestore,
  InsDelBlock,
  CryptBlock
};

int FileUtils( void )
{
  unsigned nUtils,i;
  int ret;
  static unsigned def_sel = 0;
  nUtils = sizeof(fu_names)/sizeof(char *);
  i = SelBoxA(fu_names,nUtils," File utilities: ",def_sel);
  if(i != -1)
  {
     TWindow * w;
     w = PleaseWaitWnd();
     ret = (*fu_funcs[i])();
     CloseWnd(w);
     def_sel = i;
     return ret;
  }
  return 0;
}
   