/****************************************************************************\
*   FILE_INI - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/file_ini.c                                     *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain low level implementation of .ini files services.        *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include "bbio.h"
#include "file_ini.h"
#include "biewlib.h"
#include "pmalloc.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

static int __NEAR__ is_eof(BGLOBAL h) { return bioTell(h) >= bioFLength(h); }

#define _BELL 0x07
#define _CR   0x0D
#define _LF   0x0A
#define __EOF 0x1A

#define __FI_MAXFILES 200

unsigned int  FiMaxStrLen = 255;
unsigned int  *FinCurrString;
char ** FiFileNames;
unsigned int  FiFilePtr = 0;
char *  FiUserMessage = NULL;

char *GETS(char *str,unsigned num,BGLOBAL h)
{
  unsigned i;
  char ch;
  char *ret;
  ret = str;
  for(i = 0;i < num;i++)
  {
     ch = bioReadByte(h);
     if(ch == '\n' || ch == '\r')
     {
       *str = ch;
       str++;
       ch = bioReadByte(h);
       if(ch == '\n' || ch == '\r')
       {
         *str = ch;
         str++;
       }
       else
       {
         if(is_eof(h))
         {
           if(ch != -1)
           {
             *str = ch;
             str++;
           }
           break;
         }
         bioSeek(h,-1,SEEK_CUR);
       }
       break;
     }
     if(is_eof(h))
     {
       if(ch != -1)
       {
         *str = ch;
         str++;
       }
       break;
     }
     *str = ch;
     str++;
  }
  *str = 0;
  return ret;
}

void FiAError(int nError,int row)
{
 int eret = 0;
 if(FiError) eret = (*FiError)(nError,row);
 if(eret == __FI_EXITPROC) exit(255);
}

static char * list[] = {
 "No errors detected.",
 "Can not open [.ini] file ( bad '#include' statement? ).",
 "Bad file handler ( bad '#include' statement? ).",
 "Too many opened files ( there are not enough handles ).",
 "Attempt to read at the end of file.",
 "Not enough memory (attempt to make most single expression in variables ).",
 "Opened 'if' condition ( '#endif' not found ).",
 "Unexpected 'if' condition for 'endif' statement.",
 "Unexpected 'if' condition for 'else' statement.",
 "Unrecognized command in '#' directive.\n          must be one of : if, else, elif, endif, include, set, delete, reset,\n          uppercase, lowercase, case, error, smart, nosmart.\n",
 "Unrecognized condition in 'if' statement\n          must be one of : ==, !=.\n",
 "Bad string constant : missing brackets \" string \".\n",
 "Expected opened section or subsection ( bracket : ], > missing or mismatch ).\n",
 "Expected bad character on line ( possible lost comment ).",
 "Detected bad variable in 'set' or 'delete' statement.",
 "Detected bad value of variable in 'set' statement.",
 "Unrecognized name of variable in 'delete' statement.",
 "Detected undefined variable ( case sensivity? ).",
 "Unexpected 'if' condition for 'elif' statement.",
 "Detected opened variable on line ( use even number of '%' characters ).",
 "Lost or mismatch character '=' in assigned expression.",
 "",
 "User error directive."
};

char * FiDecodeError(int nError)
{
 nError = abs(nError);
 if(nError >= 0 && nError <= 22) return (char *)list[nError];
 else return "Unknown Error";
}

static int StdError(int ne,int row)
{
  if(ne != __FI_READ_AT_END)
  {
    FILE * herr;
    char * what;
    herr = fopen("fi_syserr.$$$","wt");
    if(herr == NULL) herr = stderr;
    fprintf(herr,"About : [.Ini] file run-time support library. Written by N.Kurshev\n");
    fprintf(herr,"Detected ");
    if(ne != __FI_BADHANDLER && ne != __FI_TOOMANY && FiFilePtr)
    {
      fprintf(herr,"%s error in : %s\n",row ? "fatal" : "",FiFileNames[FiFilePtr - 1]);
    }
    else fprintf(herr,"\n");
    if(row)
    {
       fprintf(herr,"At line : %i\n",row);
    }
    what = FiDecodeError(ne);
    fprintf(herr,"Message : %s\n",what);
    if(FiUserMessage)
    {
      fprintf(herr,"User message : %s\n",FiUserMessage);
    }
    fclose(herr);
    printm("\nError in .ini file.\nFile fi_syser.$$$ created.\n");
    return __FI_EXITPROC;
  }
  return __FI_IGNORE;
}

int (*FiError)(int nError,int row) = StdError;

FiHandler FiOpen( const char * filename)
{
  FiHandler ret;
  char * ActiveFile;
  ret = bioOpen(filename,FO_READONLY | SO_DENYWRITE,0x1000);
  ActiveFile = (char *)PMalloc((strlen(filename) + 1));
  if(ActiveFile == NULL) FiAError(__FI_NOTMEM,0);
  strcpy(ActiveFile,filename);
  if(!FiFilePtr)
  {
    FiFileNames = PMalloc(sizeof(char *));
    FinCurrString = PMalloc(sizeof(unsigned int));
  }
  else
  {
    FiFileNames = PRealloc(FiFileNames,sizeof(char *)*(FiFilePtr+1));
    FinCurrString = PRealloc(FinCurrString,sizeof(unsigned int)*(FiFilePtr+1));
  }
  if(!FiFileNames || !FinCurrString) FiAError(__FI_NOTMEM,0);
  FiFileNames[FiFilePtr] = ActiveFile;
  FinCurrString[FiFilePtr++] = 0;
  if(FiFilePtr > __FI_MAXFILES-1) FiAError(__FI_TOOMANY,0);
  if(ret == &bNull )
  {
    FiAError(__FI_BADFILENAME,0);
  }
  return ret;
}

void FiClose(FiHandler h)
{
  PFREE(FiFileNames[FiFilePtr-1]);
  if(FiFilePtr)
  {
    FiFileNames = PRealloc(FiFileNames,sizeof(char *)*(FiFilePtr));
    FinCurrString = PRealloc(FinCurrString,sizeof(unsigned int)*(FiFilePtr));
    if(!FiFileNames || !FinCurrString) FiAError(__FI_NOTMEM,0);
    FiFilePtr--;
  }
  else
  {
    PFREE(FiFileNames);
    PFREE(FinCurrString);
  }
  bioClose(h);
}

void FiSeekTo( FiHandler h, int nSection, int nSubSection, int nItem )
{
  int ns = 0,nss = 0,ni = 0;
  bioSeek(h,0L,SEEKF_START);
  FinCurrString[FiFilePtr - 1] = 0;
  if(nSection)
  {
    char * buff;
    buff = __FiCMaxStr();
    while(!is_eof(h))
    {
      FiGetNextString(h,buff,FiMaxStrLen);
      if(FiisSection(buff)) ns++;
      if(ns < nSection) continue;
    }
    PFREE(buff);
  }
  if(nSubSection)
  {
   char * buff;
   buff = __FiCMaxStr();
   while(!is_eof(h))
   {
     FiGetNextString(h,buff,FiMaxStrLen);
     if(FiisSubSection(buff)) nss++;
     if(nss < nSubSection) continue;
   }
   PFREE(buff);
  }
  if(nItem)
  {
   char * buff;
   buff = __FiCMaxStr();
   while(!is_eof(h))
   {
     FiGetNextString(h,buff,FiMaxStrLen);
     if(FiisItem(buff)) ni++;
     if(nss < nItem) continue;
   }
   PFREE(buff);
  }
}

int FiSearch(FiHandler h, const char * str )
{
  char *tmp,*buff;
  tmp  = __FiCMaxStr();
  buff = __FiCMaxStr();
  bioSeek(h,0L,SEEKF_START);
  FinCurrString[FiFilePtr - 1] = 0;
  while(!is_eof(h))
  {
   FiGetNextString(h,buff,FiMaxStrLen);
   if(FiisSection(buff))
   {
     FiGetSectionName(buff,tmp);
     if(strcmp(tmp,str) == 0) { PFREE(tmp); PFREE(buff); return __FI_SECTION;}
   }
   if(FiisSubSection(buff))
   {
     FiGetSubSectionName(buff,tmp);
     if(strcmp(tmp,str) == 0) { PFREE(buff); PFREE(tmp); return __FI_SUBSECTION; }
   }
   if(FiisItem(buff))
   {
     FiGetItemName(buff,tmp);
     if(strcmp(tmp,str) == 0) { PFREE(tmp); PFREE(buff); return __FI_ITEM; }
   }
  }
  bioSeek(h,0L,SEEKF_START);
  FinCurrString[FiFilePtr - 1] = 0;
  PFREE(tmp);
  PFREE(buff);
  return __FI_NOTFOUND;
}

unsigned int  FiGetNumberOfSections( FiHandler h)
{
 long curr;
 int eret;
 unsigned int save;
 char * buff;
 eret = 0;
 save = FinCurrString[FiFilePtr -1];
 curr = bioTell(h);
 buff = __FiCMaxStr();
 bioSeek(h,0L,SEEKF_START);
 while(!is_eof(h))
 {
   FiGetNextString(h,buff,FiMaxStrLen);
   if(FiisSection(buff)) eret++;
 }
 bioSeek(h,curr,SEEKF_START);
 FinCurrString[FiFilePtr - 1] = save;
 PFREE(buff);
 return eret;
}

unsigned int FiGetTotalNumberOfSubSections( FiHandler h )
{
 long curr;
 int eret;
 unsigned int save;
 char * buff;
 eret = 0;
 curr = bioTell(h);
 buff = __FiCMaxStr();
 save = FinCurrString[FiFilePtr - 1];
 bioSeek(h,0L,SEEKF_START);
 while(!is_eof(h))
 {
   FiGetNextString(h,buff,FiMaxStrLen);
   if(FiisSubSection(buff)) eret++;
 }
 bioSeek(h,curr,SEEKF_START);
 FinCurrString[FiFilePtr - 1] = save;
 PFREE(buff);
 return eret;
}
unsigned int FiGetLocalNumberOfSubSections( FiHandler h, int nSection )
{
  long curr;
  int eret;
  unsigned int save;
  char * buff;
  save = FinCurrString[FiFilePtr - 1];
  curr = bioTell(h);
  eret = 0;
  buff = __FiCMaxStr();
  FiSeekTo(h,nSection,0,0);
  while(!is_eof(h))
  {
    FiGetNextString(h,buff,FiMaxStrLen);
    if(FiisItem(buff)) eret++;
  }
  bioSeek(h,curr,SEEKF_START);
  FinCurrString[FiFilePtr - 1] = save;
  PFREE(buff);
  return eret;
}

unsigned int  FiGetTotalNumberOfItems( FiHandler h)
{
 long curr;
 int eret;
 unsigned int save;
 char * buff;
 eret = 0;
 curr = bioTell(h);
 save = FinCurrString[FiFilePtr - 1];
 buff = __FiCMaxStr();
 bioSeek(h,0L,SEEKF_START);
 while(!is_eof(h))
 {
   FiGetNextString(h,buff,FiMaxStrLen);
   if(FiisItem(buff)) eret++;
 }
 bioSeek(h,curr,SEEKF_START);
 FinCurrString[FiFilePtr - 1] = save;
 PFREE(buff);
 return eret;
}

unsigned int  FiGetLocalNumberOfItems( FiHandler h,int nSection , int nSubSection)
{
  long curr;
  int eret;
  unsigned int save;
  char * buff;
  save = FinCurrString[FiFilePtr - 1];
  curr = bioTell(h);
  eret = 0;
  buff = __FiCMaxStr();
  FiSeekTo(h,nSection,nSubSection,0);
  while(!is_eof(h))
  {
    FiGetNextString(h,buff,FiMaxStrLen);
    if(FiisItem(buff)) eret++;
  }
  bioSeek(h,curr,SEEKF_START);
  FinCurrString[FiFilePtr - 1] = save;
  PFREE(buff);
  return eret;
}

char FiOpenComment = ';';

/*
unsigned int FiGetLengthNextString(FiHandler h)
{
  int len,i;
  char *sret;
  char inRemark;
  unsigned long where;
  where = bioTell(h);
  len = 0;
  inRemark = 0;
  while(!feof(h))
  {
    int rd;
    rd = fgetc(h);
    if((rd == EOF && !feof( h )))  FiAError(__FI_BADFILENAME,0);
    if(inRemark && (rd == _CR || rd == _LF)) { inRemark = 0; continue; }
    if(rd == FiOpenComment) inRemark = 1;
    if(rd == _CR || rd == _LF) continue;

    if(!inRemark && (rd == _CR || rd == _LF)) break;
    if(!inRemark && rd != _CR && rd != _LF) len++;
  }
  bioSeek(h,SEEKF_START,where);
  return len;
}
*/

char  FiAllWantInput = 0;

char * FiGetNextString(FiHandler h, char * str,unsigned int size)
{
  int len,i;
  char *sret;
  while(!is_eof(h))
  {
    str[0] = 0;
    sret = GETS(str,size,h);
    FinCurrString[FiFilePtr - 1]++;
    if((sret == NULL && !is_eof( h )))  FiAError(__FI_BADFILENAME,0);
    len = strlen(str);
    for(i = 0;i < len;i++) if(str[i] == FiOpenComment) { str[i] = '\0'; break; }
    len = strlen(str);
    if(!FiAllWantInput)
    {
      for(i = len - 1;i >= 0;i--)
      {
        if(str[i] == ' ' || str[i] == _CR || str[i] == _LF) str[i] = '\0';
        else break;
      }
      if(strlen(str))  break; // loop while comment
    }
  }
  return str;
}

static int __NEAR__ __isSect(const char *str,char f1)
{
  int i;
  char Flag,legal;
  Flag = 0;
  legal = 1;
  for(i = 0;i < strlen(str);i++)
  {
     char c;
     c = str[i];
     if(c != ' ' &&  c!= f1) legal = 0;
     if(c == '=') break;
     if(c == f1) { Flag++; break; }
  }
  if(Flag && !legal) FiAError(__FI_BADCHAR,FinCurrString[FiFilePtr - 1]);
  return Flag;
}

int       FiisSection( const char * str )
{
  return __isSect(str,'[');
}

int       FiisSubSection( const char * str )
{
  return __isSect(str,'<');
}

int       FiisCommand( const char * str )
{
  return __isSect(str,'#');
}

int       FiisItem( const char *str)
{
  return !FiisSection(str) && !FiisSubSection(str) && !FiisCommand(str);
}

static unsigned int __NEAR__ __GetLengthBrStr(const char * src,char obr,char cbr)
{
  STRING str_;
  int i,new_iptr;
  char found;
  char ill[3];
  char * ctrl;
   ill[0] = ' ';
   ill[1] = obr;
   ill[2] = '\0';
   str_.iptr = 0;
   str_.str = (char *)src;
   FiSkipWhiteSpace(&str_,ill);
   found = 0;
   for(i = 0;;i++)
   {
      char c;
      c = str_.str[str_.iptr++];
      if(c == cbr) { found = 1; break; }
      if(c == '\0') break;
   }
   if(!found) FiAError(__FI_OPENSECT,FinCurrString[FiFilePtr - 1]);
   new_iptr = str_.iptr;
   ctrl = __FiCNWord(&str_," ");
   if(ctrl != NULL && ctrl[0] != '\0') FiAError(__FI_BADCHAR,FinCurrString[FiFilePtr - 1]);
   PFREE(ctrl);
   str_.iptr = new_iptr;
   str_.iptr--;
   str_.iptr--;
   for(;i;)
   {
     if(isspace(str_.str[str_.iptr--])) i--;
     else break;
   }
   return i;
}

static char * __NEAR__ __GetBrStrName(const char * src,char * store,char obr,char cbr)
{
 STRING str_;
 char ill[3];
 int i,storelen;
 char found;
   storelen = __GetLengthBrStr(src,obr,cbr);
   str_.iptr = 0;
   str_.str = (char *)src;
   ill[0] = ' ';
   ill[1] = obr;
   ill[2] = '\0';
   FiSkipWhiteSpace(&str_,ill);
   found = 0;
   for(i = 0;i < storelen;i++)
   {
      char c;
      c = str_.str[str_.iptr++];
      if(c == cbr) { found = 1; break; }
      if(c == '\0') break;
      store[i] = c;
   }
   if(!found)
   {
     if(cbr == obr) ill[1] = '\0';
     FiSkipWhiteSpace(&str_,ill);
     if(str_.str[str_.iptr] == cbr) found = 1;
   }
   store[i] = 0;
   if(!found) FiAError(__FI_OPENSECT,FinCurrString[FiFilePtr - 1]);
   return store;
}

unsigned int  FiGetLengthBracketString( const char * str )
{
  return __GetLengthBrStr(str,'"','"');
}

char * FiGetBracketString(const char * str,char * store)
{
  return __GetBrStrName(str,store,'"','"');
}

char *    FiGetSectionName(const char * src,char * store)
{
  return __GetBrStrName(src,store,'[',']');
}

char *    FiGetSubSectionName(const char * src,char * store)
{
  return __GetBrStrName(src,store,'<','>');
}

unsigned int  FiGetLengthSection( const char * src )
{
  return __GetLengthBrStr(src,'[',']');
}

unsigned int  FiGetLengthSubSection( const char * src )
{
  return __GetLengthBrStr(src,'<','>');
}

unsigned int  FiGetLengthItem( const char * src )
{
  STRING str;
  str.iptr = 0;
  str.str = (char *)src;
  return FiGetLengthNextWord(&str," ");
}

unsigned int  FiGetLengthValue( const char * src )
{
  STRING str;
  unsigned int new_iptr,siptr;
  str.iptr = 0;
  str.str = (char *)src;
  FiSkipWhiteSpace(&str," ");
  siptr = str.iptr;
  new_iptr = FiGetLengthNextWord(&str," =");
  str.iptr = new_iptr + siptr;
  FiSkipWhiteSpace(&str," =");
  return strlen(src) - str.iptr;
}

char *    FiGetValueOfItem(const char * src,char * store)
{
  STRING str_;
  char * b1;
  b1 = __FiCMaxStr();
  str_.iptr = 0;
  str_.str = (char *)src;
  FiGetNextWord(&str_," =",b1);
  FiSkipWhiteSpace(&str_," ");
  if(str_.str[str_.iptr] != '=') FiAError(__FI_NOTEQU,FinCurrString[FiFilePtr - 1]);
  str_.iptr++;
  FiSkipWhiteSpace(&str_," ");
  strcpy(store,&str_.str[str_.iptr]);
  PFREE(b1);
  return store;
}

unsigned int  FiGetLengthCommandString( const char * src )
{
   STRING str;
   str.iptr = 0;
   str.str = (char *)src;
   FiSkipWhiteSpace(&str," #");
   return strlen(src) - str.iptr;
}

char *    FiGetItemName(const char * src,char * store)
{
  STRING str_;
   str_.iptr = 0;
   str_.str = (char *)src;
   str_.str = FiGetNextWord(&str_," =",store);
   return store;
}

char *   FiGetCommandString(const char * src,char * store)
{
  STRING str_;
   str_.iptr = 0;
   str_.str = (char *)src;
   str_.str = FiSkipWhiteSpace(&str_," #");
   strcpy(store,str_.str);
   return store;
}

unsigned int  FiGetLengthNextWord( STRING * str, const char * illegal_symbols)
{
 unsigned int iptr;
 char c;
 int j = 0;
 unsigned slen;
 iptr = str->iptr;
 c = str->str[str->iptr];
 FiSkipWhiteSpace(str,illegal_symbols);
 slen = strlen(str->str);
 while(str->iptr < slen)
 {
   c = str->str[str->iptr++];
   if(c == 0) break;
   if(!FiisLegal(illegal_symbols,c)) break;
   j++;
 }
 str->iptr = iptr;
 return j;
}

char * FiGetNextWord( STRING * str,const char * illegal_symbols,char * store)
{
 char c = str->str[str->iptr];
 int j = 0;
 unsigned slen;
 FiSkipWhiteSpace(str,illegal_symbols);
 slen = strlen(str->str);
 while(str->iptr < slen)
 {
   c = str->str[str->iptr];
   if(c == 0) break;
   if(!FiisLegal(illegal_symbols,c)) break;
   store[j++] = str->str[str->iptr++];
 }
 store[j++] = '\0';
 return store;
}

char * FiSkipWhiteSpace( STRING * str, const char * illegal_symbols )
{
  char c;
  unsigned slen;
  slen = strlen(str->str);
  while(str->iptr < slen)
  {
    c = str->str[str->iptr];
    if(FiisLegal(illegal_symbols,c)) break;
    str->iptr++;
  }
  return &str->str[str->iptr];
}

int FiisLegal(const char * illegal,char c)
{
  int i = 0;
  while(illegal[i])
  {
    if( illegal[i] == ' ' ) if(isspace(c)) return 0;
    if( illegal[i++] == c ) return 0;
  }
  return 1;
}  