/****************************************************************************\
*   FIPROGRS - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/fiprogrs.c                                     *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain middle level implementation of .ini files services.     *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>

#include "pmalloc.h"
#include "file_ini.h"
#include "_sys_dep.h"

static int __NEAR__ is_eof(BGLOBAL h) { return bioTell(h) >= bioFLength(h); }

static char CaseSens = 2; // 2 - case 1 - upper 0 - lower
static FiUserFunc proc;
static pVar FirstVar = NULL;
FiHandler ActiveFile = 0;
static char Smarting = 1;

/************** BEGIN of Construct section ********************/

char * __FiCMaxStr( void )
{
  char * ret;
  ret = (char *)PMalloc(FiMaxStrLen + 1);
  if(ret == NULL) FiAError(__FI_NOTMEM,0);
  return ret;
}

char * __FiCNWord( STRING *str , const char * illegal_set)
{
  char * ret;
  int lword;
  lword = FiGetLengthNextWord(str,illegal_set);
  ret = (char *)PMalloc(lword + 1);
  if(ret == NULL) FiAError(__FI_NOTMEM,0);
  if(lword) FiGetNextWord(str,illegal_set,ret);
  else      ret[0] = 0;
  return ret;
}

char * __FiCBString( const char * src )
{
  char * ret;
  int lbr;
  lbr = FiGetLengthBracketString(src);
  ret = (char *)PMalloc(lbr + 1);
  if(ret == NULL) FiAError(__FI_NOTMEM,0);
  if(lbr) FiGetBracketString(src,ret);
  else    ret[0] = 0;
  return ret;
}

char * __FiCItem( const char * src )
{
  char * ret;
  int li;
  li = FiGetLengthItem(src);
  ret = (char *)PMalloc(li + 1);
  if(ret == NULL) FiAError(__FI_NOTMEM,0);
  if(li) FiGetItemName(src,ret);
  else   ret[0] = 0;
  return ret;
}

char * __FiCValue( const char * src )
{
  char * ret;
  int lv;
  lv = FiGetLengthValue(src);
  ret = (char *)PMalloc(lv + 1);
  if(ret == NULL) FiAError(__FI_NOTMEM,0);
  if(lv) FiGetValueOfItem(src,ret);
  else   ret[0] = 0;
  return ret;
}

char * __FiCSection( const char * src )
{
  char * ret;
  int ls;
  ls = FiGetLengthSection(src);
  ret = (char *)PMalloc(ls + 1);
  if(ret == NULL) FiAError(__FI_NOTMEM,0);
  if(ls) FiGetSectionName(src,ret);
  else   ret[0] = 0;
  return ret;
}

char * __FiCSubSection( const char * src )
{
  char * ret;
  int lss;
  lss = FiGetLengthSubSection(src);
  ret = (char *)PMalloc(lss + 1);
  if(ret == NULL) FiAError(__FI_NOTMEM,0);
  if(lss) FiGetSubSectionName(src,ret);
  else    ret[0] = 0;
  return ret;
}

char * __FiCCmd( const char * src )
{
  char * ret;
  int lc;
  lc = FiGetLengthCommandString(src);
  ret = (char *)PMalloc(lc + 1);
  if(ret == NULL) FiAError(__FI_NOTMEM,0);
  if(lc) FiGetCommandString(src,ret);
  else   ret[0] = 0;
  return ret;
}
/************* END of Construct section *********************/

/************* BEGIN of List Var section ********************/

pVar FiConstructVar(const char *v,const char *a)
{
 char * str;
 pVar vv;
 vv = (pVar)PMalloc(sizeof(Var));
 if(vv == NULL) FiAError(__FI_NOTMEM,0);
 vv->next = NULL;
 vv->prev = NULL;
 str = (char *)PMalloc(strlen(v) + 1);
 if(str == NULL) FiAError(__FI_NOTMEM,0);
 strcpy(str,v);
 vv->variables = str;
 str = (char *)PMalloc(strlen(a) + 1);
 if(str == NULL) FiAError(__FI_NOTMEM,0);
 strcpy(str,a);
 vv->associate = str;
 return vv;
}

void FiDeleteVar(pVar pp)
{
  PFREE(pp->variables);
  PFREE(pp->associate);
  PFREE(pp);
}

void FiDeleteAllVar( void )
{
  pVar iter;
  if(FirstVar == NULL) return;
  iter = FirstVar;
  while( iter!= 0 )
  {
    pVar temp;
    temp = iter->next;
    FiDeleteVar(iter);
    iter = temp;
  }
}

const char * FiExpandVariables(const char * var)
{
 char *ret = 0;
 pVar iter;
 iter = FirstVar;
 if(FirstVar != NULL)
 {
  for(;;)
  {
   if(strcmp(var,iter->variables) == 0)
   {
     ret = iter->associate;
     break;
   }
   if(iter->next == NULL) FiAError(__FI_NODEFVAR,FinCurrString[FiFilePtr-1]);
   iter = iter->next;
  }
 }
 return ret;
}

void FiExpandAllVar(const char * value,char * store)
{
  if(Smarting)
  {
    char * tmp;
    char isVar,npercent;
    unsigned char buffer_ptr,tmp_ptr;
    int i;
    npercent = 0;
    buffer_ptr = tmp_ptr = isVar = 0;
    tmp = __FiCMaxStr();
    for(i = 0;i < strlen(value);i++)
    {
        char c;
        c = value[i];
        if(c == '%')
        {
          npercent++;
          isVar = !isVar;
          if(!isVar)
          {
            tmp[tmp_ptr] = '\0';
            strcpy(&store[buffer_ptr],FiExpandVariables(tmp));
            buffer_ptr = strlen(store);
            tmp_ptr = 0;
          }
        }
        else
        {
          if(isVar)  tmp[tmp_ptr++] = c;
          else store[buffer_ptr++] = c;
        }
    }
    store[buffer_ptr] = '\0';
    if( npercent%2 ) FiAError(__FI_OPENVAR,FinCurrString[FiFilePtr - 1]);
    PFREE(tmp);
  }
  else strcpy(store,value);
}

void FiAddVariables(const char * var,const char * associate)
{
 pVar iter,prev;
 if(!FirstVar)
 {
    FirstVar = FiConstructVar(var,associate);
    FirstVar->next = NULL;
    FirstVar->prev = NULL;
 }
 else
 {
  iter = FirstVar;
  for(;;)
  {
   if(strcmp(iter->variables,var) == 0)
   {
     PFREE(iter->associate);
     iter->associate = PMalloc(strlen(associate) + 1);
     if(iter->associate == NULL) FiAError(__FI_NOTMEM,0);
     strcpy(iter->associate,associate);
     return;
   }
   if(iter->next == NULL) break;
   iter = iter->next;
  }
  prev = FiConstructVar(var,associate);
  iter->next = prev;
  prev->prev = iter;
 }
}

void FiRemoveVariables(const char * var)
{
  pVar iter,p;
  if(FirstVar == NULL) return;
  iter = FirstVar;
  for(;;)
  {
    if(strcmp(var,iter->variables) == 0)
    {
      if(iter != FirstVar)
      {
        p = iter->prev;
        p->next = iter->next;
        FiDeleteVar(iter);
        return;
      }
      else
      {
        p = iter->next;
        FiDeleteVar(iter);
        FirstVar = p;
        return;
      }
    }
    if(iter->next == NULL) break;
    iter = iter->next;
  }
  FiAError(__FI_NOVAR,FinCurrString[FiFilePtr - 1]);
}

/*************** END of List VAr Section ***************/

int FiGetConditionStd( const char *condstr)
{
  char * var,cond[3],*user_ass,*tmp;
  const char * real_ass;
  STRING str;
  int ret;

  str.iptr = 0;
  str.str = (char *)condstr;

  user_ass  = __FiCMaxStr();
  tmp       = __FiCMaxStr();

  var = __FiCNWord(&str," !=");
  FiExpandAllVar(var,tmp);
  real_ass = FiExpandVariables(tmp); // real association

  PFREE(var);
  PFREE(tmp);

  FiSkipWhiteSpace(&str," ");
  cond[0] = str.str[str.iptr++];
  cond[1] = str.str[str.iptr++];
  cond[2] = '\0';

  tmp  = __FiCNWord(&str," ");
  FiExpandAllVar(tmp,user_ass);
  PFREE(tmp);
  tmp = __FiCNWord(&str," ");
  if(tmp != NULL && tmp[0] != '\0') FiAError(__FI_BADCHAR,FinCurrString[FiFilePtr - 1]);
  PFREE(tmp);
  ret = 0;
  if(strcmp(cond,"==") == 0)  ret = (strcmp(user_ass,real_ass) == 0);
  else
   if(strcmp(cond,"!=") == 0)  ret = (strcmp(user_ass,real_ass) != 0);
   else
     FiAError(__FI_BADCOND,FinCurrString[FiFilePtr - 1]);
  PFREE(user_ass);
  return ret;
}

char cond_ret = 1;

char FiCommandProcessorStd( const char * cmd )
{
 char *word;
 STRING str;
 str.iptr = 0;
 str.str = (char *)cmd;
 word = __FiCNWord(&str," ");
 strlwr(word);
 if(strcmp(word,"include") == 0)
 {
   char * bracket;
   char * v;
   bracket = __FiCBString(&str.str[str.iptr]);
   v = __FiCMaxStr();
   FiExpandAllVar(bracket,v);
   (*FiFileProcessor)(v);
   PFREE(bracket);
   PFREE(v);
   goto Exit_CP;
 }
 if(strcmp(word,"set") == 0)
 {
   char * a;
   char * v;
   v = __FiCNWord(&str," =");
   FiSkipWhiteSpace(&str," ");
   if(str.str[str.iptr] != '=') FiAError(__FI_NOTEQU,FinCurrString[FiFilePtr - 1]);
   str.iptr++;
   a = __FiCNWord(&str," ");
   if(v[0] == '\0' || v == NULL ) FiAError(__FI_BADVAR,FinCurrString[FiFilePtr - 1]);
   if(a[0] == '\0' || a == NULL ) FiAError(__FI_BADVAL,FinCurrString[FiFilePtr - 1]);
   FiAddVariables(v,a);
   PFREE(a);
   PFREE(v);
   v = __FiCNWord(&str," ");
   if(v != NULL && v[0] != '\0') FiAError(__FI_BADCHAR,FinCurrString[FiFilePtr - 1]);
   PFREE(v);
   goto Exit_CP;
 }
 if(strcmp(word,"delete") == 0)
 {
   char * v;
   char * a;
   v = __FiCNWord(&str," ");
   a = __FiCMaxStr();
   FiExpandAllVar(v,a);
   if(a[0] == '\0' || a == NULL) FiAError(__FI_BADVAR,FinCurrString[FiFilePtr - 1]);
   FiRemoveVariables(a);
   PFREE(v);
   PFREE(a);
   v = __FiCNWord(&str," ");
   if(v != NULL && v[0] != '\0') FiAError(__FI_BADCHAR,FinCurrString[FiFilePtr - 1]);
   PFREE(v);
   goto Exit_CP;
 }
 if(strcmp(word,"reset") == 0)
 {
   FiDeleteAllVar();
   goto Exit_CP;
 }
 if(strcmp(word,"case") == 0)
 {
    CaseSens = 2;
    goto Exit_CP;
 }
 if(strcmp(word,"smart") == 0)
 {
     Smarting = 1;
     goto Exit_CP;
 }
 if(strcmp(word,"nosmart") == 0)
 {
     Smarting = 0;
     goto Exit_CP;
 }
 if(strcmp(word,"uppercase") == 0)
 {
    CaseSens = 1;
    goto Exit_CP;
 }
 if(strcmp(word,"lowercase") == 0)
 {
    CaseSens = 0;
    goto Exit_CP;
 }
 if(strcmp(word,"error") == 0)
 {
   char *v,*a;
   v = __FiCBString(&str.str[str.iptr]);
   a = __FiCMaxStr();
   FiExpandAllVar(v,a);
   PFREE(v);
   FiUserMessage = a;
   FiAError(__FI_FIUSER,FinCurrString[FiFilePtr - 1]);
 }
 if(strcmp(word,"else") == 0) FiAError(__FI_ELSESTAT,FinCurrString[FiFilePtr - 1]);
 if(strcmp(word,"endif") == 0) FiAError(__FI_IFNOTFOUND,FinCurrString[FiFilePtr - 1]);
 if(strcmp(word,"elif") == 0) FiAError(__FI_ELIFSTAT,FinCurrString[FiFilePtr - 1]);
 if(strcmp(word,"if") == 0)
 {
   char Condition,BeenElse;
   unsigned int nLabel,nsave;
   char *ondelete,*sstore;
   sstore = __FiCMaxStr();
   ondelete = sstore;
   nsave = FinCurrString[FiFilePtr - 1];
   cond_ret = Condition = (*FiGetCondition)(&cmd[str.iptr]);
   nLabel = 1;
   BeenElse = 0;
   while(!is_eof(ActiveFile))
   {
     FiGetNextString(ActiveFile,sstore,FiMaxStrLen);
     if(sstore == NULL ) { PFREE(ondelete); goto Exit_CP; }
     if(sstore[0] == '\0') { PFREE(sstore); goto Exit_CP; }
     if(FiisCommand(sstore))
     {
       STRING cStr;
       char *lword;
       char * cmdstr;
       cmdstr = __FiCCmd(sstore);
       if(CaseSens == 1) strupr(cmdstr);
       if(CaseSens == 0) strlwr(cmdstr);
       cStr.iptr = 0;
       cStr.str = cmdstr;
       lword = __FiCNWord(&cStr," ");
       strlwr(lword);
       if(strcmp(lword,"if") == 0 && !Condition) nLabel++;
       if(strcmp(lword,"endif") == 0 )
       {
         nLabel--;
         if(nLabel == 0)
         {
           PFREE(sstore);
           PFREE(cmdstr);
           PFREE(lword);
           goto Exit_CP;
         }
         if(nLabel <  0) FiAError(__FI_IFNOTFOUND,FinCurrString[FiFilePtr - 1]);
       }
       if(strcmp(lword,"else") == 0 && nLabel == 1)
       {
           if( BeenElse ) FiAError(__FI_ELSESTAT,FinCurrString[FiFilePtr - 1]);
           if( nLabel == 1 ) cond_ret = Condition = (Condition ? 0 : 1);
           if( nLabel == 1 ) BeenElse = 1;
           PFREE(lword);
           PFREE(cmdstr);
           continue;
       }
       if(strcmp(lword,"elif") == 0 && nLabel == 1)
       {
           if( BeenElse ) FiAError(__FI_ELIFSTAT,FinCurrString[FiFilePtr - 1]);
           if( nLabel == 1 ) cond_ret = Condition = (*FiGetCondition)(&cmdstr[cStr.iptr]);
           if( nLabel == 1 ) BeenElse = 1;
           PFREE(lword);
           PFREE(cmdstr);
           continue;
       }
       if(Condition) (*FiCommandProcessor)(cmdstr);
       PFREE(lword);
       PFREE(cmdstr);
     }
     else { if(Condition) (*FiStringProcessor)(sstore); }
   }
   if(sstore) PFREE(sstore)
   else PFREE(ondelete)
   FiAError(__FI_OPENCOND,nsave);
 }
 FiAError(__FI_UNRECOGN,FinCurrString[FiFilePtr - 1]);
 Exit_CP:
 PFREE(word);
 return cond_ret;
}

static  char * curr_sect = NULL;
static  char * curr_subsect = NULL;

void FiStringProcessorStd(char * curr_str)
{
    if(FiisCommand(curr_str))
    {
      char * cmd;
      cmd = __FiCCmd(curr_str);
      (*FiCommandProcessor)(cmd);
      PFREE(cmd);
      return;
    }
    if(FiisSection(curr_str))
    {
      if(curr_sect) PFREE(curr_sect);
      if(curr_subsect) PFREE(curr_subsect);
      curr_sect = __FiCSection(curr_str);
      return;
    }
    if(FiisSubSection(curr_str))
    {
      if(curr_subsect) PFREE(curr_subsect);
      curr_subsect = __FiCSubSection(curr_str);
      return;
    }
    if(FiisItem(curr_str))
    {
      char *buffer,*bufferv;
      char *item,*val;
      IniInfo info;
      buffer = __FiCMaxStr();
      bufferv = __FiCMaxStr();
      item = __FiCItem(curr_str);
      if(item[0])
      {
       val = __FiCValue(curr_str);
       FiExpandAllVar(val,buffer);
       FiExpandAllVar(item,bufferv);
       if(curr_sect) info.section = curr_sect;
       else info.section = "";
       if(curr_subsect) info.subsection = curr_subsect;
       else info.subsection = "";
       info.item = bufferv;
       info.value = buffer;
       (*proc)(&info);
       PFREE(val);
      }
      PFREE(buffer);
      PFREE(bufferv);
      PFREE(item);
      return;
    }
}

void FiFileProcessorStd(const char * filename)
{
  FiHandler h;
  FiHandler oldh;
  char * work_str, * ondelete;
  oldh = ActiveFile;
  h = FiOpen(filename);
  ActiveFile = h;
  work_str = __FiCMaxStr();
  ondelete = work_str;
  while(!is_eof(h))
  {
      FiGetNextString(h,work_str,FiMaxStrLen);
      if(!work_str || !work_str[0]) break;
      if(CaseSens == 1) strupr(work_str);
      if(CaseSens == 0) strlwr(work_str);
      (*FiStringProcessor)(work_str);
  }
  FiClose(h);
  if(work_str) PFREE(work_str)
  else PFREE(ondelete)
  ActiveFile = oldh;
}

void     FiProgress(const char * filename,FiUserFunc usrproc)
{

  if(!usrproc) return;
  proc = usrproc;

  (*FiFileProcessor)(filename);

  proc = NULL;
  if(curr_sect) PFREE(curr_sect);
  if(curr_subsect) PFREE(curr_subsect);
  FiDeleteAllVar();
}

void      (*FiFileProcessor)(const char *fname) = FiFileProcessorStd;
void      (*FiStringProcessor)(char * curr_str) = FiStringProcessorStd;
char      (*FiCommandProcessor)(const char * cmd) = FiCommandProcessorStd;
int       (*FiGetCondition)( const char *condstr) = FiGetConditionStd;
 