/****************************************************************************\
*   PMALLOC - this source file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/pmalloc.c                                      *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of preemptive memory allocation.         *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#if __BITNESS__ == 16
#include <malloc.h>
#endif
#include "pmalloc.h"

static LowMemCallBack *lmstack = NULL;
unsigned               lmcount = 0;

int              PMRegLowMemCallBack(LowMemCallBack func)
{
  void *new_ptr;
  if(lmcount > UINT_MAX-2) return 0;
  if(!lmstack) new_ptr = malloc(sizeof(LowMemCallBack));
  else         new_ptr = realloc(lmstack,(lmcount+1)*sizeof(LowMemCallBack));
  if(!new_ptr) return 0;
  lmstack = new_ptr;
  lmstack[lmcount++] = func;
  return 1;
}

int              PMUnregLowMemCallBack(LowMemCallBack func)
{
  unsigned i,fidx;
  fidx = UINT_MAX;
  for(i = 0;i < lmcount;i++)
  {
    if(lmstack[i] == func)
    {
      fidx = i;
      break;
    }
  }
  if(fidx != UINT_MAX)
  {
    void *new_ptr;
    memmove(&lmstack[fidx],&lmstack[fidx+1],lmcount);
    new_ptr = realloc(lmstack,(lmcount--)*sizeof(LowMemCallBack));
    if(new_ptr) lmstack = new_ptr;
    return 1;
  }
  return 0;
}

void *           PMalloc(size_t obj_size)
{
  void *ret;
  unsigned i;
  ret = malloc(obj_size);
  if(!ret)
  {
    for(i = 0;i < lmcount;i++)
    {
      if(lmstack[i](obj_size))
      {
        ret = malloc(obj_size);
        if(ret) break;
      }
    }
  }
  return ret;
}

void *           PRealloc(void *ptr,size_t obj_size)
{
  void *ret;
  unsigned i;
  ret = realloc(ptr,obj_size);
  if(!ret)
  {
    for(i = 0;i < lmcount;i++)
    {
      if(lmstack[i](obj_size))
      {
        ret = realloc(ptr,obj_size);
        if(ret) break;
      }
    }
  }
  return ret;
}

void             PFree(void *ptr)
{
  free(ptr);
}

#if __BITNESS__ == 16
void __HUGE__ *  PHMalloc(unsigned long obj_size)
{
  void __HUGE__ *ret;
  unsigned i;
  ret = halloc(obj_size);
  if(!ret)
  {
    for(i = 0;i < lmcount;i++)
    {
      if(lmstack[i](obj_size))
      {
        ret = halloc(obj_size);
        if(ret) break;
      }
    }
  }
  return ret;
}

void __HUGE__ *  PHRealloc(void __HUGE__ *ptr,unsigned long obj_size)
{
  void __HUGE__ *ret;
  unsigned i;
  ret = hrealloc(ptr,obj_size);
  if(!ret)
  {
    for(i = 0;i < lmcount;i++)
    {
      if(lmstack[i](obj_size))
      {
        ret = hrealloc(ptr,obj_size);
        if(ret) break;
      }
    }
  }
  return ret;
}

void             PHFree(void __HUGE__ * ptr)
{
  hfree(ptr);
}
#else
void __HUGE__ *  PHMalloc(unsigned long obj_size)
{
  return PMalloc(obj_size);
}

void __HUGE__ *  PHRealloc(void __HUGE__ *ptr,unsigned long obj_size)
{
  return PRealloc(ptr,obj_size);
}

void             PHFree(void __HUGE__ * ptr)
{
  return PFree(ptr);
}
#endif

 