/****************************************************************************\
*   __OS_DEP - this header file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/__os_dep.h                              *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain all operating system depended part of BIEW project.     *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#ifndef __OS_DEP_H
#define __OS_DEP_H

#ifdef __cplusplus
extern "C" {
#endif

/* KEYBOARD handling */

#define KBD_NONSTOP_ON_MOUSE_PRESS 0x00000001L

extern void                   __init_keyboard( void );
extern void                   __term_keyboard( void );
extern int                    __kbdGetKey( unsigned long flg );
extern int                    __kbdTestKey( unsigned long flg );
extern int                    __kbdGetShiftsKey( void );

/* MOUSE handling */

#define MS_LEFTPRESS    1
#define MS_RIGHTPRESS   2
#define MS_MIDDLEPRESS  4

extern int                    __init_mouse( void );
extern void                   __term_mouse( void );
extern int                    __MsGetState( void ); /* return >0 if visible */
extern void                   __MsSetState( int is_visible );
extern void                   __MsGetPos( unsigned *x, unsigned *y );
extern int                    __MsGetBtns( void );

/* VIDEO subsystem handling */

#define __TVIO_MAXSCREENHEIGHT  100
#define __TVIO_MAXSCREENWIDTH   132

#define __TVIO_FLG_DIRECT_CONSOLE_ACCESS  0x00000001L

extern void                   __init_vio( unsigned long flags );
extern void                   __term_vio( void );
#define CUR_OFF   0
#define CUR_NORM  1
#define CUR_SOLID 2
extern int                    __vioGetCursorType( void );
extern void                   __vioSetCursorType( int );
extern void                   __vioGetCursorPos(int* x, int* y);
extern void                   __vioSetCursorPos(int x, int y);
extern void                   __vioRereadState( void );
extern void                   __vioReadBuff(int x,int y,void *buff,unsigned len);
extern void                   __vioWriteBuff(int x,int y,void *buff,unsigned len);
extern unsigned short         tvioWidth,tvioHeight,tvioNumColors;

/* OS specific functions */
extern void                   __init_sys( void );
extern void                   __term_sys( void );
extern void                   __OsYield( void );
extern char *                 __get_ini_name( void );
extern char *                 __get_help_name( void );

/* Timer functions */
/* Condition: create one timer callback per program */

typedef void timer_callback( void );
                             /* Setup user defined timer callback with given
                                time interval. Return real call back
                                interval in milliseconds */
extern unsigned              __OsSetTimerCallBack(unsigned ms,timer_callback *func);
                             /* Restore time callback function to original
                                state */
extern void                   __OsRestoreTimer(void);


/* OS File manipulating functions */
/*
   I had problem with standard i/o function like: open, read, write
   in different C libraries such as emx and possible some one.
   It cause to born this section.
*/
extern void                   __OsClose(int handle);
extern int                    __OsChSize(int handle, long size);
extern int                    __OsCreate(const char *name);
extern int                    __OsDupHandle(int handle);
extern int                    __OsDelete(const char *name);

/* Below describing open flags that present in any OSes */
/* File open flags */
#define FO_READONLY           0x0000
#define FO_WRITEONLY          0x0001
#define FO_READWRITE          0x0002

/* sharing open flags */
#define SO_COMPAT             0x0000
#define SO_DENYALL            0x0010
#define SO_DENYWRITE          0x0020
#define SO_DENYREAD           0x0030
#define SO_DENYNONE           0x0040
#define SO_PRIVATE            0x0080

extern int                    __OsOpen(const char *name,int mode);

extern int                    __IsFileExists(const char *name);
extern long                   __FileLength(int handle);

#define SEEKF_START           0
#define SEEKF_CUR             1
#define SEEKF_END             2

extern long                   __OsSeek(int handle, long newpos, int origin);
extern long                   __OsTell(int handle);

                                        /* Different C libraries have different
                                           methods for changing size of file
                                           and in many case it undocumented */
extern int                    __OsTruncFile(int handle, unsigned long newsize);
extern int                    __OsRename(const char *oldname,const char *newname);
extern int                    __OsRead(int handle,void *buff,unsigned size);
extern int                    __OsWrite(int handle,const void *buff,unsigned size);

/* National Language Support */

extern void                   __nls_Cvt866_to_Base(char *str,unsigned size);

#ifdef __cplusplus
}
#endif

#endif 