/****************************************************************************\
*   KEYBOARD - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1999-2000 Konstantin Boldyshev                     *
*   License:              See below                                          *
*   Author and developer: Konstantin Boldyshev                               *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/generic/unix/keyboard.c                 *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
* ncurses / vt100 keyboard library                                           *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#include "kbd_code.h"
#include "__os_dep.h"

#include <stdio.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_NCURSES

#warning "NCURSES version of BIEW is in alpha stage!"

#include <curses.h>

#ifdef HAVE_MOUSE
MEVENT me;
static int __mou_nbtns;
#else
#undef NCURSES_MOUSE_VERSION
#endif

#else

#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>

extern int ttyhandle;		/* from vio.c */
extern FILE *ttyfile;		/* from vio.c */

static struct termios sattr;

#define MAX_SEQ 20	/* max fields number in a category */

typedef struct {
    char c;
    int key;
} _2seq;

typedef struct {
    char pre1;
    char pre2;
    char suf;
    short num;
    _2seq s[MAX_SEQ];
} eseq;

/*
    array of translatable sequences
*/

#define NUM_SEQ 7	/* number of sequence categories */

const static eseq S[NUM_SEQ] = {
{'O', 0, 0, MAX_SEQ, {
    {'A',KE_UPARROW},
    {'B',KE_DOWNARROW},
    {'C',KE_RIGHTARROW},
    {'D',KE_LEFTARROW},
    {'P',KE_F(1)},
    {'Q',KE_F(2)},
    {'R',KE_F(3)},
    {'w',KE_F(3)},
    {'y',KE_F(3)}, 
    {'S',KE_F(4)},
    {'x',KE_F(4)},
    {'t',KE_F(5)},
    {'v',KE_F(5)},
    {'u',KE_F(6)},
    {'l',KE_F(6)},
    {'q',KE_F(7)},
    {'s',KE_F(7)},
    {'r',KE_F(8)},
    {'p',KE_F(9)},
    {'n',KE_F(9)}}
},
{'[', 0, 0, 8, {
    {'A',KE_UPARROW},
    {'B',KE_DOWNARROW},
    {'C',KE_RIGHTARROW},
    {'D',KE_LEFTARROW},
    {'H',KE_HOME},
    {'K',KE_END},
    {'L',KE_CTL_PGUP},
    {'M',KE_CTL_PGUP}}
},
{'[', '[', 0, 5, {
    {'A',KE_F(1)},
    {'B',KE_F(2)},
    {'C',KE_F(3)},
    {'D',KE_F(4)},
    {'E',KE_F(5)}}
},
{'[', 0, '~', 8, {
    {'1',KE_HOME},
    {'2',KE_INS},
    {'3',KE_DEL},
    {'4',KE_END},
    {'5',KE_PGUP},
    {'6',KE_PGDN},
    {'7',KE_HOME},
    {'8',KE_END}}
},
/*
{'[', 0, '~', 1, {
    {'J',KEY_CTRL_PGDN}}
},
*/
{'[', '1', '~', 8, {
    {'1',KE_F(1)},
    {'2',KE_F(2)},
    {'3',KE_F(3)},
    {'4',KE_F(4)},
    {'5',KE_F(5)},
    {'7',KE_F(6)},
    {'8',KE_F(7)},
    {'9',KE_F(8)}}
},	
{'[', '2', '~', 8, {
    {'0',KE_F(9)},
    {'1',KE_F(10)},
    {'3',KE_SHIFT_F(1)},
    {'4',KE_SHIFT_F(2)},
    {'5',KE_SHIFT_F(3)},
    {'6',KE_SHIFT_F(4)},
    {'8',KE_SHIFT_F(5)},
    {'9',KE_SHIFT_F(6)}}
},	
{'[', '3', '~', 4, {
    {'1',KE_SHIFT_F(7)},
    {'2',KE_SHIFT_F(8)},
    {'3',KE_SHIFT_F(9)},
    {'4',KE_SHIFT_F(10)}}
}
};

#endif

/*
    keyboard FIFO
*/

#define KBUFSIZE 64

static struct {
    int pool[KBUFSIZE];
    int current;
} keybuf;

static int    shift_status = 0;	/* status of shift keys */
int           __unix_mou_x = 0,__unix_mou_y = 0,__unix_mou_btns = 0;

static int is_key_ready()
{
#ifdef HAVE_NCURSES

    int i = getch();
    if (i != ERR) ungetch(i);
    return i == ERR ? 0 : 1;

#else

#ifdef HAVE_SELECT

    struct timeval t = { 0, 0 };
    fd_set readfds;

    FD_ZERO(&readfds);
    FD_SET(ttyhandle, &readfds);
    return select(ttyhandle + 1, &readfds, NULL, NULL, &t);

#else

    int i = getc(ttyfile);
    if (i != -1) ungetc(i, ttyfile);
    return i == -1 ? 0 : 1;

#endif

#endif
}

static void pushEvent(int event)
{
    int i;

    if (event && keybuf.current < KBUFSIZE) {
#ifdef HAVE_NCURSES
	if (event >= KEY_MIN && event <= KEY_MAX) event = _2B(event);
#endif
	if (keybuf.current) 
	    for (i = keybuf.current; i >= 0; i--)
	     keybuf.pool[i + 1] = keybuf.pool[i];
	keybuf.pool[0] = event;
	keybuf.current++;
    }
}

/*
    ReadNextEvent is non-blocking call
*/

static void ReadNextEvent(void)
{
    int key = 0;

#define ret(x)	pushEvent(x); return;
#define set_s(x) shift_status &= (x); shift_status ^= (x); ret(KE_SHIFTKEYS); 

#ifdef HAVE_NCURSES

#define get(x)  x = getch();

#else

#ifdef HAVE_SELECT
#define get(x) read(ttyhandle,&(x),1)
#else
#define get(x) x = getc(ttyfile)
#endif

#define _SEQLEN 7

  int i, j = 0, n;
  char c[_SEQLEN];
#endif

    if (!is_key_ready()) return;

#ifdef HAVE_NCURSES

    get(key); switch(key) {
        case ERR: return;
	case KE_STATUS_ALT	: set_s(KS_ALT);
	case KE_STATUS_SHIFT	: set_s(KS_SHIFT);
	case KE_STATUS_CONTROL	: set_s(KS_CTRL);
	case KE_STATUS_RESET	: set_s(0);
	case KE_ENTER2		: ret(KE_ENTER);
	case KE_C_O		: ret(KE_CTL_(O));
#ifdef HAVE_MOUSE
        case KEY_MOUSE: getmouse(&me);
	                __unix_mou_x = me.x;
			__unix_mou_y = me.y;
			__unix_mou_btns = me.bstate;
			pushEvent(KE_MOUSE);
			return;
#endif	
    }

    ret(key);

#else

/*
    VT100
*/

/*
    translate escape sequence
*/

	get(c[0]); switch(c[0]) {
#ifndef HAVE_SELECT
	    case -1			: return; /* no key is ready */
#endif
	    case KE_ESCAPE		: break;
	    case KE_STATUS_ALT		: set_s(KS_ALT);
	    case KE_STATUS_SHIFT	: set_s(KS_SHIFT);
	    case KE_STATUS_CONTROL	: set_s(KS_CTRL);
	    case KE_STATUS_RESET	: set_s(0);
	    case KE_ENTER2		: key = KE_ENTER; break;
	    case KE_DEL_BACKSPACE	: key = KE_BKSPACE; break;
	    case KE_C_O			: key = KE_CTL_(O); break;
	    default			: key = c[0];
	}

	if (key || j) goto place_key;
	for (i = 1; i < _SEQLEN - 1; i++) {
	    if(is_key_ready()) get(c[i]); 
	    else break; 
	}

	if (i < 3) { 
	    key = c[0]; 
	    goto place_key; 
	}

/*
    translate escape sequence
*/    

#define S S[i]
	for (i = 0; i < NUM_SEQ && !key; i++)  {
 	    if (c[1] != S.pre1) continue;
	    n = 2;
	    if (S.pre2) { 
		if (c[n] != S.pre2) continue; 
		n++; 
	    }
	    for(j = 0; j < S.num; j++) 
		if (c[n] == S.s[j].c) {
		    if (S.suf && c[n+1] != S.suf) break;
		    key = S.s[j].key;
		}
	}

place_key:
   if (key) ret(key);

#endif	/* HAVE_NCURSES */

}

int __kbdGetShiftsKey( void )
{
    return shift_status;
}

int __kbdTestKey(unsigned long flg)
{
#ifdef HAVE_MOUSE
    if(__MsGetBtns() && flg == KBD_NONSTOP_ON_MOUSE_PRESS) return KE_MOUSE;
#endif
    ReadNextEvent();
    return keybuf.current;
}

int __kbdGetKey(unsigned long flg)
{
    int key = 0, s = 0;

#ifdef HAVE_MOUSE
    if (__MsGetBtns() && flg == KBD_NONSTOP_ON_MOUSE_PRESS) return KE_MOUSE;
#endif

    while (!keybuf.current) { __OsYield(); ReadNextEvent(); }
    key = keybuf.pool[--keybuf.current];

    if (!(key == KE_MOUSE || key == KE_SHIFTKEYS)) {
	if ((shift_status & KS_CTRL) == KS_CTRL)	s |= ADD_CONTROL;
	if ((shift_status & KS_ALT) == KS_ALT)		s |= ADD_ALT;
	if ((shift_status & KS_SHIFT) == KS_SHIFT)	s |= ADD_SHIFT;
	shift_status = 0;
    }
    return key | s;
}

void inline __init_keyboard(void)
{
#ifdef HAVE_NCURSES

    raw();
    noecho();
//    timeout(10);
    intrflush(stdscr,FALSE);
    meta(stdscr,TRUE);
    keypad(stdscr,TRUE);
    nodelay(stdscr,TRUE);

#else

    struct termios tattr;

    tcgetattr(ttyhandle, &tattr);
    sattr = tattr;
    tattr.c_lflag &= ~(ICANON | ECHO | ISIG);
    tattr.c_cc[VMIN] = 1;
    tattr.c_cc[VTIME] = 0;
    tattr.c_iflag &= ~(INPCK | ISTRIP | IXON);
    tattr.c_oflag |= OPOST | ONLCR;
    tcsetattr(ttyhandle, TCSANOW, &tattr);

#ifndef HAVE_SELECT
    if (fcntl(ttyhandle, F_SETFL, fcntl(ttyhandle, F_GETFL) | O_NONBLOCK) < 0)
    {
	printm("Cannot set nonblocking mode: %s\nBehavior is undefined!\n", strerror(errno));
    }
#endif

#endif	/* HAVE_NCURSES */

#ifdef HAVE_MOUSE
    if ((__mou_nbtns = __init_mouse()) != INT_MAX) __MsSetState(1);
#endif
}

void inline __term_keyboard(void)
{
#ifdef HAVE_MOUSE
   if(__mou_nbtns != INT_MAX)
   {
     __MsSetState(0);
     __term_mouse();
   }   
#endif

#ifndef HAVE_NCURSES
    tcsetattr(ttyhandle,TCSANOW,&sattr);
#endif
}
