/****************************************************************************\
*   VIO - this source file is part of Binary vIEW project (BIEW)             *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1999-2000 Konstantin Boldyshev                     *
*   License:              See below                                          *
*   Author and developer: Konstantin Boldyshev                               *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/generic/unix/vio.c                      *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
* ncurses/vt100 implementation of screen functions                           *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#include "__os_dep.h"
#include "_sys_dep.h"
#include "biewutil.h" /* for printm() */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

//#define VIO_MALLOC	//problems on exit with dynamic memory allocation
#define VIO_FAST

#define VMAX_X 200
#define VMAX_Y 100

#ifdef HAVE_NCURSES

#include <curses.h>

const static int __ncurses_color[8] = {
    COLOR_BLACK,
    COLOR_BLUE,
    COLOR_GREEN,
    COLOR_CYAN,
    COLOR_RED,
    COLOR_MAGENTA,
    COLOR_YELLOW,
    COLOR_WHITE
};

#define _bg(x)		(((x) >> 4) & 7)
#define _fg(x)		((x) & 7)

/*
    convert vga color to ncurses color
*/

#define _2color(x)	__ncurses_color[((x) & 7)]

/*
    convert vga attribute to ncurses color pair
*/

#define _2cpair(x)	_bg(x) * 010 + _fg(x) + 1


#ifdef HAVE_NCURSES_PSEUDO
static unsigned char _2pseudo(unsigned char c)
{
    switch (c) {
	case 0xB0:	return ACS_CKBOARD;
	case 0xB1:	return ACS_CKBOARD; // 50%
	case 0xB2:	return ACS_CKBOARD; // 80%
	case 0xB3:	return ACS_VLINE;    // 
	case 0xB4:	return ACS_RTEE;     // 
	case 0xB5:	return ACS_RTEE;     // 
	case 0xB6:	return ACS_RTEE;     // 
	case 0xB9:	return ACS_RTEE;     // 
	case 0xBA:	return ACS_VLINE;    // 
	case 0xBB:	return ACS_URCORNER; // 
	case 0xBC:	return ACS_LRCORNER; // 
	case 0xBF:	return ACS_URCORNER; // 
	case 0xC0:	return ACS_LLCORNER; // 
	case 0xC1:	return ACS_BTEE;     // 
	case 0xC2:	return ACS_TTEE;     // 
	case 0xC3:	return ACS_LTEE;     // 
	case 0xC4:	return ACS_HLINE;    // 
	case 0xC5:	return ACS_PLUS;
	case 0xC6:	return ACS_LTEE;     // 
	case 0xC7:	return ACS_LTEE;     // 
	case 0xC8:	return ACS_LLCORNER; // 
	case 0xC9:	return ACS_ULCORNER; //  We don't have doubles in curses
	case 0xCA:	return ACS_BTEE;     // 
	case 0xCB:	return ACS_TTEE;     // 
	case 0xCC:	return ACS_LTEE;     // 
	case 0xCD:	return ACS_HLINE;    // 
	case 0xCF:	return ACS_BTEE;     // 
	case 0xD0:	return ACS_BTEE;     // 
	case 0xD1:	return ACS_TTEE;     // 
	case 0xD2:	return ACS_TTEE;     // 
	case 0xD9:	return ACS_LRCORNER; // 
	case 0xDA:	return ACS_ULCORNER; // 
	case 0xDB:	return ACS_BLOCK;
	case 0xDC:	return ' ';
	case 0xDD:	return ' ';
	case 0xDF:	return ' ';
    }
    return c;

}
extern unsigned char alt2koi[];
#endif

#else /* VT100 */

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#define VTMP_LEN 100

#ifdef VIO_MALLOC
static unsigned char *vtmp;
#else
static unsigned char vtmp[VTMP_LEN];
#endif

static int saveX = 0, saveY = 0, firstX = 0, firstY = 0;
const static int __ansi_color[8] = {0,4,2,6,1,5,3,7};

FILE *ttyfile;
int ttyhandle;

#define twrite(x)	write(ttyhandle,(x),strlen(x))

#define _bg(x) ((x) >> 4)
#define _fg(x) ((x) & 0x0f)

/*
    convert vga color to ansi color
*/

#define _2color(x) __ansi_color[((x) & 7)]

/*
    convert vga attrubute to ansi escape sequence
*/

static char *_2ansi(unsigned char attr)
{
    sprintf(
    	vtmp,
	"\033[%d;3%d;4%d%sm",
	_fg(attr) > 7  ? 1 : 0,
	_2color(_fg(attr)),
	_2color(_bg(attr)),
	_bg(attr) > 7 ? ";5" : ""
    );
    return vtmp;
}

#ifndef HAVE_PSEUDO
static unsigned char _2asc[] = {
"cueaaaaceeeiiiaaeafooouuyOUcLYPfaiounN--?--//!<>:%%|{+++++I+'++."
"`++}-+++`.+++=+++++++++++'.|-||-abipeouyooooooEn=+><()-=o..Vn2X "
};
#endif

#endif	/* HAVE_NCURSES */



#ifdef VIO_MALLOC
static unsigned char *viomem;
#else
static unsigned char viomem[VMAX_X * VMAX_Y * 2];	/* video memory */
#endif

unsigned short tvioWidth, tvioHeight, tvioNumColors=16;
static int cursor_type = CUR_NORM;

static unsigned char *calc_vioaddr(int x,int y)
{
    return viomem+(x+y*tvioWidth)*2;
}

/*

*/

int __vioGetCursorType( void )
{
    return cursor_type;
}


void __vioSetCursorType(int type)
{
    cursor_type = type;

#ifdef HAVE_NCURSES

    leaveok(stdscr, type == CUR_OFF ? TRUE : FALSE);

#else

    sprintf(vtmp, "\033[?25%c", type == CUR_OFF ? 'l' : 'h');
    twrite(vtmp);

#endif
}


void __vioGetCursorPos(int *x,int *y)
{
#ifdef HAVE_NCURSES

    getyx(stdscr,*y,*x);

#else

    *x = saveX;
    *y = saveY;

#endif
}

void __vioSetCursorPos(int x,int y)
{
#ifdef HAVE_NCURSES

    move(y,x);

#else

    saveX = x;
    saveY = y;
    sprintf(vtmp,"\033[%d;%dH", y + 1, x + 1);
    twrite(vtmp);

#endif
}


void __vioReadBuff(int x,int y,void *buff,unsigned len)
{
    memcpy(buff,calc_vioaddr(x,y),len);
}

void __vioWriteBuff(int x, int y, void *buff, unsigned len)
{
    unsigned int i;
    unsigned char c, ca, *p;

#ifdef HAVE_NCURSES

    chtype ch;

    leaveok(stdscr, TRUE);

#else

    unsigned char *pb = malloc(len * 20);
    *pb = 0;

#endif

    memcpy(calc_vioaddr(x, y), buff, len);

    __vioSetCursorPos(x, y);

    for (i = 0; i < len; i += 2) {

	p = buff + i; c = *p; ca = *(p + 1);

	if (!c) c = ' '; if (c < ' ' || c == 0x7f || c == 0x9b) c = '.';

#ifdef HAVE_NCURSES

	ch = c | COLOR_PAIR(_2cpair(ca));
	if ((ca & 0xf) > 7)	ch |= A_BOLD;
	if ((ca & 0xf0) > 0x70)	ch |= A_BLINK;
	addch(ch);

#else

#ifndef HAVE_PSEUDO
	if (c >= 0x80) c = _2asc[c - 0x80];
#endif

#ifdef VIO_FAST
	if (((i >= 2) && (ca != *(p - 1))) || i == (len - 2) || !i)
#endif
	strcat(pb, _2ansi(ca));

	strncat(pb, &c, 1);

#endif	/* HAVE_NCURSES */

    }

#ifdef HAVE_NCURSES

    refresh();

#else

    twrite(pb);
    free(pb);

#endif	/* HAVE_NCURSES */
}

/*
void __vioRereadState( void )
{
#ifdef HAVE_NCURSES

    tvioWidth = COLS;
    tvioHeight = LINES;

#else

    struct winsize w;

    ioctl(ttyhandle, TIOCGWINSZ, &w);
    tvioWidth = w.ws_col ? w.ws_col : 80;
    tvioHeight = w.ws_row ? w.ws_row : 25;

#ifndef VIO_MALLOC
    if (tvioWidth > VMAX_X) tvioWidth = VMAX_X;
    if (tvioHeight > VMAX_Y) tvioHeight = VMAX_Y;
#endif

#endif
}
*/

void __term_vio(void)
{
#ifdef HAVE_NCURSES

    endwin();

#else

    __vioSetCursorType(CUR_NORM);
    __vioSetCursorPos(firstX, firstY);

    twrite("\033[0m\033[3l");
    fclose(ttyfile);

#endif
}

void __init_vio(unsigned long flags)
{
#ifdef HAVE_NCURSES

    int i;

    initscr();
    if (has_colors()) start_color();
    if (COLOR_PAIRS > 077) for (i = 0; i < 0xff; i++)
	    init_pair(
		_2cpair(i),
		_2color(_fg(i)),
		_2color(_bg(i))
    	    );
    else tvioNumColors = 2;

    tvioWidth = COLS;
    tvioHeight = LINES;

#ifdef HAVE_NCURSES_PSEUDO
    for (i = 0x80; i < 0x100; i++) if (i != _2pseudo(i)) alt2koi[i-0x80] = _2pseudo(i);
#endif
#else

    struct winsize w;
/*
    char *buf = getenv("TERM");

    if (buf && (strcasecmp(buf, "ANSI") || strcasecmp(buf, "vt100")))
    {
	printm("I can't handle %s terminal.", buf);
	exit(2);
    }
*/
    ttyfile = fopen(ttyname(fileno(stdout)), "r+");
    if (ttyfile == NULL) ttyfile = stdout;

    ttyhandle = fileno(ttyfile);

    twrite("\033[0m\033[3h");

    ioctl(ttyhandle, TIOCGWINSZ, &w);
    tvioWidth = w.ws_col ? w.ws_col : 80;
    tvioHeight = w.ws_row ? w.ws_row : 25;
#ifndef VIO_MALLOC
    if (tvioWidth > VMAX_X) tvioWidth = VMAX_X;
    if (tvioHeight > VMAX_Y) tvioHeight = VMAX_Y;
#endif

#ifdef VIO_MALLOC
    vtmp = malloc(VTMP_LEN);
    viomem = malloc(tvioWidth * tvioHeight * 2);
#endif

    memset(vtmp, 0, VTMP_LEN);
    memset(viomem, 0, tvioWidth * tvioHeight * 2);

#endif
}
