/****************************************************************************\
*   VIO - this source file is part of Binary vIEW project (BIEW)             *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/i386/dos/vio.c                          *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of video subsystem handles for DOS-32.   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include "__os_dep.h"
#include "_sys_dep.h"
#include <dos.h>
#include <string.h>
#include <go32.h>
#include <sys/farptr.h>

unsigned short tvioWidth,tvioHeight,tvioNumColors;
static unsigned viomem;
static unsigned long vio_flags;

void __init_vio( unsigned long flg )
{
#if 0
  union REGS reg;
  _AX = 0x1111;
  _BL = 0;
  geninterrupt(0x10); /* 80x28 */
#endif
  vio_flags = flg;
  __vioRereadState();
}

void __term_vio( void )
{
#if 0
  union REGS inreg;
  memset(&inreg,0,sizeof(inreg));
  inreg.x.ax = 0x1114;
  inreg.h.bl = 0;
  int86(0x10,&inreg,&inreg);
#endif
}

void __vioRereadState( void )
{
  union REGS reg;
  reg.h.ah = 0x0F;
  int86(0x10,&reg,&reg);
  tvioNumColors = 16;
  tvioWidth = _farpeekw(_go32_conventional_mem_selector(),0x044A);
  tvioHeight = _farpeekb(_go32_conventional_mem_selector(),0x0484)+1;
  switch(reg.h.al)
  {
     case 7: viomem = 0xB0000;
             tvioNumColors = 2;
             break;
     default: viomem = 0xB8000;
  }
}

int __vioGetCursorType( void )
{
  union REGS reg;
  int ret,mode;
  memset(&reg,0,sizeof(union REGS));
  reg.h.ah = 0x0F;
  int86(0x10,&reg,&reg);
  mode = reg.h.al;
  reg.h.ah = 3;
  int86(0x10,&reg,&reg);
  if(reg.x.cx == 0x2000) ret = CUR_OFF;
  else
    if(tvioHeight >= 43 && reg.x.cx == 0x08) ret = CUR_SOLID;
    else
      if(tvioHeight < 43 && mode == 7 && reg.x.cx == 0x0C) ret = CUR_SOLID;
      else
        if(tvioHeight < 43 && mode != 7 && reg.x.cx == 0x07) ret = CUR_SOLID;
        else ret = CUR_NORM;
  return ret;
}

/* type: 0 - off  1 - normal  2 - solid */
void __vioSetCursorType(int type)
{
  union REGS inreg,outreg;
  int mode;
  memset(&inreg,0,sizeof(inreg));
  inreg.h.ah = 0x0F;
  int86(0x10,&inreg,&outreg);
  mode = outreg.h.al;
  switch(type)
  {
    case CUR_OFF:
            inreg.x.cx = 0x2000;
            break;
    case CUR_SOLID:
            if(tvioHeight >= 43) inreg.x.cx = 0x08;
            else
            {
              inreg.h.ch = 0;
              inreg.h.cl = mode == 7 ? 0x0C : 0x07;
            }
            break;
     default:
            if(mode == 7) inreg.x.cx = 0x0B0C;
            else
            {
              inreg.h.ch = 0x06;
              inreg.h.cl = tvioHeight >= 43 ? 0 : 7;
            }
  }
  inreg.h.ah = 1;
  int86(0x10,&inreg,&outreg);
}

void __vioGetCursorPos(int *x,int *y)
{
  union REGS reg;
  memset(&reg,0,sizeof(union REGS));
  reg.h.ah = 3;
  int86(0x10,&reg,&reg);
  *x = reg.h.dl;
  *y = reg.h.dh;
}

void __vioSetCursorPos(int x,int y)
{
   union REGS reg;
   memset(&reg,0,sizeof(union REGS));
   reg.h.ah = 2;
   reg.h.dh = y;
   reg.h.dl = x;
   int86(0x10,&reg,&reg);
}

#define calc_vioaddr(x,y) (viomem+(x+y*tvioWidth)*2)

void __vioWriteBuff(int x,int y,void *buff,unsigned len)
{
  if((vio_flags & __TVIO_FLG_DIRECT_CONSOLE_ACCESS) != __TVIO_FLG_DIRECT_CONSOLE_ACCESS)
  {
    union REGS reg,oreg;
    unsigned i,apage;
    memset(&reg,0,sizeof(union REGS));
    reg.h.ah = 0x0F;
    int86(0x10,&reg,&reg);
    apage = reg.h.bh;
    reg.h.ah = 0x09;
    reg.h.bh = apage;
    reg.x.cx = 1;
    for(i = 0;i < len;)
    {
      __vioSetCursorPos(x++,y);
      reg.h.al = ((char *)buff)[i++];
      reg.h.bl = ((char *)buff)[i++];
      int86(0x10,&reg,&oreg);
      if(x > tvioWidth) { x = 0; y++; }
    }
  }
  else
    movedata(_go32_my_ds(),(unsigned)buff,
             _go32_conventional_mem_selector(),
             calc_vioaddr(x,y),len);
}

void __vioReadBuff(int x,int y,void *buff,unsigned len)
{
  if((vio_flags & __TVIO_FLG_DIRECT_CONSOLE_ACCESS) != __TVIO_FLG_DIRECT_CONSOLE_ACCESS)
  {
    union REGS reg,oreg;
    unsigned i,apage,blen;
    memset(&reg,0,sizeof(union REGS));
    blen = len / 2;
    reg.h.ah = 0x0F;
    int86(0x10,&reg,&reg);
    apage = reg.h.bh;
    reg.h.ah = 0x08;
    reg.h.bh = apage;
    for(i = 0;i < blen;)
    {
      __vioSetCursorPos(x++,y);
      int86(0x10,&reg,&oreg);
      ((short *)buff)[i++] = oreg.x.ax;
      if(x > tvioWidth) { x = 0; y++; }
    }
  }
  else
    movedata(_go32_conventional_mem_selector(),
             calc_vioaddr(x,y),_go32_my_ds(),
             (unsigned)buff,len);
}
 