/****************************************************************************\
*   KEYBOARD - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1999-2000 Konstantin Boldyshev                     *
*   License:              See below                                          *
*   Author and developer: Konstantin Boldyshev                               *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/I386/linux/keyboard.c                   *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  Linux direct console / vt100 keyboard library                             *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#include "kbd_code.h"
#include "__os_dep.h"
#include "biewutil.h"

#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/kd.h>

#ifdef HAVE_MOUSE
#include <gpm.h>
#define get(x) (x) = Gpm_Getchar()

static int _mou_btns = 0;
static unsigned _mou_x = 0,_mou_y = 0;
static int __ms_nbtns;

#else

#define get(x) read(ttyhandle,&(x),1) /* read one byte */

#endif

#define KBUFSIZE	64	/* size of keyboard buffer */
#define KSCANSIZE	84	/* size of scancode table */
#define KEYNUM		128	/* */

extern int ttyhandle, on_console;	/* from vio.c */
static volatile int shift_status = 0,	/* status of shift keys */
		    keypressed[KEYNUM];	/* indicates whether key is down */
static struct termios sattr;		/* original terminal attributes */

/*
    keyboard FIFO
*/

static struct {
    int pool[KBUFSIZE];
    int current;
} keybuf;

/*
    scancode tables
*/

static unsigned int scancode_table[KSCANSIZE] =
{
    KE_ESCAPE,	'1',	'2',	'3',	'4',	'5',	'6',	'7',
    '8',	'9',	'0',	'-',	'=',	KE_BKSPACE,KE_TAB,'q',
    'w',	'e',	'r',	't',	'y',	'u',	'i',	'o',
    'p',	'[',	']',	KE_ENTER,0xffff,'a',	's',	'd',
    'f',	'g',	'h',	'j',	'k',	'l',	';',	0x27,
    '`',	0xffff,	0x5c,	'z',	'x',	'c',	'v',	'b',
    'n',	'm',	',',	'.',	'/',	0xffff,	'*',	0xffff,
    ' ',	0xffff,	KE_F(1),	KE_F(2),	KE_F(3),	KE_F(4),	KE_F(5),	KE_F(6),
    KE_F(7),	KE_F(8),	KE_F(9),	KE_F(10),	0xffff,	0xffff,	KE_HOME,KE_UPARROW,
    KE_PGUP,	'-',	KE_LEFTARROW,0,KE_RIGHTARROW,'+',KE_END,KE_DOWNARROW,
    KE_PGDN,	KE_INS,	KE_DEL,	0xfffe
};

/*
unsigned int scancode_caps_table[KSCANSIZE] =
{
    KE_ESCAPE,	'!',	'@',	'#',	'$',	'%',	'^',	'&',
    '*',	'(',	')',	'_',	'+',	KE_BKSPACE,KE_TAB,'Q',
    'W',	'E',	'R',	'T',	'Y',	'U',	'I',	'O',
    'P',	'{',	'}',	KE_ENTER,0,	'A',	'S',	'D',
    'F',	'G',	'H',	'J',	'K',	'L',	':',	0x22,
    '~',	0,	'|',	'Z',	'X',	'C',	'V',	'B',
    'N',	'M',	'<',	'>',	'?',	0,	'*',	0xff,
    ' ',	0,	KE_F(1),	KE_F(2),	KE_F(3),	KE_F(4),	KE_F(5),	KE_F(6),
    KE_F(7),	KE_F(8),	KE_F(9),	KE_F(10),	0xffff,	0xffff,	KE_HOME,KE_UPARROW,
    KE_PGUP,	'-',	KE_LEFTARROW,0,KE_RIGHTARROW,'+',KE_END,KE_DOWNARROW,
    KE_PGDN,	KE_INS,	KE_DEL,	0xfffe
};

unsigned int scancode_num_table[KSCANSIZE] =
{
    KE_ESCAPE,	'!',	'@',	'#',	'$',	'%',	'^',	'&',
    '*',	'(',	')',	'_',	'+',	KE_BKSPACE,KE_TAB,'Q',
    'W',	'E',	'R',	'T',	'Y',	'U',	'I',	'O',
    'P',	'{',	'}',	KE_ENTER,KS_CTRL,'A',	'S',	'D',
    'F',	'G',	'H',	'J',	'K',	'L',	':',	0x22,
    '~',	KS_SHIFT,'|',	'Z',	'X',	'C',	'V',	'B',
    'N',	'M',	'<',	'>',	'?',	KS_SHIFT,'*',	KS_ALT,
    ' ',	KS_CAPSLOCK,KE_F(1),KE_F(2),KE_F(3),	KE_F(4),	KE_F(5),	KE_F(6),
    KE_F(7),	KE_F(8),	KE_F(9),	KE_F(10),	KS_NUMLOCK,KS_SCRLOCK,'7','8',
    '9',	'-',	'4',	'5',	'6',	'+',	'1',	'2',
    '3',	'0',	'.',	0
};
*/

#define MAX_SEQ 20	/* max fields number in a category */
#define NUM_SEQ 7	/* number of sequence categories */

typedef struct {
    char c;
    int key;
} _2seq;

typedef struct {
    char pre1;
    char pre2;
    char suf;
    short num;
    _2seq s[MAX_SEQ];
} eseq;

/*
    array of translatable sequences
*/

const static eseq S[NUM_SEQ] = {
{'O', 0, 0, MAX_SEQ, {
    {'A',KE_UPARROW},
    {'B',KE_DOWNARROW},
    {'C',KE_RIGHTARROW},
    {'D',KE_LEFTARROW},
    {'P',KE_F(1)},
    {'Q',KE_F(2)},
    {'R',KE_F(3)},
    {'w',KE_F(3)},
    {'y',KE_F(3)}, 
    {'S',KE_F(4)},
    {'x',KE_F(4)},
    {'t',KE_F(5)},
    {'v',KE_F(5)},
    {'u',KE_F(6)},
    {'l',KE_F(6)},
    {'q',KE_F(7)},
    {'s',KE_F(7)},
    {'r',KE_F(8)},
    {'p',KE_F(9)},
    {'n',KE_F(9)}}
},
{'[', 0, 0, 8, {
    {'A',KE_UPARROW},
    {'B',KE_DOWNARROW},
    {'C',KE_RIGHTARROW},
    {'D',KE_LEFTARROW},
    {'H',KE_HOME},
    {'K',KE_END},
    {'L',KE_CTL_PGUP},
    {'M',KE_CTL_PGUP}}
},
{'[', '[', 0, 5, {
    {'A',KE_F(1)},
    {'B',KE_F(2)},
    {'C',KE_F(3)},
    {'D',KE_F(4)},
    {'E',KE_F(5)}}
},
{'[', 0, '~', 8, {
    {'1',KE_HOME},
    {'2',KE_INS},
    {'3',KE_DEL},
    {'4',KE_END},
    {'5',KE_PGUP},
    {'6',KE_PGDN},
    {'7',KE_HOME},
    {'8',KE_END}}
},
/*
{'[', 0, '~', 1, {
    {'J',KEY_CTRL_PGDN}}
},
*/
{'[', '1', '~', 8, {
    {'1',KE_F(1)},
    {'2',KE_F(2)},
    {'3',KE_F(3)},
    {'4',KE_F(4)},
    {'5',KE_F(5)},
    {'7',KE_F(6)},
    {'8',KE_F(7)},
    {'9',KE_F(8)}}
},	
{'[', '2', '~', 8, {
    {'0',KE_F(9)},
    {'1',KE_F(10)},
    {'3',KE_SHIFT_F(1)},
    {'4',KE_SHIFT_F(2)},
    {'5',KE_SHIFT_F(3)},
    {'6',KE_SHIFT_F(4)},
    {'8',KE_SHIFT_F(5)},
    {'9',KE_SHIFT_F(6)}}
},	
{'[', '3', '~', 4, {
    {'1',KE_SHIFT_F(7)},
    {'2',KE_SHIFT_F(8)},
    {'3',KE_SHIFT_F(9)},
    {'4',KE_SHIFT_F(10)}}
}
};

static int is_key_ready()
{
    struct timeval t = { 0, 0 };
    fd_set readfds;

    FD_ZERO(&readfds);
    FD_SET(ttyhandle, &readfds);
    return select(ttyhandle + 1, &readfds, NULL, NULL, &t);
}

static void pushEvent(int event)
{
    if (event && keybuf.current < KBUFSIZE) {
	int i;
	if (keybuf.current) for (i = keybuf.current; i >= 0; i--)
	     keybuf.pool[i + 1] = keybuf.pool[i];
	keybuf.pool[0] = event;
	keybuf.current++;
    }
}
/*
    ReadNextEvent is non-blocking
*/

static void ReadNextEvent(void)
{
#define _SEQLEN 7
#define ret(x)	pushEvent((x)); return;
#define set_s(x) shift_status &= (x); shift_status ^= (x); ret(KE_SHIFTKEYS);

	unsigned char c = 0;
	unsigned key = 0;
	int i;

	if (!is_key_ready()) return;

	/* If key is ready, place it into keyboard buffer */

        if (on_console) {
	    int ss = shift_status;	    

	    do { get(c); } while (c == 0xe0);
	    if (c == 0xe1) get(c);
	    if (c) {
		keypressed[c & (KEYNUM - 1)] = c & KEYNUM ? 0 : 1;
		for (i = 1; i < KSCANSIZE; i++) if (keypressed[i]) 
		    key = scancode_table[i - 1];
	    }
	    if (__kbdGetShiftsKey() != ss) key = KE_SHIFTKEYS;
        } else {

/*

    VT100 emulation

*/

	    int j = 0, n;
	    char c[_SEQLEN];

	    get(c[0]); switch(c[0]) {
		case KE_STATUS_ALT	: set_s(KS_ALT);;
		case KE_STATUS_SHIFT	: set_s(KS_SHIFT);
		case KE_STATUS_CONTROL	: set_s(KS_CTRL);
		case KE_STATUS_RESET	: set_s(0);
		case KE_ENTER2		: key = KE_ENTER; break;
		case KE_DEL_BACKSPACE	: key = KE_BKSPACE; break;
		case KE_C_O		: key = KE_CTL_(O); break;
		case KE_ESCAPE		: break;
		default			: key = c[0];
	    }
	    if (key || j) goto place_key;
	    for (i = 1; i < _SEQLEN - 1; i++) { 
		if(is_key_ready()) get(c[i]); 
		else break; 
	    }
	    if (i < 3) { 
		key = c[0]; 
		goto place_key; 
	    }

/*
    translate escape sequence
*/    

#define S S[i]
	    for (i = 0; i < NUM_SEQ && !key; i++) {
		if (c[1] != S.pre1) continue;
		n = 2;
		if (S.pre2) { 
		    if (c[n] != S.pre2) continue; 
		    n++; 
		}
		for (j = 0; j < S.num; j++) if (c[n] == S.s[j].c) {
		    if (S.suf && c[n+1] != S.suf) break;
		    key = S.s[j].key;
	        }
	    }
	}
place_key:
       if (key && key < 0xfffe) ret(key);
}

int __kbdGetShiftsKey(void)
{
	/* shift_status = 6;ioctl(ttyhandle,TIOCLINUX,&shift_status); */
    if (on_console) {
	 shift_status = 0;
	 if (keypressed[0x2a] || keypressed[0x36]) shift_status |= KS_SHIFT;
	 if (keypressed[0x1d]) shift_status |= KS_CTRL;
	 if (keypressed[0x38]) shift_status |= KS_ALT;
	 if (keypressed[0x3a]) shift_status |= KS_CAPSLOCK;
	 if (keypressed[0x45]) shift_status |= KS_NUMLOCK;
	 if (keypressed[0x46]) shift_status |= KS_SCRLOCK;
    }
    return shift_status; 
}

int __kbdTestKey(unsigned long flg)
{
#ifdef HAVE_MOUSE
    if(__MsGetBtns() && flg == KBD_NONSTOP_ON_MOUSE_PRESS) return KE_MOUSE;
#endif
    ReadNextEvent();
    return keybuf.current;
}

int __kbdGetKey (unsigned long flg)
{
    int key = 0, s = 0;

#ifdef HAVE_MOUSE
    if (__MsGetBtns() && flg == KBD_NONSTOP_ON_MOUSE_PRESS) return KE_MOUSE;
#endif

    while (!keybuf.current) { __OsYield(); ReadNextEvent(); }
    key = keybuf.pool[--keybuf.current];

    if (!(key == KE_MOUSE || key == KE_SHIFTKEYS)) {
	if ((shift_status & KS_CTRL) == KS_CTRL)	s |= ADD_CONTROL;
	if ((shift_status & KS_ALT) == KS_ALT)		s |= ADD_ALT;
	if ((shift_status & KS_SHIFT) == KS_SHIFT)	s |= ADD_SHIFT;
	if (!on_console) shift_status = 0;
    }
    return key | s;
}

/*

*/

inline void __term_keyboard(void)
{
    if (on_console) ioctl(ttyhandle, KDSKBMODE, K_XLATE);
    tcsetattr(ttyhandle,TCSANOW,&sattr);
}

inline void __init_keyboard(void)
{
    struct termios tattr;

    if (on_console && (ioctl(ttyhandle, KDSKBMODE, K_RAW) < 0)) {
	printm("Cannot set keyboard raw mode: %s\nUsing VT100 emulation..\n", strerror(errno));
	on_console = 0;
    }
    tcgetattr(ttyhandle, &tattr);
    sattr = tattr;
    tattr.c_lflag &= ~(ICANON | ECHO | ISIG);
    tattr.c_cc[VMIN] = 1;
    tattr.c_cc[VTIME] = 0;
    tattr.c_iflag &= ~(INPCK | ISTRIP | IXON | ICRNL);
    tattr.c_oflag |= OPOST | ONLCR;
    tcsetattr(ttyhandle, TCSANOW, &tattr);
}
