/****************************************************************************\
*   VIO - this source file is part of Binary vIEW project (BIEW)             *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/i386/win32/vio.c                        *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of video subsystem handles for Win32s.   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include "windows.h"
#include "__os_dep.h"
#include "_sys_dep.h"
#include <stdlib.h>
#include <string.h>

unsigned short tvioWidth,tvioHeight,tvioNumColors;
static CONSOLE_SCREEN_BUFFER_INFO csbinfo;
static HANDLE hOut;
static unsigned long vio_flags;

void __init_vio( unsigned long flg )
{
  hOut = GetStdHandle(STD_OUTPUT_HANDLE);
  vio_flags = flg;
  __vioRereadState();
}

void __term_vio( void )
{
}

void __vioRereadState( void )
{
  GetConsoleScreenBufferInfo(hOut,&csbinfo);
  if((vio_flags & __TVIO_FLG_DIRECT_CONSOLE_ACCESS) == __TVIO_FLG_DIRECT_CONSOLE_ACCESS)
  {
    tvioWidth  = csbinfo.dwSize.X;
    tvioHeight = csbinfo.dwSize.Y;
  }
  else
  {
    tvioWidth  = csbinfo.srWindow.Right-csbinfo.srWindow.Left+1;
    tvioHeight = csbinfo.srWindow.Bottom-csbinfo.srWindow.Top+1;
  }
  tvioNumColors = 16;
}

int __vioGetCursorType( void )
{
  CONSOLE_CURSOR_INFO cci;
  GetConsoleCursorInfo(hOut,&cci);
  if(cci.bVisible == FALSE) return CUR_OFF;
  if(cci.dwSize == 100) return CUR_SOLID;
  return CUR_NORM;
}

/* type: 0 - off  1 - normal  2 - solid */
void __vioSetCursorType(int type)
{
  CONSOLE_CURSOR_INFO cci;
  GetConsoleCursorInfo(hOut,&cci);
  cci.bVisible = TRUE;
  switch(type)
  {
    case CUR_OFF: cci.bVisible = FALSE;
                  break;
    case CUR_SOLID:
                  cci.dwSize = 100;
                  break;
    default:
                  cci.dwSize = 20;
  }
  SetConsoleCursorInfo(hOut,&cci);
}

void __vioGetCursorPos(int *x,int *y)
{
  CONSOLE_SCREEN_BUFFER_INFO csbinfo;
  GetConsoleScreenBufferInfo(hOut,&csbinfo);
  *x = csbinfo.dwCursorPosition.X;
  *y = csbinfo.dwCursorPosition.Y;
}

void __vioSetCursorPos(int x,int y)
{
  COORD cc;
  cc.X = x;
  cc.Y = y;
  SetConsoleCursorPosition(hOut,cc);
}

void __vioWriteBuff(int x,int y,void *buff,unsigned len)
{
  unsigned char chars[__TVIO_MAXSCREENWIDTH];
  unsigned short attr[__TVIO_MAXSCREENWIDTH];
  unsigned long w;
  unsigned rlen;
  unsigned j,i;
  COORD cc;
  rlen = min(len/2,__TVIO_MAXSCREENWIDTH);
  cc.X = x;
  cc.Y = y;
  for(j = i = 0;i < rlen;i++,j+=2)
  {
    chars[i] = ((unsigned char *)buff)[j];
    attr[i] = ((unsigned char *)buff)[j+1];
  }
  WriteConsoleOutputCharacter(hOut,chars,rlen,cc,(LPDWORD)&w);
  WriteConsoleOutputAttribute(hOut,attr,rlen,cc,(LPDWORD)&w);
}

void __vioReadBuff(int x,int y,void *buff,unsigned len)
{
  unsigned char chars[__TVIO_MAXSCREENWIDTH];
  unsigned short attr[__TVIO_MAXSCREENWIDTH];
  unsigned long r;
  unsigned i,rlen,j;
  COORD cc;
  rlen = min(len/2,__TVIO_MAXSCREENWIDTH);
  cc.X = x;
  cc.Y = y;
  ReadConsoleOutputCharacter(hOut,chars,rlen,cc,(LPDWORD)&r);
  ReadConsoleOutputAttribute(hOut,attr,rlen,cc,(LPDWORD)&r);
  for(j = i = 0;i < rlen;i++,j+=2)
  {
    ((unsigned char *)buff)[j] = chars[i];
    ((unsigned char *)buff)[j+1] = attr[i];
  }
}
 