/****************************************************************************\
*   TWIN - this source file is part of Binary vIEW project (BIEW)            *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/twin.c                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of Text Window manager.                  *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
/* Text window release */

#include "twin.h"
#include "biewlib.h"
#include "pmalloc.h"

#ifndef NDEBUG
#ifdef __TSC__
#pragma save , call(inline=>on) // must be GP in protmode
static void winInternalError( void ) = { 0xFF, 0xFF };
#pragma restore
#else
#ifdef __GNUC__
#define winInternalError(x) asm(".short 0xFFFF\n");
#endif
#endif

static int __NEAR__ test_win(TWindow *win)
{
 if(*((void **)(win->body + win->wsize)) == win->body)
 {
   if(*((void **)(win->saved + win->wsize)) == win->saved) return 1;
 }
 return 0;
}
#endif

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

#define IFLG_VISIBLE       0x00000001L
#define IFLG_ENABLED       0x00000002L
#define IFLG_CURSORBEENOFF 0x80000000L

TWindow *active;

static void __NEAR__ winerr(char *str) { printm("\n\nInternal twin library error: %s\n",str); _exit(EXIT_FAILURE); }
static void __NEAR__ wputc(char,char,int update);
static void __NEAR__ paint_internal(void);

static TWindow *head = NULL;
static TWindow *cursorwin = NULL;

#ifndef NDEBUG
#define CHECK_WINS(x) { if(!test_win(x)) winInternalError(); }
#else
#define CHECK_WINS(x)
#endif

void twInit( unsigned long vio_flags )
{
  __init_vio(vio_flags);
  __init_keyboard();
  if(tvioWidth > __TVIO_MAXSCREENWIDTH)
  {
    winerr("Size of video buffer is too large");
  }
  __vioSetCursorType(CUR_OFF);
}

void twDestroy( void )
{
  __vioSetCursorType(CUR_NORM);
  __term_keyboard();
  __term_vio();
}

static void __NEAR__ adjustColor(Color *fore,Color *back)
{
  if(tvioNumColors < 16)
  {
    if(*fore == 8 || *fore < 7) *fore = Black;
    else
      if(*fore > 8) *fore = White;
      else          *fore = LightGray;
    if(*back == 8 || *back < 7) *back = Black;
    else                        *back = LightGray;
    if(*back == *fore)
    {
      if(*back == Black) *fore = LightGray;
      else
        if(*back == LightGray) *fore = *fore == LightGray ? Black : White;
    }
  }
}

Color twSetColor(Color fore,Color back)
{
  Color ret;
  ret = active->ctext;
  adjustColor(&fore,&back);
  active->ctext = fore & 0x0F;
  active->ctext |= (back << 4) & 0xF0;
  return ret;
}

Color twGetColor( void ) { return active->ctext; }

Color twTextColor( Color col )
{
  Color back,ret;
  back = (twGetColor() >> 4) & 0x0F;
  ret = (twGetColor() & 0x0F);
  twSetColor(col, back);
  return ret;
}

Color twTextBkGnd( Color col )
{
   Color fore,ret;
   fore = (twGetColor() & 0x0F);
   ret = (twGetColor() >> 4) & 0x0F;
   twSetColor(fore, col);
   return ret;
}

void twSetFrame(TWindow *win,char *frame,Color fore,Color back)
{
  TWindow *tmp;
  win->flags |= TWS_FRAMEABLE;
  if(frame != UP3D_FRAME && frame != DN3D_FRAME) memcpy(win->Frame,frame,8);
  else
  {
    if(frame == UP3D_FRAME) memset(win->Frame,0,8);
    else                    memset(win->Frame,-1,8);
  }
  adjustColor(&fore,&back);
  win->cframe = fore & 0x0F;
  win->cframe |= (back << 4) & 0xF0;
  tmp = active;
  active = win;
  paint_internal();
  active = tmp;
}

void twGetFrame(TWindow *win,char *frame,Color* fore,Color* back)
{
  if((win->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE)
  {
    memcpy(frame,active->Frame,8);
    *fore = win->cframe & 0x0F;
    *back = (win->cframe >> 4) & 0x0F;
  }
}

void twSetTitle(TWindow *win,char *title,int mode,Color fore,Color back)
{
  TWindow *tmp;
  int slen;
  slen = strlen(title);
  if(win->Title) PFREE(win->Title);
  win->Title = PMalloc(slen+1);
  if(!win->Title) winerr("Out of memory!");
  strcpy(win->Title,title);
  if((win->flags & TWS_NLS866) == TWS_NLS866) __nls_Cvt866_to_Base(win->Title,slen);
  win->TitleMode = mode;
  adjustColor(&fore,&back);
  win->ctitle = fore & 0x0F;
  win->ctitle |= (back << 4) & 0xF0;
  tmp = active;
  active = win;
  paint_internal();
  active = tmp;
}

int twGetTitle(TWindow *win,char *title,unsigned cb_title,Color* fore,Color* back)
{
  if((win->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE && win->Title)
  {
    strncpy(title,win->Title,cb_title);
    title[cb_title-1] = '\0';
    *fore = win->ctitle & 0x0F;
    *back = (win->ctitle >> 4) & 0x0F;
    return win->TitleMode;
  }
  return 0;
}

void twSetFooter(TWindow *win,char *footer,int mode,Color fore,Color back)
{
  TWindow *tmp;
  int slen;
  slen = strlen(footer);
  if(win->Footer) PFREE(win->Footer);
  win->Footer = PMalloc(slen+1);
  if(!win->Footer) winerr("Out of memory!");
  strcpy(win->Footer,footer);
  if((win->flags & TWS_NLS866) == TWS_NLS866) __nls_Cvt866_to_Base(win->Footer,slen);
  win->FooterMode = mode;
  adjustColor(&fore,&back);
  win->cfooter = fore & 0x0F;
  win->cfooter |= (back << 4) & 0xF0;
  tmp = active;
  active = win;
  paint_internal();
  active = tmp;
}

int twGetFooter(TWindow *win,char *footer,unsigned cb_footer,Color* fore,Color* back)
{
  if((win->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE && win->Footer)
  {
    strncpy(footer,win->Footer,cb_footer);
    footer[cb_footer-1] = '\0';
    *fore = win->cfooter & 0x0F;
    *back = (win->cfooter >> 4) & 0x0F;
    return win->FooterMode;
  }
  return 0;
}

static void __NEAR__ __unlistwin(TWindow *win)
{
  if(head == win) head = win->next;
  else
  {
    TWindow *iter;
    iter = head;
    while(1)
    {
       if(!iter) break;
       if(iter->next) if(iter->next == win) iter->next = win->next;
       iter = iter->next;
    }
  }
}

#define __ATHEAD(win)\
{\
  if(head) ((TWindow *)win)->next = head;\
  head = (TWindow *)win;\
}

#define __ATWIN(win,prev)\
{\
  if(!prev) __ATHEAD(win)\
  else\
  {\
     ((TWindow *)win)->next = ((TWindow *)prev)->next;\
     ((TWindow *)prev)->next = (TWindow *)win;\
  }\
}

static TWindow * __NEAR__ __prevwin(TWindow *win)
{
  if(head == win) return NULL;
  else
  {
    TWindow *iter;
    iter = head;
    while(1)
    {
       if(!iter) break;
       if(iter->next) if(iter->next == win) return iter;
       iter = iter->next;
    }
  }
  return NULL;
}

static TWindow * __NEAR__ __findcursorablewin(void)
{
  TWindow *iter;
  iter = head;
  while(1)
  {
     if(!iter) break;
     if((iter->flags & TWS_CURSORABLE) == TWS_CURSORABLE) return iter;
     iter = iter->next;
  }
  return NULL;
}

#define __AT_POINT(iter,x,y)\
{\
  iter = head;\
  while( 1 )\
  {\
    if(!iter) break;\
    if((iter->iflags & IFLG_VISIBLE) == IFLG_VISIBLE)\
    {\
      if((y >= iter->Y1) && (y < iter->Y2) &&\
         (x >= iter->X1) && (x < iter->X2))\
                       break;\
    }\
    iter = iter->next;\
  }\
}

TWindow * twGetWinAtPos(int x,int y)
{
  TWindow *ret;
  __AT_POINT(ret,x-1,y-1);
  return ret;
}

#define __FIND_OVER(ret,win,x,y)\
{\
  TWindow *iter;\
  int xx,yy;\
  iter = head;\
  ret = NULL;\
  xx = win->X1+x;\
  yy = win->Y1+y;\
  while( iter && iter != win )\
  {\
    if((iter->iflags & IFLG_VISIBLE) == IFLG_VISIBLE)\
    {\
      if((yy >= iter->Y1) && (yy < iter->Y2) &&\
         (xx >= iter->X1) && (xx < iter->X2))\
                       ret = iter;\
    }\
    iter = iter->next;\
  }\
}

#define IS_VALID_XY(win,x,y)\
  (win->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE ?\
   x && x < win->wwidth && y && y < win->wheight :\
   x >= 0 && x <= win->wwidth && y >= 0 && y < win->wheight

static void __NEAR__ iGotoXY(int x,int y)
{
  if(!x || !y) return;
  active->cur_x = x-1;
  active->cur_y = y-1;
  if(active->cur_x >= active->wwidth) active->cur_x = active->wwidth-1;
  if(active->cur_y >= active->wheight) active->cur_y = active->wheight-1;
}

static void __NEAR__ set_xy(int x,int y)
{
  int width,height;
  if(!x || !y) return;
  x--; y--;
  active->cur_x = x;
  active->cur_y = y;
  width = active->wwidth;
  height = active->wheight;
  if((active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE)
  {
    width -= 1;
    height -= 1;
    active->cur_x++;
    active->cur_y++;
  }
  if(active->cur_x >= width) active->cur_x = width-1;
  if(active->cur_y >= height) active->cur_y = height-1;
}

static void __NEAR__ paint_cursor( void )
{
  TWindow * top;
  if(cursorwin && (cursorwin->iflags & IFLG_ENABLED) == IFLG_ENABLED)
  {
    __FIND_OVER(top,cursorwin,cursorwin->X1,cursorwin->Y1);
    if(!(top || !((cursorwin->iflags & IFLG_VISIBLE) == IFLG_VISIBLE) || cursorwin != active))
    {
        __vioSetCursorPos(active->X1+active->cur_x,active->Y1+active->cur_y);
    }
  }
}

void twGotoXY(int x,int y)
{
  set_xy(x,y);
  paint_cursor();
}

static TWindow * __NEAR__ makewin(int x1, int y1, int width, int height)
{
  TWindow *win;
  int size;
  win = PMalloc(sizeof(TWindow));
  if(!win) return NULL;
  memset(win,0,sizeof(TWindow));
  size = width*height*2;
  win->wsize = size;
  win->wwidth = width;
  win->wheight = height;
  if(!(win->body = PMalloc(size+sizeof(void *))))
  {
    PFREE(win);
    return NULL;
  }
  if(!(win->saved = PMalloc(size+sizeof(void *))))
  {
    PFREE(win->body);
    PFREE(win);
    return NULL;
  }
  *((void **)(win->body + size)) = win->body;
  *((void **)(win->saved + size)) = win->saved;
  win->X1 = x1;
  win->Y1 = y1;
  win->X2 = x1+width;
  win->Y2 = y1+height;
  __ATHEAD(win);
  return win;
}

/*
 *  Three basic functions for copying from buffer to screen:
 *  ========================================================
 *  updatescreencharfrombuff: low level implementation updatescreen family
 *  updatescreenchar:  correctly copyed user record from win to screen
 *  restorescreenchar: correctly copyed window memory from win to screen
 */

static void __NEAR__ updatescreencharfrombuff(TWindow *win,
                                              int x,
                                              int y,
                                              char *buff,
                                              char *accel)
{
  unsigned idx,aidx;
  if((win->iflags & IFLG_VISIBLE) == IFLG_VISIBLE)
  {
    TWindow * top;
    idx = (y*win->wwidth+x)*2;
    aidx = x*2;
    __FIND_OVER(top,win,x,y);
    if(top)
    {
      unsigned tidx,tx,ty;
      tx = win->X1 - top->X1 + x;
      ty = win->Y1 - top->Y1 + y;
      tidx = (tx + ty*top->wwidth)*2;
      top->saved[tidx] = buff[idx];
      top->saved[tidx+1] = buff[idx+1];
      if(accel)
      {
        TWindow *vis;
        __AT_POINT(vis,x+win->X1,y+win->Y1);
        tx = (win->X1 + x) - vis->X1;
        ty = (win->Y1 + y) - vis->Y1;
        tidx = (tx + ty*vis->wwidth)*2;
        accel[aidx] = vis->body[tidx];
        accel[aidx+1] = vis->body[tidx+1];
      }
      CHECK_WINS(top);
    }
    else
    {
      int ms_vis,is_hidden = 0;
      if(accel)
      {
        accel[aidx] = buff[idx];
        accel[aidx+1] = buff[idx+1];
      }
      else
      {
        if((win->iflags & IFLG_ENABLED) == IFLG_ENABLED)
        {
          ms_vis = __MsGetState();
          if(ms_vis > 0)
          {
             unsigned mx,my;
             __MsGetPos(&mx,&my);
             if(mx == win->X1+x && my == win->Y1+y)
             {
               is_hidden = 1;
               __MsSetState(0);
             }
          }
          __vioWriteBuff(win->X1+x,win->Y1+y,&buff[idx],2);
          if(is_hidden) __MsSetState(1);
          CHECK_WINS(win);
        }
      }
    }
  }
}

static void __NEAR__ updatescreenchar(TWindow *win,int x,int y,char *accel)
{
  updatescreencharfrombuff(win,x-1,y-1,win->body,accel);
}

static void __NEAR__ restorescreenchar(TWindow *win,int x,int y,char *accel)
{
  updatescreencharfrombuff(win,x-1,y-1,win->saved,accel);
}

/*
 *  Three basic functions for copying from buffer to screen:
 *  ========================================================
 *  updatewinmemcharfromscreen: correctly copied screen to window memory
 *  screen2win: quick implementation of copying screen to window memory
 *  snapshot:   snap shot of screen to win body
 */

void __NEAR__ updatewinmemcharfromscreen(TWindow *win,int x,int y,char *accel)
{
  unsigned idx,aidx;
  if((win->iflags & IFLG_VISIBLE) == IFLG_VISIBLE)
  {
    TWindow * top;
    idx = (y*win->wwidth+x)*2;
    aidx = x*2;
    __FIND_OVER(top,win,x,y);
    if(top)
    {
      unsigned tidx,tx,ty;
      tx = win->X1 - top->X1 + x;
      ty = win->Y1 - top->Y1 + y;
      tidx = (tx + ty*top->wwidth)*2;
      win->saved[idx] = top->saved[tidx];
      win->saved[idx+1] = top->saved[tidx+1];
      top->saved[tidx] = win->body[idx];
      top->saved[tidx+1] = win->body[idx+1];
      CHECK_WINS(top);
    }
    else
    {
      int ms_vis,is_hidden = 0;
      if(accel)
      {
        win->saved[idx] = accel[aidx];
        win->saved[idx+1] = accel[aidx+1];
      }
      else
      {
        ms_vis = __MsGetState();
        if(ms_vis > 0)
        {
           unsigned mx,my;
           __MsGetPos(&mx,&my);
           if(mx == win->X1+x && my == win->Y1+y)
           {
             is_hidden = 1;
             __MsSetState(0);
           }
        }
        __vioReadBuff(win->X1+x,win->Y1+y,&win->saved[idx],2);
        if(is_hidden) __MsSetState(1);
        CHECK_WINS(win);
      }
    }
  }
}

static void __NEAR__ screen2win(TWindow *win)
{
  unsigned i,lwidth;
  int ms_vis,is_hidden = 0;
  ms_vis = __MsGetState();
  if(ms_vis > 0)
  {
    unsigned mx,my;
    __MsGetPos(&mx,&my);
    if(mx >= win->X1 && mx <= win->X2 && my >= win->Y1 && my <= win->Y2)
    {
      is_hidden = 1;
      __MsSetState(0);
    }
  }
  lwidth = win->wwidth*2;
  for(i = 0;i < win->wheight;i++)
    __vioReadBuff(win->X1,win->Y1+i,&win->saved[i*lwidth],lwidth);
  if(is_hidden) __MsSetState(1);
  CHECK_WINS(win);
}

void twSnapShot(TWindow *win) /* for snapshot */
{
  unsigned i,lwidth;
  int ms_vis,is_hidden = 0;
  ms_vis = __MsGetState();
  if(ms_vis > 0)
  {
    unsigned mx,my;
    __MsGetPos(&mx,&my);
    if(mx >= win->X1 && mx <= win->X2 && my >= win->Y1 && my <= win->Y2)
    {
      is_hidden = 1;
      __MsSetState(0);
    }
  }
  lwidth = win->wwidth*2;
  for(i = 0;i < win->wheight;i++)
    __vioReadBuff(win->X1,win->Y1+i,&win->body[i*lwidth],lwidth);
  if(is_hidden) __MsSetState(1);
  CHECK_WINS(win);
}

/*
 *  Helpful functions:
 *  ==================
 *  savedwin2screen:  restore entire window memory to screen
 *  updatescreen:     restore entire user record of window to screen
 *  updatescreenline: restore one line from user record of window to screen
 *  updatewinmem:     update entire window memory from screen
 */

static void __NEAR__ savedwin2screen(TWindow *win)
{
  unsigned i,j;
  char accel[__TVIO_MAXSCREENWIDTH*2];
  int ms_vis,is_hidden = 0;
  if((win->iflags & IFLG_VISIBLE) == IFLG_VISIBLE)
  {
    unsigned mx,my;
    ms_vis = __MsGetState();
    if(ms_vis)
    {
      __MsGetPos(&mx,&my);
      if(mx >= win->X1 && mx <= win->X2 && my >= win->Y1 && my <= win->Y2)
      {
        is_hidden = 1;
        __MsSetState(0);
      }
    }
    for(i = 0;i < win->wheight;i++)
    {
      for(j = 0;j < win->wwidth;j++)
      {
        restorescreenchar(win,j+1,i+1,accel);
      }
      __vioWriteBuff(win->X1,win->Y1+i,accel,win->wwidth*2);
    }
    if(is_hidden) __MsSetState(1);
    CHECK_WINS(win);
  }
}

static void __NEAR__ updatescreen(TWindow *win)
{
  unsigned i,j;
  char accel[__TVIO_MAXSCREENWIDTH*2];
  int ms_vis,is_hidden = 0;
  if((win->iflags & IFLG_VISIBLE) == IFLG_VISIBLE)
  {
    unsigned mx,my;
    ms_vis = __MsGetState();
    if(ms_vis)
    {
      __MsGetPos(&mx,&my);
      if(mx >= win->X1 && mx <= win->X2 && my >= win->Y1 && my <= win->Y2)
      {
        is_hidden = 1;
        __MsSetState(0);
      }
    }
    for(i = 0;i < win->wheight;i++)
    {
      for(j = 0;j < win->wwidth;j++)
      {
        updatescreenchar(win,j+1,i+1,accel);
      }
      if((win->iflags & IFLG_ENABLED) == IFLG_ENABLED) __vioWriteBuff(win->X1,win->Y1+i,accel,win->wwidth*2);
    }
    if(is_hidden) __MsSetState(1);
    CHECK_WINS(win);
  }
}

static void __NEAR__ updatescreenline(TWindow *win,int y)
{
  unsigned i,line;
  char accel[__TVIO_MAXSCREENWIDTH*2];
  int ms_vis,is_hidden = 0;
  if((win->iflags & IFLG_VISIBLE) == IFLG_VISIBLE)
  {
    unsigned mx,my;
    ms_vis = __MsGetState();
    __MsGetPos(&mx,&my);
    line = win->Y1+y-1;
    if(ms_vis > 0 && my == line)
    {
      is_hidden = 1;
      __MsSetState(0);
    }
    for(i = 0;i < win->wwidth;i++)
      updatescreenchar(win,i+1,y,accel);
    if((win->iflags & IFLG_ENABLED) == IFLG_ENABLED) __vioWriteBuff(win->X1,line,accel,win->wwidth*2);
    if(is_hidden) __MsSetState(1);
    CHECK_WINS(win);
  }
}

static void __NEAR__ updatewinmem(TWindow *win)
{
  unsigned i,j;
  char accel[__TVIO_MAXSCREENWIDTH*2];
  int ms_vis,is_hidden = 0;
  if((win->iflags & IFLG_VISIBLE) == IFLG_VISIBLE)
  {
    unsigned mx,my;
    ms_vis = __MsGetState();
    if(ms_vis)
    {
      __MsGetPos(&mx,&my);
      if(mx >= win->X1 && mx <= win->X2 && my >= win->Y1 && my <= win->Y2)
      {
        is_hidden = 1;
        __MsSetState(0);
      }
    }
    for(i = 0;i < win->wheight;i++)
    {
      __vioReadBuff(win->X1,win->Y1+i,accel,win->wwidth*2);
      for(j = 0;j < win->wwidth;j++)
      {
        updatewinmemcharfromscreen(win,j,i,accel);
      }
    }
    if(is_hidden) __MsSetState(1);
  }
}

static void __NEAR__ WORDFILL(void *pointer,unsigned short filler,unsigned int len)
{
  unsigned long i,lim,lfill;
  lfill = (((long)filler) << 16) | ((long)filler);
  lim = len / 2;
  for(i = 0;i < lim;i++) ((long *)pointer)[i] = lfill;
  if(lim*2 != len) ((unsigned short *)pointer)[i*2] = filler;
}

static int __NEAR__ calc_title_off(int mode,int w,int slen)
{
 int stx;
   switch(mode)
   {
     case TMODE_LEFT:
                      stx = 2;
                      break;
     case TMODE_CENTER:
                      stx = (w - slen)/2 + 1;
                      break;
     default:
     case TMODE_RIGHT:
                      stx = w - slen;
                      break;
   }
   return stx;
}

static void __NEAR__ make_frame( void )
{
 unsigned i,w,h;
 char cfr;
 int sx,sy;
 char frm[8];
 int  up;
 char lt = 15,gr = 7,bl = 0;
 sx = active->cur_x;
 sy = active->cur_y;
 w = active->wwidth;
 h = active->wheight;
 cfr = active->cframe;
 up = 0;
 if(memcmp(active->Frame,"\0\0\0\0\0\0\0",8) != 0 &&
    memcmp(active->Frame,"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF",8) != 0)
                                             memcpy(frm,active->Frame,8);
 else
 {
   memcpy(frm,SINGLE_FRAME,8);
   if(tvioNumColors >= 16 && ((cfr >> 4) & 0x0F) != Black)
   {
     Color cfbk;
     if(memcmp(active->Frame,"\0\0\0\0\0\0\0",8) == 0) up = 1;
     else                                              up = 2;
     cfbk = (cfr >> 4) & 0x0F;
     lt = cfr;
     lt &= 0xF0;
     lt |= cfbk < 8 ? cfbk + 8 : White;
     gr = cfr;
     gr &= 0xF0;
     gr |= cfbk == LightGray ? Gray : LightGray;
     bl = cfr;
     bl &= 0xF0;
     bl |= cfbk == LightGray ? Black : Gray;
   }
 }
 __nls_Cvt866_to_Base(frm,sizeof(frm));
 iGotoXY(1,1);
 wputc(frm[0],up ? up == 1 ? lt : bl : cfr,0);
 for(i = 1; i < w - 1; i++)
 {
   iGotoXY(1+i,1);
   wputc(frm[1],up ? up == 1 ? lt : bl : cfr,0);
 }
 iGotoXY(w,1);
 wputc(frm[2],up ? gr : cfr,0);
 for(i = 1;i < h - 1;i++)
 {
   iGotoXY(1,i+1);
   wputc(frm[3],up ? up == 1 ? lt : bl : cfr,1);
 }
 for(i = 1;i < h - 1;i++)
 {
   iGotoXY(w,i+1);
   wputc(frm[4],up ? up == 1 ? bl : lt : cfr,1);
 }
 iGotoXY(1,h);
 wputc(frm[5],up ? gr : cfr,0);
 for(i = 1; i < w - 1; i++)
 {
   iGotoXY(1+i,h);
   wputc(frm[6],up ? up == 1 ? bl : lt : cfr,0);
 }
 iGotoXY(w,h);
 wputc(frm[7],up ? up == 1 ? bl : lt : cfr,0);
 if(active->Title)
 {
   int slen,stx;
   slen = strlen(active->Title);
   if(slen > w) slen = w;
   stx = calc_title_off(active->TitleMode,w,slen);
   for(i = 0;i < slen;i++) { iGotoXY(stx+i,1); wputc(active->Title[i],active->ctitle,0); }
 }
 if(active->Footer)
 {
   int slen,stx;
   slen = strlen(active->Footer);
   if(slen > w) slen = w;
   stx = calc_title_off(active->FooterMode,w,slen);
   for(i = 0;i < slen;i++) { iGotoXY(stx+i,h); wputc(active->Footer[i],active->cfooter,0); }
 }
 updatescreenline(active,1);
 updatescreenline(active,h);
 active->cur_x = sx;
 active->cur_y = sy;
}

static void __NEAR__ paint_internal( void )
{
  if((active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE) make_frame();
}

void twinDraw3DFrame(int x1, int y1, int x2, int y2,int flags)
{
 static char *frame[4] = { THIN_FRAME,MEDIUM_FRAME,THICK_FRAME,THICK_FRAME };
 int width,height;
 int i;
 char *frameptr = frame[(flags >> 4) & 0x03];
 width = x2 - x1;
 height = y2 - y1;
 twTextColor((flags & FRAME_DOWN) == FRAME_DOWN ? Black : White);
 twGotoXY(x1,y1); twPutChar(frameptr[0]);
 for(i = 0; i < width; i++)
 {
   twGotoXY(x1+1+i,y1); twPutChar(frameptr[1]);
 }
 for(i = 0;i < height;i++)
 {
   twGotoXY(x1,y1+i+1);
   twPutChar(frameptr[3]);
   if((flags & FRAME_THICK) == FRAME_THICK)
   {
     twGotoXY(x1+1,y1+i+1);
     twPutChar(frameptr[3]);
   }
 }
 if(!(flags & FRAME_HEIGHTMASK))
 {
   twTextColor(Gray);
   twGotoXY(x1,y1+height); twPutChar(frameptr[5]);
   twGotoXY(x1+width,y1); twPutChar(frameptr[2]);
 }
 else
 {
   twTextColor((flags & FRAME_DOWN) == FRAME_DOWN ? Black : White);
   twGotoXY(x1,y1+height); twPutChar(frameptr[5]);
   twTextColor((flags & FRAME_DOWN) == FRAME_DOWN ? White : Black);
   twGotoXY(x1+width,y1); twPutChar(frameptr[2]);
 }
 twTextColor((flags & FRAME_DOWN) == FRAME_DOWN ? White : Black);
 for(i = 0; i < width; i++)
 {
   twGotoXY(x1+1+i,y1+height); twPutChar(frameptr[6]);
 }
 for(i = 0;i < height;i++)
 {
   twGotoXY(x1+width,y1+i+1);
   twPutChar(frameptr[4]);
   if((flags & FRAME_THICK) == FRAME_THICK)
   {
     twGotoXY(x1-1+width,y1+i+1);
     twPutChar(frameptr[4]);
   }
 }
 twGotoXY(x1+width,y1+height); twPutChar(frameptr[7]);
 twTextColor(Black);
}


TWindow * twCreateWin(int x1, int y1, int width, int height, int flags)
{
  TWindow *win;
  if((flags & TWS_FRAMEABLE) == TWS_FRAMEABLE) { width ++; height ++; }
  if(!(win = makewin(x1-1,y1-1,width,height))) return NULL;
  win->flags = flags;
  win->iflags = IFLG_ENABLED;
  memcpy(win->Frame,SINGLE_FRAME,8);
  memset(win->body,0,win->wsize);
  CHECK_WINS(win);
  win->ctext = win->cframe = win->ctitle = win->cfooter = LightGray;
  win->cur_x = win->cur_y = 1;
  twUseWin(win);
  paint_internal();
  if((flags & TWS_VISIBLE) == TWS_VISIBLE) twShowWin(win);
  return win;
}

void twShowWin(TWindow* w)
{
  if(!(w->iflags & IFLG_VISIBLE) == IFLG_VISIBLE)
  {
    w->iflags |= IFLG_VISIBLE;
    __unlistwin(w);
    __ATHEAD(w);
    screen2win(w);
    updatescreen(w);
    if((w->flags & TWS_CURSORABLE) == TWS_CURSORABLE)
    {
      cursorwin = w;
    }
    paint_cursor();
  }
}

void twShowWinOnTop(TWindow* w)
{
  if((w->iflags & IFLG_VISIBLE) == IFLG_VISIBLE) twHideWin(w);
  w->iflags |= IFLG_VISIBLE;
  __unlistwin(w);
  __ATHEAD(w);
  updatewinmem(w);
  updatescreen(w);
  if((w->flags & TWS_CURSORABLE) == TWS_CURSORABLE)
  {
    cursorwin = w;
  }
  paint_cursor();
}

void twShowWinBeneath(TWindow* w,TWindow *prev)
{
  if((w->iflags & IFLG_VISIBLE) == IFLG_VISIBLE) twHideWin(w);
  w->iflags |= IFLG_VISIBLE;
  __unlistwin(w);
  __ATWIN(w,prev);
  updatewinmem(w);
  updatescreen(w);
}

void twHideWin(TWindow *w)
{
  if(cursorwin == w) __vioSetCursorType(CUR_OFF);
  savedwin2screen(w);
  w->iflags &= ~IFLG_VISIBLE;
}

void twDestroyWin(TWindow *win)
{
  twHideWin(win);
  if(win->Title) PFREE(win->Title);
  if(win->Footer) PFREE(win->Footer);
  PFREE(win->body);
  PFREE(win->saved);
  if(active == win)
  {
     if(win == head)   active = win->next;
     else              active = __prevwin(win);
  }
  __unlistwin(win);
  if(cursorwin == win) cursorwin = __findcursorablewin();
  paint_cursor();
  PFREE(win);
}

void twGetWinPos(TWindow *win,int *x1,int *y1,int *x2,int *y2)
{
  *x1 = win->X1+1;
  *y1 = win->Y1+1;
  *x2 = win->X2;
  *y2 = win->Y2;
  if((win->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE)
  {
   (*x2)--;
   (*y2)--;
  }
}

void twMoveWin(TWindow *win,int dx,int dy)
{
  int x,y,vis;
  TWindow *prev;
  vis = (win->iflags & IFLG_VISIBLE) == IFLG_VISIBLE;
  x = twWhereX();
  y = twWhereY();
  prev = __prevwin(win);
  if(vis) twHideWin(win);
  win->X1 += dx;
  win->Y1 += dy;
  win->X2 += dx;
  win->Y2 += dy;
  if(vis) twShowWinBeneath(win,prev);
  twGotoXY(x,y);
}

void twResizeWin(TWindow *win,int width,int height)
{
  int oldw,oldh,osize,ncopy,delta,fillsize;
  int from,to,size,i,loop,start;
  int x,y,vis;
  short wfill;
  TWindow * tmp,*prev;
  char * newbody;
  x = twWhereX();
  y = twWhereY();
  vis = (win->iflags & IFLG_VISIBLE) == IFLG_VISIBLE;
  prev = __prevwin(win);
  if(vis) twHideWin(win);
  size = width*height*2;
  newbody = PMalloc(size+sizeof(void *));
  if(!newbody) winerr("Out of memory!");
  *((void **)(newbody + size)) = newbody;
  oldw = win->wwidth;
  oldh = win->wheight;
  /* --- Compute copy parameters --- */
  to = (win->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE ? 2 : 0;
  from = (win->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE ? 2 : 0;
  size = width*2;
  osize = oldw*2;
  delta = 0;
  if((win->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE) delta = 4;
  start = (win->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE ? 1 : 0;
  loop = min(oldh,height);
  ncopy = min(size,osize);
  wfill = (' ' & 0x00FF) | ((win->ctext << 8) & 0xFF00);
  fillsize = (size-osize)/2;
  for(i = start;i < loop;i++)
  {
      memcpy(&newbody[to],&win->body[from],ncopy);
      if(oldw < width) WORDFILL(&newbody[to+ncopy-delta],wfill,fillsize);
      CHECK_WINS(win);
      to += size;
      from += osize;
  }
  if(oldh < height)
  {
    for(;i < height;i++)
    {
       WORDFILL(&newbody[to],wfill,size/2);
       to += size;
    }
  }
  PFREE(win->saved);
  win->wsize = width*height*2;
  win->wwidth = width;
  win->wheight = height;
  win->saved = PMalloc(win->wsize+sizeof(char *));
  if(!win->saved) winerr("Out of memory!");
  PFREE(win->body);
  *((void **)(win->saved + win->wsize)) = win->saved;
  win->body = newbody;
  win->X2 = win->X1 + width;
  win->Y2 = win->Y1 + height;
  CHECK_WINS(win);
  tmp = active;
  active = win;
  paint_internal();
  active = tmp;
  if(vis) twShowWinBeneath(win,prev);
  twGotoXY(x,y);
}

void twCentredWin(TWindow * win,TWindow *parent)
{
  unsigned ww,wh,pww,pwh;
  int vis = (win->iflags & IFLG_VISIBLE) == IFLG_VISIBLE;
  if(vis) twHideWin(win);
  ww = win->wwidth;
  wh = win->wheight;
  if(!parent)
  {
    pww = tvioWidth;
    pwh = tvioHeight;
  }
  else
  {
    pww = parent->wwidth;
    pwh = parent->wheight;
  }
  win->X1 = ( pww - ww ) / 2;
  win->X2 = ( pww + ww ) / 2;
  win->Y1 = ( pwh - wh ) / 2;
  win->Y2 = ( pwh + wh ) / 2;
  if(vis) twShowWin(win);
}

void twClearWinEx( unsigned char filler )
{
  int to,size,lsize,i,loop,start,delta;
  unsigned short wfill;
  to = (active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE ? 2 : 0;
  size = active->wwidth;
  delta = 0;
  if((active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE) delta = 2;
  start = (active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE ? 1 : 0;
  loop = active->wheight;
  if((active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE) loop--;
  wfill = (filler & 0x00FF) | ((active->ctext << 8) & 0xFF00);
  lsize = size*2;
  for(i = start;i < loop;i++)
  {
      WORDFILL(&active->body[to + i*lsize],wfill,(size-delta));
      CHECK_WINS(active);
      updatescreenline(active,i+1);
  }
  active->cur_x = active->cur_y = 1;
}

void twClearWin( void )
{
  twClearWinEx(' ');
}

void twClrEOLEx(unsigned char filler )
{
  int size;
  short wfill;
  wfill = (filler & 0x00FF) | ((active->ctext << 8) & 0xFF00);
  size = active->wwidth - active->cur_x;
  if((active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE) size--;
  WORDFILL(&active->body[(active->cur_x + active->cur_y*active->wwidth)*2],wfill,size);
  CHECK_WINS(active);
  updatescreenline(active,active->cur_y+1);
}

void twClrEOL( void )
{
  twClrEOLEx(' ');
}

TWindow * twUseWin( TWindow * win)
{
  TWindow *ret;
  ret = active;
  active = win;
  if((active->flags & TWS_CURSORABLE) == TWS_CURSORABLE)
  {
     cursorwin = win;
  }
  paint_cursor();
  return ret;
}

TWindow * twUsedWin( void )
{
  return active;
}

int twWhereX( void )
{
  return (active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE ? active->cur_x : active->cur_x+1;
}

int twWhereY( void )
{
  return (active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE ? active->cur_y : active->cur_y+1;
}

static void __NEAR__ wputc(char ch,char color,int update)
{
  int idx;
  idx = (active->cur_x + active->cur_y*active->wwidth)*2;
  active->body[idx] = ch;
  active->body[idx+1] = color;
  CHECK_WINS(active);
  if(update) updatescreenchar(active,active->cur_x+1,active->cur_y+1,NULL);
}

void twPutChar(char ch)
{
  unsigned cx,cy;
  if((active->flags & TWS_NLS866) == TWS_NLS866) __nls_Cvt866_to_Base(&ch,1);
  cx = active->cur_x;
  cy = active->cur_y;
  if(IS_VALID_XY(active,cx,cy))
  {
    wputc(ch,active->ctext,1);
    active->cur_x++;
  }
}

char twGetChar( void )
{
  int idx,cx,cy;
  cx = active->X1 + active->cur_x;
  cy = active->Y1 + active->cur_y;
  idx = (cx + cy*active->wwidth)*2;
  return active->body[idx];
}

int twPutS(const char *str)
{
  int freq = 0;
  unsigned vidx,cx,cy;
  char ch;
  char *__nls = NULL,* __nls_ptr;
  if((active->flags & TWS_NLS866) == TWS_NLS866)
  {
     int len;
     len = strlen(str);
     __nls = PMalloc(len+1);
     if(__nls)
     {
       strcpy(__nls,str);
       __nls_Cvt866_to_Base(__nls,len);
     }
     else __nls = (char *)str;
  }
  else __nls = (char *)str;
  __nls_ptr = __nls;
  vidx = (active->cur_x + active->cur_y*active->wwidth)*2;
  while((ch=*__nls++)!=0)
  {
    if(ch == '\n')
    {
        updatescreenline(active,active->cur_y+1);
        set_xy(1,active->cur_y+1);
        vidx = (active->cur_x + active->cur_y*active->wwidth)*2;
        continue;
    }
    if(ch == '\r')
    {
        set_xy(1,active->cur_y);
        vidx = (active->cur_x + active->cur_y*active->wwidth)*2;
        continue;
    }
    cx = active->cur_x;
    cy = active->cur_y;
    if(IS_VALID_XY(active,cx,cy))
    {
        active->body[vidx++] = ch;
        active->body[vidx++] = active->ctext;
        active->cur_x++;
        freq++;
    }
  }
  updatescreenline(active,active->cur_y+1);
  paint_cursor();
  if(__nls_ptr != str) PFREE(__nls_ptr);
  return freq;
}

int twPrintF(const char *fmt,...)
{
  char *buff;
  int ret;
  va_list args;
  buff = PMalloc(tvioWidth*tvioHeight); /*Danger - format can be great that
                                         allocated buffer, but it meaningless */
  if(!buff) return 0;
  va_start(args,fmt);
  vsprintf(buff,fmt,args);
  va_end(args);
  ret = twPutS(buff);
  PFREE(buff);
  return ret;
}

int twDirectWrite(int x, int y,const void *str,unsigned len)
{
  int rlen,j,ioff;
  char *__nls = NULL;
  if((active->flags & TWS_NLS866) == TWS_NLS866)
  {
     __nls = PMalloc(len);
     if(__nls)
     {
       memcpy(__nls,str,len);
       __nls_Cvt866_to_Base(__nls,len);
     }
     else __nls = (char *)str;
  }
  else __nls = (char *)str;
  set_xy(x,y);
  rlen = active->wwidth;
  if((active->flags & TWS_FRAMEABLE) == TWS_FRAMEABLE) rlen-=2;
  rlen -= active->cur_x;
  rlen = min(rlen,len);
  ioff = ((active->cur_x)+(active->cur_y)*active->wwidth)*2;
  for(j = 0;j < rlen;j++,ioff+=2)
  {
    active->body[ioff] = ((const char *)__nls)[j];
    active->body[ioff+1] = active->ctext;
  }
  CHECK_WINS(active);
  updatescreenline(active,active->cur_y+1);
  paint_cursor();
  active->cur_x+=rlen;
  if(__nls != str) PFREE(__nls);
  return rlen;
}

void twWriteBuffer(TWindow *win,int x,int y,const void *buff,unsigned len)
{
  int rlen;
  rlen = x+len > win->wwidth ? win->wwidth-x+1 : len;
  if(y <= win->wheight)
  {
    memcpy(&win->body[((x-1)+(y-1)*win->wwidth)*2],buff,rlen*2);
    CHECK_WINS(win);
    updatescreenline(win,y);
    paint_cursor();
  }
}

void twReadBuffer(TWindow *win,int x,int y,void *buff,unsigned len)
{
  int rlen;
  rlen = x+len > win->wwidth ? win->wwidth-x+1 : len;
  if(y <= win->wheight)
  {
    memcpy(buff,&win->body[((x-1)+(y-1)*win->wwidth)*2],rlen*2);
    CHECK_WINS(win);
    paint_cursor();
  }
}

void twFreezeWin(TWindow *win)
{
  win->iflags &= ~IFLG_ENABLED;
}

void twRefreshWin(TWindow *win)
{
  win->iflags |= IFLG_ENABLED;
  updatescreen(win);
  paint_cursor();
}

void * twGetUsrData(TWindow *win) { return win->usrData; }

void * twSetUsrData(TWindow *win,void *data)
{
  void *ret;
  ret = win->usrData;
  win->usrData = data;
  return ret;
}  