/****************************************************************************\
*   NLM386 - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/nlm386.c                               *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of NLM-32 (Novell Loadable Module) file  *
* format decoder.                                                            *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>

#include "pmalloc.h"
#include "bin_util.h"
#include "plugins/bin/nlm386.h"
#include "kbd_code.h"
#include "codeguid.h"
#include "bmfile.h"
#include "bconsole.h"
#include "reg_form.h"
#include "tstrings.h"
#include "biewutil.h"
#include "biewhelp.h"
#include "__os_dep.h"
#include "_sys_dep.h"

static Nlm_Internal_Fixed_Header nlm;

static int __NEAR__ FindPubName(char *buff,unsigned cb_buff,unsigned long pa);
static void nlm_ReadPubNameList(BGLOBAL handle,int interact);

static BGLOBAL nlm_cache = &bNull;

int nlmLowMemFunc( unsigned long need_mem )
{
  if(!fmtActiveState)
  {
    if(PubNames)
    {
       la_Destroy(PubNames);
       PubNames = NULL;
       return 1;
    }
  }
  return 0;
}

unsigned long ShowNLMHeader( void )
{
  unsigned long fpos;
  char modName[NLM_MODULE_NAME_SIZE];
  TWindow * w;
  unsigned keycode;
  fpos = BMGetCurrFilePos();
  w = CrtHlpWnd(" NetWare Loadable Module ",54,23);
  twGotoXY(1,1);
  strncpy(modName,&nlm.nlm_moduleName[1],(int)nlm.nlm_moduleName[0]);
  modName[(unsigned)nlm.nlm_moduleName[0]] = 0;
  twPrintF("Module (Version)              = %s (%02hu.%02hu)\n"
           "Code image offset             = %08lXH\n"
           "Code image size               = %08lXH\n"
           "Data image offset             = %08lXH\n"
           "Data image size               = %08lXH\n"
           "Uninitialized data size       = %08lXH\n"
           "Custom data offset            = %08lXH\n"
           "Custom data size              = %08lXH\n"
           "Module dependency offset      = %08lXH\n"
           "Number of Module dependencies = %08lXH\n"
           "Relocation fixup offset       = %08lXH\n"
           "Number of relocations fixup   = %08lXH\n"
           "External reference offset     = %08lXH\n"
           "Number of external references = %08lXH\n"
           "Public offset                 = %08lXH\n"
           "Number of public              = %08lXH\n"
           "Debug info offset             = %08lXH\n"
           "Number of debug records       = %08lXH\n"
           ,modName
           ,(unsigned short)nlm.nlm_version
           ,(unsigned short)(nlm.nlm_version >> 16)
           ,nlm.nlm_codeImageOffset
           ,nlm.nlm_codeImageSize
           ,nlm.nlm_dataImageOffset
           ,nlm.nlm_dataImageSize
           ,nlm.nlm_uninitializedDataSize
           ,nlm.nlm_customDataOffset
           ,nlm.nlm_customDataSize
           ,nlm.nlm_moduleDependencyOffset
           ,nlm.nlm_numberOfModuleDependencies
           ,nlm.nlm_relocationFixupOffset
           ,nlm.nlm_numberOfRelocationFixups
           ,nlm.nlm_externalReferencesOffset
           ,nlm.nlm_numberOfExternalReferences
           ,nlm.nlm_publicsOffset
           ,nlm.nlm_numberOfPublics
           ,nlm.nlm_debugInfoOffset
           ,nlm.nlm_numberOfDebugRecords);
  twSetColor(White,Black);
  twPrintF("Code start offset             = %08lXH [Enter]",nlm.nlm_codeStartOffset);
  twClrEOL(); twPrintF("\n");
  twSetColor(Yellow,Blue);
  twPrintF("Exit procedure offset         = %08lXH [Ctrl+Enter]",nlm.nlm_exitProcedureOffset);
  twClrEOL(); twPrintF("\n");
  twSetColor(Black,LightGray);
  twPrintF("Check unload procedure offset = %08lXH\n"
           "Module type                   = %08lXH\n"
           "Flags                         = %08lXH"
           ,nlm.nlm_checkUnloadProcedureOffset
           ,nlm.nlm_moduleType
           ,nlm.nlm_flags);
  do
  {
    keycode = GetEvent(drawEmptyPrompt,w);
    if(keycode == KE_ENTER)
    {
      fpos = nlm.nlm_codeImageOffset + nlm.nlm_codeStartOffset;
      break;
    }
    else
      if(keycode == KE_CTL_ENTER)
      {
        fpos = nlm.nlm_codeImageOffset + nlm.nlm_exitProcedureOffset;
        break;
      }
  }while(keycode != KE_ESCAPE);
  CloseWnd(w);
  return fpos;
}

unsigned long ShowNewNLM( void )
{
  unsigned long fpos,ssize,m,d,sharedEntry,sharedExit;
  char modName[256];
  unsigned char len;
  TWindow *w;
  unsigned keycode;
  sharedEntry = sharedExit = 0;
  fpos = BMGetCurrFilePos();
  w = CrtHlpWnd(" NetWare Loadable Module ",74,23);
  twGotoXY(1,1);
  BMSeek(sizeof(Nlm_Internal_Fixed_Header),BM_SEEK_SET);
  len = BMReadByte();
  BMReadBuffer(modName,len + 1);
  ssize = BMReadDWord();
  BMSeek(4,BM_SEEK_CUR); /* skip reserved */
  twPrintF("%s\n"
           "Stack size                    = %08lXH\n"
           ,modName
           ,ssize);
  BMReadBuffer(modName,5);
  modName[5] = 0;
  twPrintF("Old thread name               = %s\n",modName);
  len = BMReadByte();
  BMReadBuffer(modName,len + 1);
  twPrintF("Screen name                   = %s\n",modName);
  len = BMReadByte();
  BMReadBuffer(modName,len + 1);
  twPrintF("Thread name                   = %s",modName);
  while(1)
  {
    BMReadBuffer(modName,9);
    modName[9] = 0;
    if(memcmp(modName,"VeRsIoN#",8) == 0)
    {
      BMSeek(-1,BM_SEEK_CUR);
      ssize = BMReadDWord();
      d = BMReadDWord();
      m = BMReadDWord();
      twPrintF("\nVersion ( Revision )          = %lu.%lu ( %08lXH )\n",ssize,d,m);
      ssize = BMReadDWord();
      m     = BMReadDWord();
      d     = BMReadDWord();
      twPrintF("Date (DD.MM.YY)               = %lu.%lu.%lu",d,m,ssize);
    }
    else
      if(memcmp(modName,"CoPyRiGhT",9) == 0)
      {
        BMSeek(1,BM_SEEK_CUR);
        len = BMReadByte();
        BMReadBuffer(modName,len + 1);
        twPrintF("\nCopyright = %s",modName);
      }
      else
        if(memcmp(modName,"MeSsAgEs",8) == 0)
        {
          BMSeek(-1,BM_SEEK_CUR);
          ssize = BMReadDWord();
          twPrintF("\nLanguage                      = %08lXH\n",ssize);
          ssize = BMReadDWord();
          m = BMReadDWord();
          d = BMReadDWord();
          twPrintF("Messages (offset/length/count)= %08lXH/%08lXH/%08lXH\n",ssize,m,d);
          ssize = BMReadDWord();
          m = BMReadDWord();
          d = BMReadDWord();
          twPrintF("Help (offset/length/dataOff)  = %08lXH/%08lXH/%08lXH\n",ssize,m,d);
          ssize = BMReadDWord();
          m = BMReadDWord();
          twPrintF("SharedCode (offset/length)    = %08lXH/%08lXH\n",ssize,m);
          ssize = BMReadDWord();
          m = BMReadDWord();
          twPrintF("SharedData (offset/length)    = %08lXH/%08lXH\n",ssize,m);
          ssize = BMReadDWord();
          m = BMReadDWord();
          twPrintF("SharedReloc (offset/count)    = %08lXH/%08lXH\n",ssize,m);
          ssize = BMReadDWord();
          m = BMReadDWord();
          twPrintF("SharedExtRef (offset/count)   = %08lXH/%08lXH\n",ssize,m);
          ssize = BMReadDWord();
          m = BMReadDWord();
          twPrintF("SharedPublics (offset/count)  = %08lXH/%08lXH\n",ssize,m);
          ssize = BMReadDWord();
          m = BMReadDWord();
          twPrintF("SharedDebugRec (offset/count) = %08lXH/%08lXH\n",ssize,m);
          sharedEntry = BMReadDWord();
          twSetColor(White,Black);
          twPrintF("Shared initialization offset  = %08lXH [Enter]",sharedEntry);
          twClrEOL(); twPrintF("\n");
          sharedExit = BMReadDWord();
          twSetColor(Yellow,Blue);
          twPrintF("Shared exit procedure offset  = %08lXH [Ctrl+Enter]",sharedExit);
          twClrEOL(); twPrintF("\n");
          twSetColor(Black,LightGray);
          ssize = BMReadDWord();
          twPrintF("Product ID                    = %08lXH",ssize);
        }
        else
          if(memcmp(modName,"CuStHeAd",8) == 0)
          {
            unsigned long hdr;
            BMSeek(-1,BM_SEEK_CUR);
            ssize = BMReadDWord();
            d = BMReadDWord();
            m = BMReadDWord();
            BMReadBuffer(modName,8);
            modName[8] = 0;
            hdr = BMReadDWord();
            twPrintF("\nCustHead (name/hdrOff/hdrLen/dataOff/dataLen) = %s/%08lXH/%08lXH/%08lXH/%08lHX",modName,hdr,ssize,d,m);
          }
          else
            if(memcmp(modName,"CyGnUsEx",8) == 0)
            {
              BMSeek(-1,BM_SEEK_CUR);
              d = BMReadDWord();
              m = BMReadDWord();
              twPrintF("\nCygnus (offset/length) = %08lXH/%08lXH",d,m);
            }
            else break;
  }
  do
  {
    keycode = GetEvent(drawEmptyPrompt,w);
    if(keycode == KE_ENTER)
    {
       if(sharedEntry) fpos = sharedEntry + nlm.nlm_codeImageOffset;
       break;
    }
    else
      if(keycode == KE_CTL_ENTER && sharedExit)
      {
        fpos = sharedExit + nlm.nlm_codeImageOffset;
        break;
      }
  }while(keycode != KE_ESCAPE);
  CloseWnd(w);
  return fpos;
}

unsigned NLMExtRefNumItems(BGLOBAL handle)
{
  return (unsigned)nlm.nlm_numberOfExternalReferences;
}

int      __ReadExtRefNamesNLM(BGLOBAL handle,memArray * obj,unsigned n)
{
 unsigned i;
 bioSeek(handle,nlm.nlm_externalReferencesOffset,SEEKF_START);
 for(i = 0;i < n;i++)
 {
   char stmp[256];
   unsigned char length;
   unsigned long nrefs;
   length = bioReadByte(handle);
   if(IsKbdTerminate()) break;
   bioReadBuffer(handle,stmp,length);
   stmp[length] = 0;
   nrefs = bioReadDWord(handle);
   bioSeek(handle,nrefs*4,SEEKF_CUR);
   if(!ma_AddString(obj,stmp,1)) break;
 }
 return 1;
}

static unsigned long __NEAR__ CalcEntryNLM(unsigned ord,int dispmsg)
{
 unsigned char length;
 unsigned i;
 unsigned long ret,fpos,cpos;
 fpos = BMGetCurrFilePos();
 cpos = nlm.nlm_publicsOffset;
 for(i = 0;i < ord;i++)
 {
   length = BMReadByteEx(cpos,SEEKF_START); cpos+=length + 5;
 }
 length = BMReadByteEx(cpos,SEEKF_START); cpos+=length + 1;
 ret = BMReadDWordEx(cpos,SEEKF_START);
 ret &= 0x00FFFFFFL;
 ret += nlm.nlm_codeImageOffset;
 BMSeek(fpos,SEEKF_START);
 if(ret > BMGetFLength())
 {
    ret = fpos;
    if(dispmsg) ErrMessageBox("Entry point is outside of this file",NULL);
 }
 return ret;
}

unsigned NLMNamesNumItems(BGLOBAL handle)
{
  return (unsigned)nlm.nlm_numberOfPublics;
}

int      NLMNamesReadItems(BGLOBAL handle,memArray * obj,unsigned nnames)
{
 unsigned char length;
 unsigned i;
 bioSeek(handle,nlm.nlm_publicsOffset,SEEKF_START);
 for(i = 0;i < nnames;i++)
 {
   char stmp[256];
   length = bioReadByte(handle);
   if(IsKbdTerminate()) break;
   if(length > 66)
   {
     bioReadBuffer(handle,stmp,66);
     bioSeek(handle,length - 66,SEEKF_CUR);
     strcat(stmp,">>>");
   }
   else { bioReadBuffer(handle,stmp,length); stmp[length] = 0; }
   bioSeek(handle,4L,SEEKF_CUR);
   if(!ma_AddString(obj,stmp,1)) break;
 }
 return 1;
}

unsigned long ShowExtRefNLM( void )
{
  fmtShowList(NLMExtRefNumItems,
              __ReadExtRefNamesNLM,
              EXT_REFER,
              LB_SORTABLE,
              NULL);
   return BMGetCurrFilePos();
}

unsigned NLMModRefNumItems(BGLOBAL handle)
{
  return (unsigned)nlm.nlm_numberOfModuleDependencies;
}

int      __ReadModRefNamesNLM(BGLOBAL handle,memArray * obj,unsigned nnames)
{
 unsigned char length;
 unsigned i;
 bioSeek(handle,nlm.nlm_moduleDependencyOffset,SEEKF_START);
 for(i = 0;i < nnames;i++)
 {
   char stmp[256];
   length = bioReadByte(handle);
   if(IsKbdTerminate()) break;
   if(length > 66)
   {
     bioReadBuffer(handle,stmp,66);
     bioSeek(handle,length - 66,SEEKF_CUR);
     strcat(stmp,">>>");
   }
   else { bioReadBuffer(handle,stmp,length); stmp[length] = 0; }
   if(!ma_AddString(obj,stmp,1)) break;
 }
 return 1;
}


unsigned long ShowModRefNLM( void )
{
  fmtShowList(NLMModRefNumItems,
              __ReadModRefNamesNLM,
              MOD_REFER,
              LB_SORTABLE,
              NULL);
   return BMGetCurrFilePos();
}

unsigned long ShowPubNamNLM( void )
{
  unsigned long fpos = BMGetCurrFilePos();
  int ret;
  ret = fmtShowList(NLMNamesNumItems,NLMNamesReadItems,
                    EXP_TABLE,
                    LB_SELECTIVE,NULL);
  if(ret != -1)
  {
    fpos = CalcEntryNLM(ret,1);
  }
  return fpos;
}

/***************************************************************************/
/************************  FOR NLM  ****************************************/
/***************************************************************************/
static char __codelen;

typedef struct tagRELOC_NLM
{
  unsigned long offset;
  unsigned long nameoff; // if refnum == -1 then internal
}RELOC_NLM;

static linearArray *RelocNlm = NULL;

int nlm_compare_s(const void __HUGE__ *e1,const void __HUGE__ *e2)
{
  const RELOC_NLM __HUGE__ *r1,__HUGE__ *r2;
  int ret;
  r1 = e1;
  r2 = e2;
  if(r1->offset < r2->offset) ret = -1;
  else
    if(r1->offset > r2->offset) ret = 1;
    else                      ret = 0;
  return ret;
}

int nlm_compare_f(const void __HUGE__ *e1,const void __HUGE__ *e2)
{
  const RELOC_NLM __HUGE__ *r1,__HUGE__ *r2;
  int ret;
  r1 = e1;
  r2 = e2;
  if(r1->offset < r2->offset) return -1;
  else
    if(r1->offset > r2->offset) ret = 1;
    else                        ret = 0;
  return ret;
}

static void __NEAR__ BuildRelocNlm( void )
{
//  void *tmp;
  unsigned i,j;
  unsigned long fpos,cpos,val,niter,noff;
  unsigned char len;
  TWindow * w,*usd;
  RELOC_NLM rel;
  usd = twUsedWin();
  if(!(RelocNlm = la_Build(0,sizeof(RELOC_NLM),1))) return;
  w = CrtHlpWnd(" System is busy ",49,1);
  if(!PubNames) nlm_ReadPubNameList(BMbioHandle(),1);
  twUseWin(w);
  twGotoXY(1,1);
  twPutS(" Building reference chains for NLM format");
  twUseWin(usd);
  fpos = BMGetCurrFilePos();
  // -- for external references
  cpos = nlm.nlm_externalReferencesOffset;
  for(j = 0;j < nlm.nlm_numberOfExternalReferences;j++)
  {
    noff = cpos;
    len = BMReadByteEx(cpos,SEEKF_START); cpos += len + 1;
    niter = BMReadDWordEx(cpos,SEEKF_START); cpos += 4;
    for(i = 0;i < niter;i++)
    {
      val = BMReadDWordEx(cpos,SEEKF_START); cpos += 4;
      rel.offset = (val&0x00FFFFFFL) + nlm.nlm_codeImageOffset;
      rel.nameoff = noff;
      if(!la_AddData(RelocNlm,&rel,1)) goto next;
    }
  }
  // -- for internal references
  cpos = nlm.nlm_relocationFixupOffset;
  for(j = 0;j < nlm.nlm_numberOfRelocationFixups;j++)
  {
    val = BMReadDWord();
    rel.offset = (val&0x00FFFFFFL) + nlm.nlm_codeImageOffset;
    rel.nameoff = -1;
    if(!la_AddData(RelocNlm,&rel,1)) break;
  }
  next:
  la_Sort(RelocNlm,nlm_compare_s);
  BMSeek(fpos,SEEKF_START);
  CloseWnd(w);
}

static int __NEAR__ BuildReferStrNLM(char *str,RELOC_NLM*rne)
{
  unsigned long fpos,val;
  unsigned char len;
  char name[256];
  BGLOBAL b_cache;
  b_cache = nlm_cache;
  fpos = BMGetCurrFilePos();
  bioSeek(b_cache,rne->nameoff,BM_SEEK_SET);
  if(rne->nameoff != 0xFFFFFFFFL)
  {
    len = bioReadByte(b_cache);
    bioReadBuffer(b_cache,name,len);
    name[len] = 0;
    strcat(str,name);
  }
  else
  {
    val = BMReadDWordEx(rne->offset,SEEKF_START);
    if(FindPubName(name,sizeof(name),val))
    {
      strcat(str,name);
    }
    else
    {
      strcat(str,"(*this)+");
      strcat(str,Get8Digit(val));
    }
  }
  BMSeek(fpos,SEEKF_START);
  return 1;
}

static int AppendNLMRef(char *str,unsigned long ulShift,int flags,int codelen,unsigned long r_sh)
{
  RELOC_NLM *rnlm,key;
  int ret;
  char buff[400];
  if(!nlm.nlm_numberOfExternalReferences || nlm.nlm_externalReferencesOffset >= BMGetFLength()) return 0;
  if(!RelocNlm) BuildRelocNlm();
  key.offset = ulShift;
  __codelen = codelen;
  rnlm = la_Find(RelocNlm,&key,nlm_compare_f);
  if(!rnlm) ret = 0;
  else      ret = BuildReferStrNLM(str,rnlm);
  if(!ret && (flags & APREF_TRY_LABEL))
  {
     if(!PubNames) nlm_ReadPubNameList(BMbioHandle(),1);
     if(FindPubName(buff,sizeof(buff),r_sh))
     {
       strcat(str,buff);
       if(!DumpMode && !EditMode) GidAddGoAddress(str,r_sh);
       ret = 1;
     }
  }
  return ret;
}

int IsNLM( void )
{
  char ctrl[NLM_SIGNATURE_SIZE];
  BMReadBufferEx(ctrl,NLM_SIGNATURE_SIZE,0,SEEKF_START);
  return memcmp(ctrl,NLM_SIGNATURE,NLM_SIGNATURE_SIZE) == 0;
}

void NLMinit( void )
{
  BGLOBAL main_handle;
  PMRegLowMemCallBack(nlmLowMemFunc);
  BMReadBufferEx(&nlm,sizeof(Nlm_Internal_Fixed_Header),0,SEEKF_START);
  main_handle = BMbioHandle();
  if((nlm_cache = bioDup(main_handle)) == &bNull) nlm_cache = main_handle;
}

void NLMdestroy( void )
{
  BGLOBAL main_handle;
  main_handle = BMbioHandle();
  if(nlm_cache != &bNull && nlm_cache != main_handle) bioClose(nlm_cache);
  PMUnregLowMemCallBack(nlmLowMemFunc);
}

int NLMbitness(unsigned long off) { return DAB_USE32; }

void NLMAddrResolv(char *addr,unsigned long cfpos)
{
  if(cfpos < sizeof(Nlm_Internal_Fixed_Header))
  {
    sprintf(addr,"nlm32h:%s",Get2Digit(cfpos));
  }
}

unsigned long HelpNLM( void )
{
  hlpDisplay(10007);
  return BMGetCurrFilePos();
}

static void nlm_ReadPubName(BGLOBAL b_cache,unsigned long nameoff,char *buff,unsigned cb_buff)
{
    unsigned char length;
    bioSeek(b_cache,nameoff,SEEK_SET);
    length = bioReadByte(b_cache);
    length = min(length,cb_buff);
    bioReadBuffer(b_cache,buff,length);
    buff[length] = 0;
}

static int __NEAR__ FindPubName(char *buff,unsigned cb_buff,unsigned long pa)
{
  return fmtFindPubName(nlm_cache,buff,cb_buff,pa,
                        nlm_ReadPubNameList,
                        nlm_ReadPubName);
}

static void nlm_ReadPubNameList(BGLOBAL handle,int interact)
{
 unsigned char length;
 unsigned i;
 unsigned nnames = (unsigned)nlm.nlm_numberOfPublics;
 if(!PubNames)
   if(!(PubNames = la_Build(0,sizeof(struct PubName),interact))) return;
 bioSeek(handle,nlm.nlm_publicsOffset,SEEKF_START);
 for(i = 0;i < nnames;i++)
 {
   struct PubName nlm_pn;
   nlm_pn.nameoff = bioTell(handle);
   length         = bioReadByte(handle);
   bioSeek(handle,length,SEEK_CUR);
   nlm_pn.pa      = (bioReadDWord(handle) & 0x00FFFFFFL) + nlm.nlm_codeImageOffset;
   if(!la_AddData(PubNames,&nlm_pn,interact)) break;
 }
 if(PubNames->nItems) la_Sort(PubNames,fmtComparePubNames);
}

unsigned long NLMGetPubSym(char *str,unsigned cb_str,unsigned *func_class,
                           unsigned long pa,int as_prev)
{
  return fmtGetPubSym(nlm_cache,str,cb_str,func_class,pa,as_prev,
                      nlm_ReadPubNameList,
                      nlm_ReadPubName);
}

unsigned NLMGetObjAttr(unsigned long pa,char *name,unsigned cb_name,
                      unsigned long *start,unsigned long *end,int *class,int *bitness)
{
  unsigned ret;
  *start = 0;
  *end = BMGetFLength();
  *class = OC_NOOBJECT;
  *bitness = NLMbitness(pa);
  name[0] = 0;
  if(pa < nlm.nlm_codeImageOffset)
  {
    *end = nlm.nlm_codeImageOffset;
    ret = 0;
  }
  else
    if(pa >= nlm.nlm_codeImageOffset && pa < nlm.nlm_codeImageOffset + nlm.nlm_codeImageSize)
    {
      *start = nlm.nlm_codeImageOffset;
      *class = OC_CODE;
      *end = nlm.nlm_codeImageOffset + nlm.nlm_codeImageSize;
      ret = 1;
    }
    else
    if(pa >= nlm.nlm_dataImageOffset && pa < nlm.nlm_dataImageOffset + nlm.nlm_dataImageSize)
    {
      *class = OC_DATA;
      *start = nlm.nlm_dataImageOffset;
      *end = nlm.nlm_dataImageOffset + nlm.nlm_dataImageSize;
      ret = 2;
    }
    else
    {
      *class = OC_NOOBJECT;
      *start = nlm.nlm_dataImageOffset + nlm.nlm_dataImageSize;
      ret = 3;
    }
  return ret;
}

unsigned long NLMVA2PA(unsigned long va)
{
  return va + min(nlm.nlm_codeImageOffset,nlm.nlm_dataImageOffset);
}

unsigned long NLMPA2VA(unsigned long pa)
{
  unsigned long base = min(nlm.nlm_codeImageOffset,nlm.nlm_dataImageOffset);
  return pa > base ? pa - base : 0L;
}

REGISTRY_BIN nlm386Table =
{
  "nlm-i386 (Novell Loadable Module)",
  { "NlmHlp", "ModRef", "PubDef", NULL, "ExtNam", NULL, NULL, "NlmHdr", NULL, NULL },
  { HelpNLM, ShowModRefNLM, ShowPubNamNLM, NULL, ShowExtRefNLM, NULL, NULL, ShowNewNLM, NULL, NULL },
  IsNLM, NLMinit, NLMdestroy,
  ShowNLMHeader,
  AppendNLMRef,
  fmtSetState,
  NLMbitness,
  NLMAddrResolv,
  NLMVA2PA,
  NLMPA2VA,
  NLMGetPubSym,
  NLMGetObjAttr,
  NULL,
  NULL
};
  