/****************************************************************************\
*   NLM386 - this header file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/nlm386.h                               *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain NLM file format definitions                             *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#ifndef __NLM_INC
#define __NLM_INC

#include "biewtype.h"

#ifdef __GNUC__
#pragma pack(1)
#endif

typedef tDWord file_ptr;
typedef tDWord bfd_size_type;
typedef tDWord bfd_vma;
typedef tDWord PTR;
#define NLM_SIGNATURE "NetWare Loadable Module\x1A"
#define NLM_SIGNATURE_SIZE 24
#define NLM_MODULE_NAME_SIZE 14
#define NLM_MAX_DESCRIPTION_LENGTH 41
#define NLM_OLD_THREAD_NAME_LENGTH 5
#define NLM_MAX_SCREEN_NAME_LENGTH 4
#define NLM_MAX_THREAD_NAME_LENGTH 7
#define NLM_MAX_COPYRIGHT_MESSAGE_LENGTH 1


typedef struct nlm_internal_fixed_header
{
  tChar         nlm_signature[NLM_SIGNATURE_SIZE];
  tLong         nlm_version;
  tChar         nlm_moduleName[NLM_MODULE_NAME_SIZE];
  file_ptr      nlm_codeImageOffset;
  bfd_size_type nlm_codeImageSize;
  file_ptr      nlm_dataImageOffset;
  bfd_size_type nlm_dataImageSize;
  bfd_size_type nlm_uninitializedDataSize;
  file_ptr      nlm_customDataOffset;
  bfd_size_type nlm_customDataSize;
  file_ptr      nlm_moduleDependencyOffset;
  tLong         nlm_numberOfModuleDependencies;
  file_ptr      nlm_relocationFixupOffset;
  tLong         nlm_numberOfRelocationFixups;
  file_ptr      nlm_externalReferencesOffset;
  tLong         nlm_numberOfExternalReferences;
  file_ptr      nlm_publicsOffset;
  tLong         nlm_numberOfPublics;
  file_ptr      nlm_debugInfoOffset;
  tLong         nlm_numberOfDebugRecords;
  file_ptr      nlm_codeStartOffset;
  file_ptr      nlm_exitProcedureOffset;
  file_ptr      nlm_checkUnloadProcedureOffset;
  tLong         nlm_moduleType;
  tLong         nlm_flags;
} Nlm_Internal_Fixed_Header;

typedef struct nlm_internal_variable_header
{
  tByte         descriptionLength;
  tChar         descriptionText[NLM_MAX_DESCRIPTION_LENGTH + 1];
  tLong         stackSize;
  tLong         reserved; /* should contain zero */
  tChar         oldThreadName[NLM_OLD_THREAD_NAME_LENGTH]; // " LONG"
  tByte         screenNameLength;
  tChar         screenName[NLM_MAX_SCREEN_NAME_LENGTH + 1];
  tByte         threadNameLength;
  tChar         threadName[NLM_MAX_THREAD_NAME_LENGTH + 1];
} Nlm_Internal_Variable_Header;

typedef struct nlm_internal_version_header
{
  /* The header is recognized by "VeRsIoN#" in the stamp field. */
  tChar         stamp[8];
  tLong         majorVersion;
  tLong         minorVersion;
  tLong         revision;
  tLong         year;
  tLong         month;
  tLong         day;
} Nlm_Internal_Version_Header;

typedef struct nlm_internal_copyright_header
{
  /* The header is recognized by "CoPyRiGhT=" in the stamp field. */
  tChar         stamp[10];
  tByte         copyrightMessageLength;
  tChar         copyrightMessage[NLM_MAX_COPYRIGHT_MESSAGE_LENGTH];
} Nlm_Internal_Copyright_Header;

typedef struct nlm_internal_extended_header
{
  /* The header is recognized by "MeSsAgEs" in the stamp field. */
  tChar         stamp[8];
  tLong         languageID;
  file_ptr      messageFileOffset;
  bfd_size_type messageFileLength;
  tLong         messageCount;
  file_ptr      helpFileOffset;
  bfd_size_type helpFileLength;
  file_ptr      RPCDataOffset;
  bfd_size_type RPCDataLength;
  file_ptr      sharedCodeOffset;
  bfd_size_type sharedCodeLength;
  file_ptr      sharedDataOffset;
  bfd_size_type sharedDataLength;
  file_ptr      sharedRelocationFixupOffset;
  tLong         sharedRelocationFixupCount;
  file_ptr      sharedExternalReferenceOffset;
  tLong         sharedExternalReferenceCount;
  file_ptr      sharedPublicsOffset;
  tLong         sharedPublicsCount;
  file_ptr      sharedDebugRecordOffset;
  tLong         sharedDebugRecordCount;
  bfd_vma       SharedInitializationOffset;
  bfd_vma       SharedExitProcedureOffset;
  tLong         productID;
  tLong         reserved[6];
} Nlm_Internal_Extended_Header;

/* The format of a custom header as stored internally is different
   from the external format.  This is how we store a custom header
   which we do not recognize.  */

typedef struct nlm_internal_custom_header
{
  /* The header is recognized by "CuStHeAd" in the stamp field. */
  tChar         stamp[8];
  bfd_size_type hdrLength;
  file_ptr      dataOffset;
  bfd_size_type dataLength;
  tChar         dataStamp[8];
  PTR           hdr;
} Nlm_Internal_Custom_Header;

/* The internal Cygnus header is written out externally as a custom
   header.  We don't try to replicate that structure here.  */

typedef struct nlm_internal_cygnus_ext_header
{
  /* The header is recognized by "CyGnUsEx" in the stamp field. */
  tChar         stamp[8];
  file_ptr      offset;  /* File location of debugging information.  */
  bfd_size_type length;   /* Length of debugging information.  */
} Nlm_Internal_Cygnus_Ext_Header;

/*
   Public names table:
   ===================
   +0 ( 1 byte )         - name length
   +1 ( length bytes )   - name
   +length+1 ( 4 bytes ) - physical offset from begin of code section
*/

/*
   External name table:
   ====================
   +0 ( 1 byte )         - name length
   +1 ( length bytes )   - name
   +length+1 ( 4 bytes ) - number of fixups
   +length+5 ( 4 bytes array) - physical offsets from begin of code section
*/


typedef tDWord	Nlm32_Addr;	/* Unsigned program address */
typedef tDWord	Nlm32_Off;	/* Unsigned file offset */
typedef tLong	Nlm32_Sword;	/* Signed large integer */
typedef tDWord	Nlm32_Word;	/* Unsigned large integer */
typedef tWord	Nlm32_Half;	/* Unsigned medium integer */
typedef tByte	Nlm32_Char;	/* Unsigned tiny integer */

typedef tLong		Nlm64_Sword;
typedef tDWord		Nlm64_Word;
typedef tWord		Nlm64_Half;

#ifdef __GNUC__
#pragma pack()
#endif


#endif 