/****************************************************************************\
*   PE - this source file is part of Binary vIEW project (BIEW)              *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/pe.c                                   *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of PE (Portable Executable) file format  *
* decoder.                                                                   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <limits.h>

#include "plugins/bin/pe.h"
#include "bin_util.h"
#include "codeguid.h"
#include "kbd_code.h"
#include "bmfile.h"
#include "biewhelp.h"
#include "tstrings.h"
#include "biewutil.h"
#include "bconsole.h"
#include "reg_form.h"
#include "__os_dep.h"
#include "_sys_dep.h"
#include "pmalloc.h"

static PE_ADDR * peVA;
PEHEADER pe;

static BGLOBAL pe_cache = &bNull;
static BGLOBAL pe_cache1 = &bNull;
static BGLOBAL pe_cache2 = &bNull;
static BGLOBAL pe_cache3 = &bNull;
static BGLOBAL pe_cache4 = &bNull;

static int __NEAR__ FindPubName(char *buff,unsigned cb_buff,unsigned long pa);
static void pe_ReadPubNameList(BGLOBAL handle,int interact);
unsigned long peVA2PA(unsigned long va);

int peLowMemFunc( unsigned long need_mem )
{
  if(!fmtActiveState)
  {
    if(PubNames)
    {
       la_Destroy(PubNames);
       PubNames = NULL;
       return 1;
    }
  }
  return 0;
}

static unsigned long __NEAR__ NormalizeLValue(long value,long align)
{
 unsigned long intp;
 intp = value / align;
 if(value % align) value = ( value / intp ) * intp;
 return value;
}

static unsigned long __NEAR__ RVA2Phys(unsigned long rva)
{
 int i;
 unsigned long ret,npages;
 unsigned long poff;
 for(i = pe.peObjects - 1;i >= 0;i--)
 {
   if(rva >= peVA[i].rva) break;
   if(IsKbdTerminate()) return 0;
 }
 // each page is 4096 bytes
 // each object can contain several pages
 // it now contain previous object that contain this page
 if(!peVA[i].phys) return 0;
 npages = (rva - peVA[i].rva) / 4096;
 poff = (rva - peVA[i].rva) % 4096;
 ret = peVA[i].phys + npages*4096 + poff;
 return ret;
}

static unsigned long __NEAR__ fioReadDWord(BGLOBAL handle,unsigned long offset,int origin)
{
 bioSeek(handle,offset,origin);
 return bioReadDWord(handle);
}

static unsigned __NEAR__ fioReadWord(BGLOBAL handle,unsigned long offset,int origin)
{
 bioSeek(handle,offset,origin);
 return bioReadWord(handle);
}

static unsigned long __NEAR__ fioReadDWord2Phys(BGLOBAL handle,unsigned long offset,int origin)
{
 unsigned long dword;
 dword = fioReadDWord(handle,offset,origin);
 return RVA2Phys(dword);
}

char * __pecputype[] =
{
 "80386",
 "80486",
 "80586",
 "80686",
 "80787",
 "80887",
 "80987",
 "801087"
};

static char * __NEAR__ PECPUType(void)
{
  char * cptr;
  if(pe.peCPUType >= 0x014C && pe.peCPUType < 0x0150) cptr = __pecputype[(pe.peCPUType - 0x014C) & 0x0007];
  else
    if(pe.peCPUType == 0x0162) cptr = "__MIPS Mark I (R2000, R3000)";
    else
      if(pe.peCPUType == 0x0163) cptr = "__MIPS Mark II (R6000)";
      else
        if(pe.peCPUType == 0x0166) cptr = "__MIPS Mark III (R4000)";
        else                            cptr = __pecputype[0];
  return cptr;
}

static unsigned long entryPE;

static void __NEAR__ PaintNewHeaderPE_1( void )
{
  twPrintF("Signature                      = '%c%c'\n"
           "Required CPU Type              = %s\n"
           "Number of object entries       = %hu\n"
           "Time/Data Stamp                = %s"
           "NT header size                 = %hu bytes\n"
           "Image flags :                    [%04hXH]\n"
           "    [%c] < Relocation info stripped >\n"
           "    [%c] Image is executable\n"
           "    [%c] < Line number stripped >\n"
           "    [%c] < Local symbols stripped >\n"
           "    [%c] < Minimal object >\n"
           "    [%c] < Update object >\n"
           "    [%c] < 16 bit word machine >\n"
           "    [%c] < 32 bit word machine >\n"
           "    [%c] Fixed\n"
           "    [%c] < System file >\n"
           "    [%c] Library image\n"
           "Linker version                 = %u.%02u\n"
           ,pe.peSignature[0],pe.peSignature[1]
           ,PECPUType()
           ,pe.peObjects
           ,ctime((time_t *)&pe.peTimeDataStamp)
           ,pe.peNTHdrSize
           ,pe.peFlags
           ,GetBool(pe.peFlags & 0x0001)
           ,GetBool(pe.peFlags & 0x0002)
           ,GetBool(pe.peFlags & 0x0004)
           ,GetBool(pe.peFlags & 0x0008)
           ,GetBool(pe.peFlags & 0x0010)
           ,GetBool(pe.peFlags & 0x0020)
           ,GetBool(pe.peFlags & 0x0040)
           ,GetBool(pe.peFlags & 0x0100)
           ,GetBool(pe.peFlags & 0x0200)
           ,GetBool(pe.peFlags & 0x1000)
           ,GetBool(pe.peFlags & 0x2000)
           ,(int)pe.peLMajor,(int)pe.peLMinor);
  twSetColor(White,Black);
  twPrintF("EntryPoint RVA    %s = %08lXH (Offset: %08lXH)",pe.peFlags & 0x2000 ? "[ LibEntry ]" : "[ EXEEntry ]",pe.peEntryPointRVA,entryPE); twClrEOL();
  twSetColor(Black,LightGray);
  twPrintF("\nImage base                     = %08lXH\n"
           "Object aligning                = %08lXH"
           ,pe.peImageBase
           ,pe.peObjectAlign);
}

char * __pesubsyst[] =
{
  "Unknown",
  "Native",
  "Windows GUI",
  "Windows Character",
  "OS/2 GUI",
  "OS/2 Character",
  "Posix GUI",
  "Posix Character"
};

static void __NEAR__ PaintNewHeaderPE_2( void )
{
  twPrintF("File align                     = %08lXH\n"
           "OS version                     = %hu.%hu\n"
           "User version                   = %hu.%hu\n"
           "Subsystem version              = %hu.%hu\n"
           "Image size                     = %lu bytes\n"
           "Header size                    = %lu bytes\n"
           "File checksum                  = %08lXH\n"
           "NT subsystem                   = %s\n"
           "DLL Flags :                      [%04hXH]\n"
           " [%c] Per-Process Library initialization\n"
           " [%c] Per-Process Library termination\n"
           " [%c] Per-Thread  Library initialization\n"
           " [%c] Per-Thread  Library termination\n"
           "Stack reserve size             = %lu bytes\n"
           "Stack commit size              = %lu bytes\n"
           "Heap reserve size              = %lu bytes\n"
           "Heap commit size               = %lu bytes\n"
           "Number interesting VA/Sizes    = %lu bytes\n"
           "Export Table RVA               = %08lXH\n"
           "Total export data size         = %lu bytes\n"
           "Import Table RVA               = %08lXH"
           ,pe.peFileAlign
           ,pe.peOSMajor,pe.peOSMinor
           ,pe.peUserMajor,pe.peUserMinor
           ,pe.peSubSystMajor,pe.peSubSystMinor
           ,pe.peImageSize
           ,pe.peHeaderSize
           ,pe.peFileChecksum
           ,pe.peSubSystem <= 0x0007 ? __pesubsyst[pe.peSubSystem] : "Unknown"
           ,pe.peDLLFlags
           ,GetBool(pe.peDLLFlags & 0x0001)
           ,GetBool(pe.peDLLFlags & 0x0002)
           ,GetBool(pe.peDLLFlags & 0x0004)
           ,GetBool(pe.peDLLFlags & 0x0008)
           ,pe.peStackReserveSize
           ,pe.peStackCommitSize
           ,pe.peHeapReserveSize
           ,pe.peHeapCommitSize
           ,pe.peInterestingVASize
           ,pe.peExportTableRVA
           ,pe.peTotalExportDataSize
           ,pe.peImportTableRVA);
}

static void __NEAR__ PaintNewHeaderPE_3( void )
{
  twPrintF("Total import data size            = %lu bytes\n"
           "Resource Table RVA                = %08lXH\n"
           "Total resource data size          = %lu bytes\n"
           "Exception Table RVA               = %08lXH\n"
           "Total exception data size         = %lu bytes\n"
           "Security Table RVA                = %08lXH\n"
           "Total security data size          = %lu bytes\n"
           "Fixup Table RVA                   = %08lXH\n"
           "Total fixup data size             = %lu bytes\n"
           "Debug Table RVA                   = %08lXH\n"
           "Total number of debug directories = %lu bytes\n"
           "Image Description string RVA      = %08lXH\n"
           "Total description data size       = %lu bytes\n"
           "Machine Specific value RVA        = %08lXH\n"
           "Machine specific value size       = %lu bytes\n"
           "Thread Local Storage RVA          = %08lXH\n"
           "Total Thread Local Storage size   = %lu bytes"
           ,pe.peTotalImportDataSize
           ,pe.peResourceTableRVA
           ,pe.peTotalResourceDataSize
           ,pe.peExceptionTableRVA
           ,pe.peTotalExceptionDataSize
           ,pe.peSecurityTableRVA
           ,pe.peTotalSecurityDataSize
           ,pe.peFixupTableRVA
           ,pe.peTotalFixupDataSize
           ,pe.peDebugTableRVA
           ,pe.peTotalDebugDirectories
           ,pe.peImageDescriptionRVA
           ,pe.peTotalDescriptionSize
           ,pe.peMachineSpecificRVA
           ,pe.peMachineSpecificSize
           ,pe.peThreadLocalStorageRVA
           ,pe.peTotalTLSSize);
}

static void (__NEAR__ * pephead[])( void ) =
{
 PaintNewHeaderPE_1,
 PaintNewHeaderPE_2,
 PaintNewHeaderPE_3
};

void PaintNewHeaderPE(TWindow * win,void * *ptr,unsigned npage,unsigned tpage)
{
  char text[80];
  twUseWin(win);
  twFreezeWin(win);
  twClearWin();
  sprintf(text," Portable Executable Header [%d/%d] ",npage + 1,tpage);
  twSetTitle(win,text,TMODE_CENTER,Black,LightGray);
  twSetFooter(win,PAGEBOX_SUB,TMODE_RIGHT,White,Black);
  if(npage < 3) (*(pephead[npage]))();
  twRefreshWin(win);
}

unsigned long ShowNewHeaderPE( void )
{
 long fpos;
 fpos = BMGetCurrFilePos();
 entryPE = RVA2Phys(pe.peEntryPointRVA);
 if(PageBox(70,21,NULL,3,PaintNewHeaderPE) != -1 && entryPE && entryPE < BMGetFLength()) fpos = entryPE;
 return fpos;
}

static void __NEAR__ objpaintPE(PE_OBJECT *nam)
{
 twPrintF("Virtual Size                   = %lX bytes\n"
          "RVA (relative virtual address) = %08lX\n"
          "Physical size                  = %08lX bytes\n"
          "Physical offset                = %08lX bytes\n"
          "<Pointer to relocations>       = %08lX\n"
          "<Pointer to line numbers>      = %08lX\n"
          "<Number of relocations>        = %hu\n"
          "<Number of line numbers>       = %hu\n"
          "FLAGS: %lX\n"
          "   [%c] Code object\n"
          "   [%c] Initialized data object\n"
          "   [%c] Uninitialized data object\n"
          "   [%c] Object must not be cashed\n"
          "   [%c] Object is not pageable\n"
          "   [%c] Object is shared\n"
          "   [%c] Executable object\n"
          "   [%c] Readable object\n"
          "   [%c] Writable object"
          ,nam->oVirtualSize
          ,nam->oRVA
          ,nam->oPhysicalSize
          ,nam->oPhysicalOffset
          ,nam->oRelocPtr
          ,nam->oLineNumbPtr
          ,nam->oNReloc
          ,nam->oNLineNumb
          ,nam->oFlags
          ,GetBool((nam->oFlags & 0x00000020L) == 0x00000020L)
          ,GetBool((nam->oFlags & 0x00000040L) == 0x00000040L)
          ,GetBool((nam->oFlags & 0x00000080L) == 0x00000080L)
          ,GetBool((nam->oFlags & 0x04000000L) == 0x04000000L)
          ,GetBool((nam->oFlags & 0x08000000L) == 0x08000000L)
          ,GetBool((nam->oFlags & 0x10000000L) == 0x10000000L)
          ,GetBool((nam->oFlags & 0x20000000L) == 0x20000000L)
          ,GetBool((nam->oFlags & 0x40000000L) == 0x40000000L)
          ,GetBool((nam->oFlags & 0x80000000L) == 0x80000000L));
}

void ObjPaintPE(TWindow * win,void * * names,unsigned start,unsigned nlist)
{
 char buffer[81];
 PE_OBJECT * * nam = (PE_OBJECT * *)names;
 twUseWin(win);
 twFreezeWin(win);
 twClearWin();
 sprintf(buffer," Object Table [ %u / %u ] ",start + 1,nlist);
 twSetTitle(win,buffer,TMODE_CENTER,Black,LightGray);
 twSetFooter(win,PAGEBOX_SUB,TMODE_RIGHT,White,Black);
 twGotoXY(1,1);
 twPrintF("Object Name                    = %8s\n",nam[start]->oName);
 objpaintPE(nam[start]);
 twRefreshWin(win);
}

static int __NEAR__ __ReadObjectsPE(BGLOBAL handle,memArray * obj,unsigned n)
{
 int i;
  for(i = 0;i < n;i++)
  {
    PE_OBJECT po;
    if(IsKbdTerminate()) break;
    bioReadBuffer(handle,&po,sizeof(PE_OBJECT));
    if(!ma_AddData(obj,&po,sizeof(PE_OBJECT),1)) break;
  }
  return 1;
}

static unsigned long __NEAR__ CalcPEObjectEntry(long offset)
{
 return NormalizeLValue(offset,pe.peFileAlign);
}

unsigned long ShowObjectsPE( void )
{
 BGLOBAL handle;
 unsigned nnames;
 unsigned long fpos;
 memArray * obj;
 fpos = BMGetCurrFilePos();
 nnames = pe.peObjects;
 if(!nnames) { NotifyBox(NOT_ENTRY," Objects Table "); return fpos; }
 if(!(obj = ma_Build(nnames,1))) return fpos;
 handle = pe_cache;
 bioSeek(handle,0x18 + pe.peNTHdrSize + headshift,SEEK_SET);
 if(__ReadObjectsPE(handle,obj,nnames))
 {
  int ret;
    ret = PageBox(70,19,(void **)obj->data,obj->nItems,ObjPaintPE);
    if(ret != -1)  fpos = CalcPEObjectEntry(((PE_OBJECT *)obj->data[ret])->oPhysicalOffset);

 }
 ma_Destroy(obj);
 return fpos;
}

static unsigned __NEAR__ GetImportCountPE(BGLOBAL handle,unsigned long phys)
{
  unsigned count;
  unsigned long fpos = bioTell(handle);
  unsigned long ctrl;
  count = 0;
  bioSeek(handle,phys,SEEKF_START);
  while(1)
  {
    ctrl = fioReadDWord(handle,12L,SEEKF_CUR);
    bioSeek(handle,4L,SEEKF_CUR);
    if(ctrl == 0 || count > 0xFFFD || IsKbdTerminate()) break;
    count++;
  }
  bioSeek(handle,fpos,SEEKF_START);
  return count;
}

static char * __NEAR__ __ReadASCIIZ(BGLOBAL handle,unsigned long offset,int align)
{
  int j,size,key,shift;
  char ch;
  char * asciiz;
  unsigned long fpos;
  fpos = bioTell(handle);
  j = 0;
  shift = 0;
  if(align == 2) shift = 2;
  bioSeek(handle,offset + shift,SEEKF_START);
  while(1)
  {
    j++;
    if(j > 300) break;
    ch = bioReadByte(handle);
    if(ch == 0) break;
  }
  size = j;
  key = 0;
  if(align == 1)
  {
    if(size > 66) { j = 66; key = 1; }
  }
  asciiz = PMalloc(size > 66 ? 73 : size + 7);
  if( asciiz != 0 )
  {
    bioSeek(handle,offset + shift,SEEKF_START);
    j--;
    bioReadBuffer(handle,asciiz,j);
    if(key) { asciiz[j++] = '\x10'; asciiz[j++] = '\x10'; asciiz[j++] = '\x10'; }
    bioSeek(handle,fpos,SEEKF_START);
    if(align == 1) asciiz[j++] = '';
    asciiz[j] = 0;
  }
  return asciiz;
}

static char * __NEAR__ __ReadASCIIZFull(BGLOBAL handle,unsigned long offset,int align)
{
  int j,size,shift;
  char ch;
  char * asciiz;
  unsigned long fpos;
  fpos = bioTell(handle);
  j = 0;
  shift = 0;
  if(align == 2) shift = 2;
  bioSeek(handle,offset + shift,SEEKF_START);
  while(1)
  {
    j++;
    if(j > 300) break;
    ch = bioReadByte(handle);
    if(ch == 0) break;
  }
  size = j;
  if(align == 1)
  {
    if(size > 256) j = 256;
    size = 80;
  }
  asciiz = PMalloc(size);
  if( asciiz != 0 )
  {
    bioSeek(handle,offset + shift,SEEKF_START);
    j--;
    bioReadBuffer(handle,asciiz,j);
    bioSeek(handle,fpos,SEEKF_START);
    asciiz[j] = 0;
  }
  return asciiz;
}

static unsigned __NEAR__ __ReadImportPE(BGLOBAL handle,unsigned long phys,memArray *obj,unsigned nnames)
{
  unsigned i;
  unsigned long fpos = bioTell(handle);
  unsigned long rva,addr;
  bioSeek(handle,phys,SEEKF_START);
  for(i = 0;i < nnames;i++)
  {
    char *tmp;
    rva = fioReadDWord(handle,12L,SEEKF_CUR);
    bioSeek(handle,4L,SEEKF_CUR);
    addr = RVA2Phys(rva);
    if(!(tmp = __ReadASCIIZ(handle,addr,0))) break;
    if(IsKbdTerminate()) break;
    if(!ma_AddString(obj,tmp,1)) { free(tmp); break; }
    free(tmp);
  }
  bioSeek(handle,fpos,SEEKF_START);
  return obj->nItems;
}

static unsigned long addr_shift_pe;

unsigned GetImpCountPE(BGLOBAL handle)
{
 unsigned count;
 unsigned long Hint;
 count = 0;
 if(addr_shift_pe)
 {
   bioSeek(handle,addr_shift_pe,SEEKF_START);
   while(1)
   {
     Hint = bioReadDWord(handle);
     if(Hint == 0L || count > 0xFFFD || IsKbdTerminate()) break;
     count++;
   }
 }
 return count;
}

int  __ReadImpContPE(BGLOBAL handle,memArray * obj,unsigned nnames)
{
  int i;
  unsigned long Hint,rphys;
  bioSeek(handle,addr_shift_pe,SEEKF_START);
  for(i = 0;i < nnames;i++)
  {
    Hint = bioReadDWord(handle);
    if(!(Hint & 0x80000000L))
    {
       rphys = RVA2Phys(Hint & 0x7FFFFFFFL);
       if(rphys > BMGetFLength())
       {
          if(!ma_AddString(obj,"*** !!! ??? The .PE is incorrect ??? !!! ***",1)) break;
       }
       else
       {
         char *tmp;
         if(!(tmp = __ReadASCIIZ(handle,rphys,2))) break;
         if(!ma_AddString(obj,tmp,1)) { free(tmp); break; }
         free(tmp);
         if(IsKbdTerminate()) break;
       }
    }
    else
    {
      char stmp[80];
      if(IsKbdTerminate()) break;
      sprintf(stmp,"< By ordinal >   @%lu",Hint & 0x7FFFFFFFL);
      if(!ma_AddString(obj,stmp,1)) break;
    }
  }
  return 1;
}

unsigned long ShowModRefPE( void )
{
  BGLOBAL handle;
  char petitle[80];
  memArray * obj;
  unsigned nnames;
  unsigned long phys,fret;
  fret = BMGetCurrFilePos();
  if(!pe.peImportTableRVA) { NotifyBox(NOT_ENTRY," Module References "); return fret; }
  handle = pe_cache;
  bioSeek(handle,0L,SEEK_SET);
  phys = RVA2Phys(pe.peImportTableRVA);
  nnames = GetImportCountPE(handle,phys);
  if(!(obj = ma_Build(nnames,1))) goto exit;
  if(__ReadImportPE(handle,phys,obj,nnames))
  {
    int i;
    i = 0;
    while(1)
    {
     ImportDirPE imp_pe;
     unsigned long magic;

     i = ma_Display(obj,MOD_REFER,LB_SELECTIVE,i);
     if(i == -1) break;
     strcpy(petitle,IMPPROC_TABLE);
     strcat(petitle,obj->data[i]);
     strcat(petitle," ");
     bioSeek(handle,phys + i*sizeof(ImportDirPE),SEEKF_START);
     bioReadBuffer(handle,&imp_pe,sizeof(ImportDirPE));
     if(!(imp_pe.idMajVer == 0 && imp_pe.idMinVer == 0 && imp_pe.idDateTime != 0xFFFFFFFFL))
                                   magic = imp_pe.idFlags;
     else                          magic = imp_pe.idLookupTableRVA;
     addr_shift_pe = magic ? RVA2Phys(magic) : magic;
     fmtShowList(GetImpCountPE,__ReadImpContPE,petitle,0,NULL);
    }
  }
  ma_Destroy(obj);
  exit:
  return fret;
}

static ExportTablePE et;

int PEExportReadItems(BGLOBAL handle,memArray * obj,unsigned nnames)
{
  unsigned long nameaddr;
  unsigned long expaddr,nameptr;
  unsigned i,ord;
  char buff[10];
  nameptr = RVA2Phys(et.etNamePtrTableRVA);
  expaddr  = RVA2Phys(et.etOrdinalTableRVA);
  for(i = 0;i < nnames;i++)
  {
    char *stmp;
    nameaddr = fioReadDWord2Phys(handle,nameptr + 4*i,SEEKF_START);
    if(!(stmp = __ReadASCIIZ(handle,nameaddr,1)) || IsKbdTerminate()) break;
    ord = fioReadWord(handle,expaddr + i*2,SEEKF_START) + (unsigned)et.etOrdinalBase;
    sprintf(buff,"%i",ord);
    strcat(stmp,buff);
    if(!ma_AddString(obj,stmp,1)) { free(stmp); break; }
    free(stmp);
  }
  return 1;
}

unsigned PEExportReadFullName(BGLOBAL handle,char * names,unsigned index)
{
  unsigned long nameaddr;
  unsigned long expaddr,nameptr;
  unsigned ord;
  char * name;
  nameptr = RVA2Phys(et.etNamePtrTableRVA);
  expaddr  = RVA2Phys(et.etOrdinalTableRVA);
  nameaddr = fioReadDWord2Phys(handle,nameptr + 4*index,SEEKF_START);
  name = __ReadASCIIZFull(handle,nameaddr,1);
  if(name)
  {
    strncpy(names,name,255);
    free(name);
  }
  ord = fioReadWord(handle,expaddr + index*2,SEEKF_START);
  return ord;
}

unsigned PEExportNumItems(BGLOBAL handle)
{
  unsigned long addr;
  if(!pe.peExportTableRVA) return 0;
  addr = RVA2Phys(pe.peExportTableRVA);
  bioSeek(handle,addr,SEEKF_START);
  bioReadBuffer(handle,(void *)&et,sizeof(et));
  return (unsigned)(et.etNumNamePtrs);
}

static unsigned long __NEAR__ CalcEntryPE(unsigned ordinal,int dispmsg)
{
 unsigned long fret,rva;
 unsigned ord;
 BGLOBAL handle;
 fret = BMGetCurrFilePos();
 handle = pe_cache1;
 {
   unsigned long eret;
   rva = RVA2Phys(et.etAddressTableRVA);
   ord = (unsigned)ordinal - (unsigned)et.etOrdinalBase;
   eret = fioReadDWord2Phys(handle,rva + 4*ord,SEEKF_START);
   if(eret && eret < BMGetFLength()) fret = eret;
   else if(dispmsg) ErrMessageBox("Entry do not refers to physical page of this file","Bad entry");
 }
 return fret;
}

unsigned long ShowExpNamPE( void )
{
  unsigned long fpos = BMGetCurrFilePos();
  int ret;
  unsigned ordinal;
  ret = fmtShowList(PEExportNumItems,PEExportReadItems,
                    EXP_TABLE,
                    LB_SELECTIVE | LB_SORTABLE,&ordinal);
  if(ret != -1)
  {
    fpos = CalcEntryPE(ordinal,1);
  }
  return fpos;
}

/***************************************************************************/
/************************  FOR PE  *****************************************/
/***************************************************************************/
static char __codelen;

typedef struct tagRELOC_PE
{
  unsigned long modidx;
  union
  {
    unsigned long funcidx; // if modidx != -1
    unsigned long type;    // if modidx == -1
  }import;
  unsigned long laddr; //lookup addr
  unsigned long reserved;
}RELOC_PE;

static linearArray *CurrPEChain = NULL;

extern unsigned long RVA2Phys(unsigned long);
extern unsigned      GetImportCountPE(BGLOBAL handle,unsigned long phys);
extern unsigned long fioReadDWord(BGLOBAL handle,unsigned long offset,int origin);

static int compare_pe_reloc_s(const void __HUGE__ *e1,const void __HUGE__ *e2)
{
  const RELOC_PE __HUGE__ *p1,__HUGE__ *p2;
  p1 = e1;
  p2 = e2;
  if(p1->laddr < p2->laddr) return -1;
  else
    if(p1->laddr > p2->laddr) return 1;
    else                      return 0;
}

static int compare_pe_reloc_f(const void __HUGE__ *e1,const void __HUGE__ *e2)
{
  const RELOC_PE __HUGE__ *p1,__HUGE__ *p2;
  p1 = e1;
  p2 = e2;
  if(p2->laddr >= p1->laddr && p2->laddr < p1->laddr + __codelen) return 0;
  else                                                            return 1;
}

static void __NEAR__ BuildPERefChain( void )
{
  unsigned long phys = 0,i,j,cpos;
  RELOC_PE rel;
  ImportDirPE ipe;
  unsigned nnames;
  BGLOBAL handle;
  TWindow *w,*usd;
  usd = twUsedWin();
  if(!(CurrPEChain = la_Build(0,sizeof(RELOC_PE),1))) return;
  w = CrtHlpWnd(" System is busy ",49,1);
  if(!PubNames) pe_ReadPubNameList(BMbioHandle(),1);
  twUseWin(w);
  twGotoXY(1,1);
  twPutS(" Building reference chains for PE format");
  twUseWin(usd);
  handle = pe_cache;
  bioSeek(handle,0L,SEEK_SET);
  /*
     building references chain for external
  */
  if(pe.peImportTableRVA)
  {
    phys = RVA2Phys(pe.peImportTableRVA);
    nnames = GetImportCountPE(handle,phys);
  }
  else nnames = 0;
  for(i = 0;i < nnames;i++)
  {
   unsigned long magic,addr_shift_pe,Hint;
   bioSeek(handle,phys + i*sizeof(ImportDirPE),SEEKF_START);
   bioReadBuffer(handle,&ipe,sizeof(ImportDirPE));
   if(!(ipe.idMajVer == 0 && ipe.idMinVer == 0 && ipe.idDateTime != 0xFFFFFFFFL))
                                 magic = ipe.idLookupTableRVA;
    else                         magic = ipe.idFlags;
    addr_shift_pe = magic ? RVA2Phys(magic) : magic;
    if(addr_shift_pe)
    {
      bioSeek(handle,addr_shift_pe,SEEKF_START);
      j = 0;
      while(1)
      {
        Hint = bioReadDWord(handle);
        if(!Hint || IsKbdTerminate()) break;
        rel.modidx = i;
        rel.laddr = Hint;
        rel.import.funcidx = j;
        if(!la_AddData(CurrPEChain,&rel,1)) goto bye;
        j++;
      }
    }
  }
  /*
     building references chain for internal
  */
  if(pe.peTotalFixupDataSize)
  {
    phys = RVA2Phys(pe.peFixupTableRVA);
    bioSeek(handle,phys,SEEKF_START);
    cpos = bioTell(handle);
    while(bioTell(handle) < cpos + pe.peTotalFixupDataSize)
    {
      unsigned short typeoff;
      unsigned long page,physoff,size,ccpos;
      ccpos = bioTell(handle);
      page = bioReadDWord(handle);
      physoff = RVA2Phys(page);
      size = bioReadDWord(handle);
      while(bioTell(handle) < ccpos + size)
      {
        typeoff = bioReadWord(handle);
        if(IsKbdTerminate()) break;
        rel.modidx = -1;
        rel.import.type = typeoff;
        rel.import.type >>= 12;
        rel.laddr = physoff + (typeoff & 0x0FFF);
        if(!la_AddData(CurrPEChain,&rel,1)) goto bye;
      }
    }
  }
  bye:
  la_Sort(CurrPEChain,compare_pe_reloc_s);
  CloseWnd(w);
}

static RELOC_PE * __NEAR__ __found_RPE(unsigned long laddr,char codelen)
{
  RELOC_PE key;
  key.laddr = laddr;
  __codelen = codelen;
  if(!CurrPEChain) BuildPERefChain();
  return la_Find(CurrPEChain,&key,compare_pe_reloc_s);
}

static int __NEAR__ BuildReferStrPE(char *str,RELOC_PE *rpe,int flags)
{
   BGLOBAL handle,handle2,handle3;
   unsigned long phys,rva,magic,Hint;
   ImportDirPE ipe;
   char buff[400];
   handle = pe_cache;
   handle2 = pe_cache4;
   handle3 = pe_cache3;
   phys = RVA2Phys(pe.peImportTableRVA);
   bioSeek(handle,phys + 20L*rpe->modidx,SEEKF_START);
   rva = fioReadDWord(handle,12L,SEEKF_CUR);
   if(rpe->modidx != -1)
   {
     if(flags & APREF_USE_TYPE) strcat(str," off32");
     strcat(str, " <");
     bioSeek(handle2,RVA2Phys(rva),SEEKF_START);
     bioReadBuffer(handle2,buff,400);
     buff[399] = 0;
     strcat(str,buff);
     strcat(str,">.");
     bioSeek(handle3,phys + rpe->modidx*sizeof(ImportDirPE),SEEKF_START);
     bioReadBuffer(handle3,&ipe,sizeof(ImportDirPE));
     if(!(ipe.idMajVer == 0 && ipe.idMinVer == 0 && ipe.idDateTime != 0xFFFFFFFFL))
                                  magic = ipe.idFlags;
     else                         magic = ipe.idLookupTableRVA;
     magic = magic ? RVA2Phys(magic) : magic;
     if(magic)
     {
       bioSeek(handle,magic + rpe->import.funcidx*sizeof(long),SEEKF_START);
       Hint = bioReadDWord(handle);
       if(Hint & 0x80000000L)
       {
         char dig[15];
         sprintf(dig,"@%lu",Hint & 0x7FFFFFFFL);
         strcat(str,dig);
       }
       else
       {
         phys = RVA2Phys(Hint & 0x7FFFFFFFL);
         if(phys > BMGetFLength()) strcat(str,"???");
         else
         {
           bioSeek(handle2,phys + 2,SEEKF_START);
           bioReadBuffer(handle2,buff,400);
           buff[399] = 0;
           strcat(str,buff);
         }
       }
     }
   }
   else // internal refs
   {
     unsigned long delta,value,point_to;
     char *pe_how;
       strcat(str,"(*this).");
       bioSeek(handle3,rpe->laddr,SEEKF_START);
       value = bioReadDWord(handle3);
       delta = pe.peImageBase;
       point_to = 0;
       switch(rpe->import.type)
       {
          default:
          case 0: /* ABSOLUTE, fixup is skipped */
                  pe_how = "";
                  point_to = value;
                  break;
          case 1: /* HIGH 16-bit */
                  pe_how = "high16 < ";
                  break;
          case 2: /* LOW 16-bit */
                  pe_how = "low16 < ";
                  break;
          case 3: /* HIGHLOW */
                  point_to = peVA2PA(value);
                  pe_how = "off32 < ";
                  break;
          case 4: /* HIGHADJUST */
                  bioSeek(handle,value,SEEK_SET);
                  value = bioReadDWord(handle);
                  point_to = peVA2PA(value);
                  pe_how = "full32 < ";
                  break;
          case 5: /* MIPS JUMP ADDR */
                  pe_how = "mips < ";
                  break;
       }
     if(flags & APREF_USE_TYPE)  strcat(str,pe_how);
     strcat(str,Get8Digit(point_to ? point_to : value-delta)); /* if out of physical image */
     if(flags & APREF_USE_TYPE) strcat(str," >");
   }
   return 1;
}

static int AppendPERef(char *str,unsigned long ulShift,int flags,int codelen,unsigned long r_sh)
{
  RELOC_PE *rpe;
  unsigned long laddr;
  int ret;
  BGLOBAL b_cache;
  char buff[400];
  b_cache = pe_cache3;
  if(!pe.peImportTableRVA && !pe.peFixupTableRVA) return 0;
  laddr = BMReadDWordEx(ulShift,SEEK_SET);
  laddr -= pe.peImageBase;
  bioSeek(b_cache,RVA2Phys(laddr),SEEK_SET);
  laddr = bioReadDWord(b_cache);
  rpe = __found_RPE(laddr,codelen);
  ret = 0;
  if(!rpe ) rpe = __found_RPE(ulShift,codelen);
  if(rpe) ret = BuildReferStrPE(str,(void *)rpe,flags);
  if(!ret && (flags & APREF_TRY_LABEL))
  {
     if(!PubNames) pe_ReadPubNameList(BMbioHandle(),1);
     if(FindPubName(buff,sizeof(buff),r_sh))
     {
       strcat(str,buff);
       if(!DumpMode && !EditMode) GidAddGoAddress(str,r_sh);
       ret = 1;
     }
  }
  return ret;
}

int IsPE( void )
{
   char id[2];
   headshift = IsNewExe();
   if(headshift)
   {
     BMReadBufferEx(id,sizeof(id),headshift,SEEKF_START);
     if(id[0] == 'P' && id[1] == 'E') return 1;
   }
   return 0;
}

void initPE( void )
{
   int i;
   BGLOBAL main_handle;
   unsigned long cfpos;
   PMRegLowMemCallBack(peLowMemFunc);
   BMReadBufferEx(&pe,sizeof(PEHEADER),headshift,SEEKF_START);
   if(!(peVA = PMalloc(sizeof(PE_ADDR)*pe.peObjects)))
   {
     MemOutBox("PE initialization");
     exit(EXIT_FAILURE);
   }
   cfpos = BMGetCurrFilePos();
   main_handle = BMbioHandle();
   bioSeek(main_handle,0x18 + pe.peNTHdrSize + headshift,SEEKF_START);
   for(i = 0;i < pe.peObjects;i++)
   {
     peVA[i].rva = fioReadDWord(main_handle,12L,SEEKF_CUR);
     peVA[i].phys = fioReadDWord(main_handle,4L,SEEKF_CUR);
     bioSeek(main_handle,16L,SEEKF_CUR);
   }
   BMSeek(cfpos,SEEK_SET);
   if((pe_cache = bioDup(main_handle)) == &bNull) pe_cache = main_handle;
   if((pe_cache1 = bioDup(main_handle)) == &bNull) pe_cache1 = main_handle;
   if((pe_cache2 = bioDup(main_handle)) == &bNull) pe_cache2 = main_handle;
   if((pe_cache3 = bioDup(main_handle)) == &bNull) pe_cache3 = main_handle;
   if((pe_cache4 = bioDup(main_handle)) == &bNull) pe_cache4 = main_handle;
}

void destroyPE( void )
{
  BGLOBAL main_handle;
  if(peVA) free(peVA);
  if(PubNames) { la_Destroy(PubNames); PubNames = 0; }
  main_handle = BMbioHandle();
  if(pe_cache != &bNull && pe_cache != main_handle) bioClose(pe_cache);
  if(pe_cache1 != &bNull && pe_cache1 != main_handle) bioClose(pe_cache1);
  if(pe_cache2 != &bNull && pe_cache2 != main_handle) bioClose(pe_cache2);
  if(pe_cache3 != &bNull && pe_cache3 != main_handle) bioClose(pe_cache3);
  if(pe_cache4 != &bNull && pe_cache4 != main_handle) bioClose(pe_cache4);
  PMUnregLowMemCallBack(peLowMemFunc);
}

int bitnessPE(unsigned long off)
{
   if(off >= headshift)
   {
     return (pe.peFlags & 0x0040) ? DAB_USE16 : DAB_USE32;
   }
   else return DAB_USE16;
}

unsigned long PEHelp( void )
{
  hlpDisplay(10009);
  return BMGetCurrFilePos();
}

void peAddressResolv(char *addr,unsigned long cfpos)
{
 if(cfpos >= headshift && cfpos < headshift + sizeof(PEHEADER))
 {
    sprintf(addr,"PEH :%s",Get4Digit(cfpos - headshift));
 }
 else
 if(cfpos >= headshift + pe.peNTHdrSize + 0x18 &&
    cfpos <  headshift + pe.peNTHdrSize + 0x18 + pe.peObjects*sizeof(PE_OBJECT))
 {
    sprintf(addr,"PEOD:%s",Get4Digit(cfpos - headshift - pe.peNTHdrSize - 0x18));
 }
}

unsigned long peVA2PA(unsigned long va)
{
  return va >= pe.peImageBase ? RVA2Phys(va-pe.peImageBase) : 0L;
}

unsigned long pePA2VA(unsigned long pa)
{
 int i;
  unsigned long cfp,ret_addr;
  cfp = BMGetCurrFilePos();
  BMSeek(0x18 + pe.peNTHdrSize + headshift,SEEK_SET);
  ret_addr = 0;
  for(i = 0;i < pe.peObjects;i++)
  {
    PE_OBJECT po;
    unsigned long obj_pa;
    if(IsKbdTerminate()) break;
    BMReadBuffer(&po,sizeof(PE_OBJECT));
    obj_pa = CalcPEObjectEntry(po.oPhysicalOffset);
    if(pa >= obj_pa && pa < obj_pa + po.oPhysicalSize)
    {
      ret_addr = po.oRVA + (pa - obj_pa) + pe.peImageBase;
      break;
    }
  }
  BMSeek(cfp,SEEK_SET);
  return ret_addr;
}

static void pe_ReadPubName(BGLOBAL b_cache,unsigned long nameoff,char *buff,unsigned cb_buff)
{
    char *stmp;
    stmp = __ReadASCIIZFull(b_cache,nameoff,0);
    if(stmp)
    {
      strncpy(buff,stmp,cb_buff);
      buff[cb_buff-1] = 0;
      free(stmp);
    }
    else buff[0] = 0;
}

static int __NEAR__ FindPubName(char *buff,unsigned cb_buff,unsigned long pa)
{
  return fmtFindPubName(pe_cache4,buff,cb_buff,pa,
                        pe_ReadPubNameList,
                        pe_ReadPubName);
}

static void pe_ReadPubNameList(BGLOBAL handle,int interact)
{
  unsigned long i,nitems,expaddr,nameptr,nameaddr,entry_pa;
  unsigned ord;
  struct PubName pn;
  BGLOBAL b_cache;
  b_cache = pe_cache4 == &bNull ? handle : pe_cache4;
  nitems = PEExportNumItems(handle);
  if(!(PubNames = la_Build(0,sizeof(struct PubName),interact))) return;
  expaddr  = RVA2Phys(et.etOrdinalTableRVA);
  nameptr = RVA2Phys(et.etNamePtrTableRVA);
  for(i = 0;i < nitems;i++)
  {
    nameaddr = fioReadDWord2Phys(handle,nameptr + 4*i,SEEKF_START);
    ord = fioReadWord(b_cache,expaddr + i*2,SEEKF_START) + (unsigned)et.etOrdinalBase;
    entry_pa = CalcEntryPE((unsigned)ord,0);
    pn.pa = entry_pa;
    pn.nameoff = nameaddr;
    if(!la_AddData(PubNames,&pn,interact)) break;
  }
  la_Sort(PubNames,fmtComparePubNames);
}

unsigned long peGetPubSym(char *str,unsigned cb_str,unsigned *func_class,
                          unsigned long pa,int as_prev)
{
  return fmtGetPubSym(pe_cache,str,cb_str,func_class,pa,as_prev,
                      pe_ReadPubNameList,
                      pe_ReadPubName);
}

unsigned peGetObjAttr(unsigned long pa,char *name,unsigned cb_name,
                      unsigned long *start,unsigned long *end,int *class,int *bitness)
{
  unsigned i,nitems,ret;
  unsigned long cfpos;
  *start = 0;
  *end = BMGetFLength();
  *class = OC_NOOBJECT;
  *bitness = bitnessPE(pa);
  name[0] = 0;
  nitems = pe.peObjects;
  ret = 0;
  cfpos = BMGetCurrFilePos();
  BMSeek(0x18 + pe.peNTHdrSize + headshift,SEEK_SET);
  for(i = 0;i < nitems;i++)
  {
    PE_OBJECT po;
    if(IsKbdTerminate()) break;
    BMReadBuffer(&po,sizeof(PE_OBJECT));
    if(pa >= *start && pa < po.oPhysicalOffset)
    {
      /* means between two objects */
      *end = po.oPhysicalOffset;
      ret = 0;
      break;
    }
    if(pa >= po.oPhysicalOffset && pa < po.oPhysicalOffset + po.oPhysicalSize)
    {
      *start = po.oPhysicalOffset;
      *end = po.oPhysicalOffset + po.oPhysicalSize;
      *class = po.oFlags & 0x00000020L ? OC_CODE : OC_DATA;
      strncpy(name,po.oName,cb_name);
      ret = i+1;
      break;
    }
    *start = po.oPhysicalOffset + po.oPhysicalSize;
  }
  BMSeek(cfpos,SEEK_SET);
  return ret;
}

REGISTRY_BIN peTable =
{
  "PE (Portable Executable)",
  { "PEHelp", "Import", "Export", NULL, NULL, NULL, NULL, "PEHead", NULL, "Object" },
  { PEHelp, ShowModRefPE, ShowExpNamPE, NULL, NULL, NULL, NULL, ShowNewHeaderPE, NULL, ShowObjectsPE },
  IsPE, initPE, destroyPE,
  NULL,
  AppendPERef,
  fmtSetState,
  bitnessPE,
  peAddressResolv,
  peVA2PA,
  pePA2VA,
  peGetPubSym,
  peGetObjAttr,
  NULL,
  NULL
};
  