/****************************************************************************\
*   PHARLAP - this source file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/pharlap.c                              *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of PharLap file format decoder.          *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "__os_dep.h"
#include "bconsole.h"
#include "bin_util.h"
#include "biewutil.h"
#include "biewhelp.h"
#include "tstrings.h"
#include "kbd_code.h"
#include "bmfile.h"
#include "reg_form.h"
#include "plugins/bin/pharlap.h"

static newPharLap nph;

static BGLOBAL pl_cache = &bNull;

unsigned long ShowPharLapHeader( void )
{
  unsigned long fpos;
  TWindow *w;
  unsigned keycode;
  char sign[3];
  fpos = BMGetCurrFilePos();
  strncpy(sign,(char *)nph.plSignature,2);
  sign[2] = 0;
  w = CrtHlpWnd(" New PharLap executable ",59,23);
  twGotoXY(1,1);
  twPrintF("Signature                        = %s\n"
           "Level                            = %s\n"
           "Header size                      = %04XH\n"
           "File size                        = %08lXH\n"
           "Check sum (32-bits checksum)     = %04XH (%08lXH)\n"
           "Run-time parameters (offset/size)= %08lXH/%08lXH\n"
           "Relocations (offset/size)        = %08lXH/%08lXH\n"
           "Segment info table (offset/size) = %08lXH/%08lXH\n"
           "Image (offset/size)              = %08lXH/%08lXH\n"
           "Symbol table (offset/size)       = %08lXH/%08lXH\n"
           "GDT (offset/size)                = %08lXH/%08lXH\n"
           "LDT (offset/size)                = %08lXH/%08lXH\n"
           "IDT (offset/size)                = %08lXH/%08lXH\n"
           "TSS (offset/size)                = %08lXH/%08lXH\n"
           "Min. number of extra 4K pages    = %08lXH\n"
           "Max. number of extra 4K pages    = %08lXH\n"
           "Image base (flat level only)     = %08lXH\n"
           "Initial stack (SS:ESP)           = %04XH:%08lXH\n"
           "Initail code  (CS:EIP)           = %04XH:%08lXH\n"
           "Initial LDT/TSS                  = %04XH/%04XH\n"
           "Flags                            = %04XH\n"
           "Memory requirement for image     = %08lXH\n"
           "Stack size                       = %08lXH"
           ,sign
           ,nph.plLevel == 0x01 ? "Flat" : nph.plLevel == 0x02 ? "Multisegmented" : "Unknown"
           ,nph.plHeaderSize
           ,nph.plFileSize
           ,nph.plCheckSum,nph.plChecksum32
           ,nph.plRunTimeParms,nph.plRunTimeSize
           ,nph.plRelocOffset,nph.plRelocSize
           ,nph.plSegInfoOffset,nph.plSegInfoSize
           ,nph.plImageOffset,nph.plImageSize
           ,nph.plSymTabOffset,nph.plSymTabSize
           ,nph.plGDTOffset,nph.plGDTSize
           ,nph.plLDTOffset,nph.plLDTSize
           ,nph.plIDTOffset,nph.plIDTSize
           ,nph.plTSSOffset,nph.plTSSSize
           ,nph.plMinExtraPages
           ,nph.plMaxExtraPages
           ,nph.plBase
           ,nph.plSS,nph.plESP
           ,nph.plCS,nph.plEIP
           ,nph.plLDT,nph.plTSS
           ,nph.plFlags
           ,nph.plMemReq
           ,nph.plStackSize);
  while(1)
  {
    keycode = GetEvent(drawEmptyPrompt,w);
    if(keycode == KE_ENTER) { /* fpos = entrypoint*/; break; }
    else
      if(keycode == KE_ESCAPE) break;
  }
  CloseWnd(w);
  return fpos;
}

void PLSegPaint(TWindow * win,void * * names,unsigned start,unsigned nlist)
{
 char buffer[81];
 PLSegInfo * * nam = (PLSegInfo * *)names;
 twUseWin(win);
 twFreezeWin(win);
 twClearWin();
 sprintf(buffer," Segment Table [ %u / %u ] ",start + 1,nlist);
 twSetTitle(win,buffer,TMODE_CENTER,Black,LightGray);
 twSetFooter(win,PAGEBOX_SUB,TMODE_RIGHT,White,Black);
 twGotoXY(1,1);
 twPrintF("Selector number            = %04hXH\n"
          "Flags                      = %04hXH\n"
          "Base offset of selector    = %08lXH\n"
          "Min extra memory alloc     = %08lXH"
          ,nam[start]->siSelector
          ,nam[start]->siFlags
          ,nam[start]->siBaseOff
          ,nam[start]->siMinAlloc);
 twRefreshWin(win);
}


int __PLReadSegInfo(BGLOBAL handle,memArray * obj,unsigned nnames)
{
 unsigned i;
 for(i = 0;i < nnames;i++)
 {
   PLSegInfo plsi;
   if(IsKbdTerminate()) break;
   bioReadBuffer(handle,&plsi,sizeof(PLSegInfo));
   if(!ma_AddData(obj,&plsi,sizeof(PLSegInfo),1)) break;
 }
 return 1;
}

unsigned long PharLapSegInfo( void )
{
 BGLOBAL handle;
 unsigned nnames;
 unsigned long fpos;
 memArray * obj;
 if(nph.plSegInfoOffset && nph.plSegInfoSize) nnames = (unsigned)(nph.plSegInfoSize / sizeof(PLSegInfo));
 else                                           nnames = 0;
 fpos = BMGetCurrFilePos();
 if(!nnames) { NotifyBox(NOT_ENTRY," Segment Info table "); return fpos; }
 if(!(obj = ma_Build(nnames,1))) return fpos;
 handle = pl_cache;
 bioSeek(handle,nph.plSegInfoOffset,SEEK_SET);
 if(__PLReadSegInfo(handle,obj,nnames))
 {
    int i;
    i = PageBox(50,4,(void **)obj->data,obj->nItems,PLSegPaint) + 1;
    if(i > 0)
    {
      fpos = ((unsigned long)((PLSegInfo *)obj->data[i - 1])->siBaseOff)+nph.plImageOffset;
    }
 }
 ma_Destroy(obj);
 return fpos;
}

void PLRunTimePaint(TWindow * win,void * * names,unsigned start,unsigned nlist)
{
 char buffer[81];
 char sign[3];
 PLRunTimeParms * * nam = (PLRunTimeParms * *)names;
 twUseWin(win);
 twFreezeWin(win);
 twClearWin();
 sprintf(buffer," Run-time Parameters Table [ %u / %u ] ",start + 1,nlist);
 twSetTitle(win,buffer,TMODE_CENTER,Black,LightGray);
 twSetFooter(win,PAGEBOX_SUB,TMODE_RIGHT,White,Black);
 strncpy(sign,(char *)nam[start]->rtSignature,2);
 sign[2] = 0;
 twGotoXY(1,1);
 twPrintF("Signature                      = %s\n"
          "Min. number of real-mode parms = %04hXH\n"
          "Max. number of real-mode parms = %04hXH\n"
          "Min. interrupt buffer size     = %04hXH\n"
          "Max. interrupt buffer size     = %04hXH\n"
          "Number of interrupt stacks     = %04hXH\n"
          "Size of each interrupt stack   = %04hXH\n"
          "Offset of end of real-mode data= %08lXH\n"
          "Call buffer size               = %04hXH\n"
          "Flags                          = %04hXH\n"
          "Unpriviledge flags             = %04hXH"
          ,sign
          ,nam[start]->rtMinRModeParms
          ,nam[start]->rtMaxRModeParms
          ,nam[start]->rtMinIBuffSize
          ,nam[start]->rtMaxIBuffSize
          ,nam[start]->rtNIStacks
          ,nam[start]->rtIStackSize
          ,nam[start]->rtEndRModeOffset
          ,nam[start]->rtCallBuffSize
          ,nam[start]->rtFlags
          ,nam[start]->rtUnprivFlags);
 twRefreshWin(win);
}


int __PLReadRunTime(BGLOBAL handle,memArray * obj,unsigned nnames)
{
 unsigned i;
 for(i = 0;i < nnames;i++)
 {
   PLRunTimeParms plrtp;
   if(IsKbdTerminate()) break;
   bioReadBuffer(handle,&plrtp,sizeof(PLRunTimeParms));
   if(!ma_AddData(obj,&plrtp,sizeof(PLRunTimeParms),1)) break;
 }
 return 1;
}

unsigned long PharLapRunTimeParms( void )
{
 BGLOBAL handle;
 unsigned nnames;
 unsigned long fpos;
 memArray * obj;
 if(nph.plRunTimeParms && nph.plRunTimeSize) nnames = (unsigned)(nph.plRunTimeSize / sizeof(PLRunTimeParms));
 else                                          nnames = 0;
 fpos = BMGetCurrFilePos();
 if(!nnames) { NotifyBox(NOT_ENTRY," Run-time parameters "); return fpos; }
 if(!(obj = ma_Build(nnames,1))) return fpos;
 handle = pl_cache;
 bioSeek(handle,nph.plRunTimeParms,SEEK_SET);
 if(__PLReadRunTime(handle,obj,nnames))
 {
    int i;
    i = PageBox(50,11,(void **)obj->data,obj->nItems,PLRunTimePaint) + 1;
    if(i > 0)
    {
      fpos = nph.plRunTimeParms+i*sizeof(PLRunTimeParms);
    }
 }
 ma_Destroy(obj);
 return fpos;
}

int IsPharLap( void )
{
   char sign[2];
   BMReadBufferEx(sign,2,0,BM_SEEK_SET);
   if(sign[0] == 'P' && (sign[1] == '2' || sign[1] == '3')) return 1;
   return 0;
}

void PharLapInit( void )
{
  BGLOBAL main_handle;
  BMReadBufferEx(&nph,sizeof(nph),0,BM_SEEK_SET);
  main_handle = BMbioHandle();
  if((pl_cache = bioDup(main_handle)) == &bNull) pl_cache = main_handle;
}

void PharLapDestroy( void )
{
  BGLOBAL main_handle;
  main_handle = BMbioHandle();
  if(pl_cache != &bNull && pl_cache != main_handle) bioClose(pl_cache);
}

void PharLapAddrResolv(char *addr,unsigned long cfpos)
{
  if(cfpos < sizeof(newPharLap))
  {
    sprintf(addr,"nplhdr:%s",Get2Digit(cfpos));
  }
}

unsigned long PharLapHelp( void )
{
  hlpDisplay(10010);
  return BMGetCurrFilePos();
}

REGISTRY_BIN PharLapTable =
{
  "PharLap",
  { "PLHelp", NULL, NULL, NULL, NULL, NULL, NULL, NULL, "RunTim", "SegInf" },
  { PharLapHelp, NULL, NULL, NULL, NULL, NULL, NULL, NULL, PharLapRunTimeParms, PharLapSegInfo },
  IsPharLap,
  PharLapInit,
  PharLapDestroy,
  ShowPharLapHeader,
  NULL,
  NULL,
  NULL,
  PharLapAddrResolv,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL
};
  