/****************************************************************************\
*   RDOFF - this source file is part of Binary vIEW project (BIEW)           *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/rdoff.c                                *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of decoder for RDOFF v1 file format      *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stddef.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

#include "reg_form.h"
#include "bin_util.h"
#include "pmalloc.h"
#include "bmfile.h"
#include "kbd_code.h"
#include "codeguid.h"
#include "biewhelp.h"
#include "biewutil.h"
#include "bconsole.h"
#include "twin.h"
#include "tstrings.h"

#include "plugins/bin/rdoff.h"

struct rdoff_ImpName
{
  unsigned short lsegno;
  unsigned short reserv;
  unsigned long nameoff;
};

static unsigned long rdoff_hdrlen,cs_start,ds_start,ds_len;

static linearArray *rdoffReloc = NULL;
static linearArray *rdoffImpNames = NULL;

static void __NEAR__ ReadImpNameList(BGLOBAL handle,int interact);

static void rdoff_ReadPubNameList(BGLOBAL handle,int interact);
static int __NEAR__ FindPubName(char *buff,unsigned cb_buff,unsigned long pa);

static int compare_impnames(const void __HUGE__ *v1,const void __HUGE__ *v2);

int rdoffLowMemFunc( unsigned long need_mem )
{
  int ret;
  ret = 0;
  if(!fmtActiveState)
  {
    if(PubNames)
    {
       la_Destroy(PubNames);
       PubNames = NULL;
       ret = 1;
    }
    if(rdoffImpNames)
    {
       la_Destroy(rdoffImpNames);
       rdoffImpNames = NULL;
       ret = 1;
    }
    if(rdoffReloc)
    {
       la_Destroy(rdoffReloc);
       rdoffReloc = NULL;
       ret = 1;
    }
  }
  return ret;
}

unsigned long rdoff_Help( void )
{
  hlpDisplay(10011);
  return BMGetCurrFilePos();
}

static int __NEAR__ rdoff_skiprec(unsigned char type) /* return 0 if error */
{
  unsigned i;
  int ret;
  unsigned char nulch;
  ret = 1;
  switch(type)
  {
    case 1: /* reloc record */
            BMSeek(8,BM_SEEK_CUR);
            break;
    case 2: /* import symbol */
            BMSeek(2,BM_SEEK_CUR);
            for(i = 0;i < 32;i++)
            {
              nulch = BMReadByte();
              if(!nulch) break;
            }
            break;
    case 3: /* export symbol */
            BMSeek(5,BM_SEEK_CUR);
            for(i = 0;i < 32;i++)
            {
              nulch = BMReadByte();
              if(!nulch) break;
            }
            break;
    case 4: /* impotr library */
            for(i = 0;i < 128;i++)
            {
              nulch = BMReadByte();
              if(!nulch) break;
            }
            break;
    case 5: /* reserve bss */
            BMSeek(4,BM_SEEK_CUR);
            break;
    default: /* unknown ??? */
            ErrMessageBox("Broken RDOFF file",NULL);
            ret = 0;
            break;
  }
  return ret;
}

unsigned long rdoff_ShowExport( void )
{
  unsigned long fpos;
  unsigned char rec;
  unsigned i;
  char str[33],sout[256];
  unsigned char segno;
  unsigned long segoff;
  unsigned long abs_off;
  memArray * rdoff_et;
  fpos = BMGetCurrFilePos();
  if(!(rdoff_et = ma_Build(0,1))) return fpos;
  BMSeek(10,BM_SEEK_SET);
  while(BMGetCurrFilePos() < rdoff_hdrlen + 5)
  {
    rec = BMReadByte();
    if(rec == 3)
    {
      char ch;
      segno = BMReadByte();
      segoff = BMReadDWord();
      for(i = 0;i < 32;i++)
      {
        ch = BMReadByte();
        str[i] = ch;
        if(!ch) break;
      }
      str[i] = 0;
      abs_off = segno == 0 ? cs_start : segno == 1 ? ds_start : ULONG_MAX;
      if(abs_off < ULONG_MAX) abs_off += segoff;
      sprintf(sout,"%-50s offset=%08lXH",str,abs_off);
      if(!ma_AddString(rdoff_et,sout,1)) break;
    }
    else
      if(!rdoff_skiprec(rec)) goto exit;
  }
  if(rdoff_et->nItems)
  {
    int ret;
    ret = ma_Display(rdoff_et,EXP_TABLE,LB_SELECTIVE,0);
    if(ret != -1)
    {
      char *rets;
      rets = ((char **)rdoff_et->data)[ret];
      rets = strstr(rets,"offset=");
      if(rets) fpos = strtoul(&rets[7],NULL,16);
    }
  }
  else                 NotifyBox(NOT_ENTRY,EXP_TABLE);
  exit:
  ma_Destroy(rdoff_et);
  BMSeek(fpos,BM_SEEK_SET);
  return fpos;
}

unsigned long rdoff_FindExport(char *name )
{
  unsigned long fpos,ret;
  unsigned char rec;
  unsigned i;
  char str[33];
  unsigned char segno;
  unsigned long segoff;
  unsigned long abs_off;
  fpos = BMGetCurrFilePos();
  BMSeek(10,BM_SEEK_SET);
  ret = 0L;
  while(BMGetCurrFilePos() < rdoff_hdrlen + 5)
  {
    rec = BMReadByte();
    if(rec == 3)
    {
      char ch;
      segno = BMReadByte();
      segoff = BMReadDWord();
      for(i = 0;i < 32;i++)
      {
        ch = BMReadByte();
        str[i] = ch;
        if(!ch) break;
      }
      str[i] = 0;
      if(strcmp(str,name) == 0)
      {
        abs_off = segno == 0 ? cs_start : segno == 1 ? ds_start : NULL;
        if(abs_off != NULL) abs_off += segoff;
        ret = abs_off;
        break;
      }
    }
    else
      if(!rdoff_skiprec(rec)) break;
  }
  BMSeek(fpos,BM_SEEK_SET);
  return ret;
}

unsigned long rdoff_ShowModRef( void )
{
  unsigned long fpos;
  unsigned char rec;
  unsigned i;
  char str[129];
  memArray * rdoff_mr;
  fpos = BMGetCurrFilePos();
  if(!(rdoff_mr = ma_Build(0,1))) return fpos;
  BMSeek(10,BM_SEEK_SET);
  while(BMGetCurrFilePos() < rdoff_hdrlen + 5)
  {
    rec = BMReadByte();
    if(rec == 4)
    {
      char ch;
      for(i = 0;i < 128;i++)
      {
        ch = BMReadByte();
        str[i] = ch;
        if(!ch) break;
      }
      str[i] = 0;
      if(!ma_AddString(rdoff_mr,str,1)) break;
    }
    else
      if(!rdoff_skiprec(rec)) goto exit;
  }
  if(rdoff_mr->nItems) ma_Display(rdoff_mr,MOD_REFER,LB_SORTABLE,0);
  else                 NotifyBox(NOT_ENTRY,MOD_REFER);
  exit:
  ma_Destroy(rdoff_mr);
  BMSeek(fpos,BM_SEEK_SET);
  return fpos;
}

unsigned long rdoff_ShowImport( void )
{
  unsigned long fpos;
  unsigned char rec;
  unsigned i;
  char str[33];
  memArray * rdoff_it;
  fpos = BMGetCurrFilePos();
  if(!(rdoff_it = ma_Build(0,1))) return fpos;
  BMSeek(10,BM_SEEK_SET);
  while(BMGetCurrFilePos() < rdoff_hdrlen + 5)
  {
    rec = BMReadByte();
    if(rec == 2)
    {
      char ch;
      BMSeek(2,BM_SEEK_CUR);
      for(i = 0;i < 32;i++)
      {
        ch = BMReadByte();
        str[i] = ch;
        if(!ch) break;
      }
      str[i] = 0;
      if(!ma_AddString(rdoff_it,str,1)) break;
    }
    else
      if(!rdoff_skiprec(rec)) goto exit;
  }
  if(rdoff_it->nItems) ma_Display(rdoff_it,IMPPROC_TABLE,LB_SORTABLE,0);
  else                 NotifyBox(NOT_ENTRY,EXP_TABLE);
  exit:
  ma_Destroy(rdoff_it);
  BMSeek(fpos,BM_SEEK_SET);
  return fpos;
}

unsigned long rdoff_ShowHeader( void )
{
  int endian;
  unsigned long fpos,entry;
  unsigned long hs_len,cs_len;
  TWindow *w;
  fpos = BMGetCurrFilePos();
  endian = BMReadByteEx(5,BM_SEEK_SET);
  hs_len = BMReadDWord();
  BMSeek(hs_len,BM_SEEK_CUR);
  cs_len = BMReadDWord();
  BMSeek(cs_len,BM_SEEK_CUR);
  ds_len = BMReadDWord();
  cs_start = hs_len + 14;
  ds_start = cs_start + cs_len + 4;
  w = CrtHlpWnd(endian == 0x01 ? " RDOFF big endian " : " RDOFF little endian ",54,6);
  twGotoXY(1,1);
  twPrintF("Length of header section    = %08lXH\n"
           "Length of code section      = %08lXH\n"
           "segment .code               = %08lXH\n"
           "Length of data secion       = %08lXH\n"
           "segment .data               = %08lXH\n"
           ,hs_len
           ,cs_len
           ,cs_start
           ,ds_len
           ,ds_start);
  twSetColor(White,Black);
  entry = rdoff_FindExport("_main");
  twPrintF("Entry point                 = %08lXH"
           ,entry);
  twClrEOL();
  twSetColor(Black,LightGray);
  while(1)
  {
    int keycode;
    keycode = GetEvent(drawEmptyPrompt,w);
    switch(keycode)
    {
      case KE_ENTER:      fpos = entry; goto exit;
      case KE_ESCAPE:     goto exit;
      default:            break;
    };
  }
  exit:
  CloseWnd(w);
  BMSeek(fpos,BM_SEEK_SET);
  return fpos;
}

static int rdoff_compare_reloc(const void __HUGE__ *e1,const void __HUGE__ *e2)
{
  const RDOFF_RELOC __HUGE__ *r1,__HUGE__ *r2;
  int ret;
  r1 = e1;
  r2 = e2;
  if(r1->offset < r2->offset) ret = -1;
  else
    if(r1->offset > r2->offset) ret = 1;
    else                      ret = 0;
  return ret;
}

static void __NEAR__ BuildRelocRDOFF( void )
{
  unsigned long fpos;
  unsigned char rec;
  fpos = BMGetCurrFilePos();
  if(!(rdoffReloc = la_Build(0,sizeof(RDOFF_RELOC),1))) return;
  BMSeek(10,BM_SEEK_SET);
  while(BMGetCurrFilePos() < rdoff_hdrlen + 5)
  {
    RDOFF_RELOC rdf_r;
    rec = BMReadByte();
    if(rec == 1)
    {
      unsigned char segno;
      unsigned long off;
      segno = BMReadByte();
      rdf_r.is_rel = 0;
      if(segno >= 64)
      {
        rdf_r.is_rel = 1;
        segno -= 64;
      }
      rdf_r.offset = segno == 0 ? cs_start : segno == 1 ? ds_start : ULONG_MAX;
      off = BMReadDWord();
      if(rdf_r.offset < ULONG_MAX) rdf_r.offset += off;
      rdf_r.reflen = BMReadByte();
      rdf_r.segto = BMReadWord();
      if(!la_AddData(rdoffReloc,&rdf_r,1)) break;
    }
    else
      if(!rdoff_skiprec(rec)) goto exit;
  }
  exit:
  BMSeek(fpos,BM_SEEK_SET);
  la_Sort(rdoffReloc,rdoff_compare_reloc);
}

static unsigned char __codelen;

static int rdoffBuildReferStr(char *str,RDOFF_RELOC *reloc,unsigned long ulShift,int flags)
{
   char *ptr_type;
   char name[400],ch,buff[400];
   unsigned i;
   unsigned long field_val,foff,base_seg_off;
   if(flags & APREF_USE_TYPE)
   {
     switch(reloc->reflen)
     {
       case 1: ptr_type = "(b) "; break;
       case 2: ptr_type = "off16 "; break;
       default:
       case 4: ptr_type = "off32 "; break;
     }
     strcat(str,ptr_type);
   }
   base_seg_off = ULONG_MAX;
   switch(reloc->segto)
   {
     case 0:  ptr_type = "cs:";
              base_seg_off = cs_start;
              break;
     case 1:  ptr_type = "ds:";
              base_seg_off = ds_start;
              break;
     case 2:  ptr_type = "bss:";
              break;
     default:
              {
                 struct rdoff_ImpName key,*ret;
                 key.lsegno = reloc->segto;
                 ret = la_Find(rdoffImpNames,&key,compare_impnames);
                 if(ret)
                 {
                   BMSeek(ret->nameoff,BM_SEEK_SET);
                   name[0] = 0;
                   for(i = 0;i < sizeof(name);i++)
                   {
                     ch = BMReadByte();
                     name[i] = ch;
                     if(!ch) break;
                   }
                   name[i] = 0;
                   ptr_type = name;
                 }
                 else
                 {
                   ptr_type = "???";
                 }
              }
   }
   if(reloc->segto >= 3) strcat(str,ptr_type); /* case extern symbol */
   if(reloc->segto < 3 || reloc->is_rel)
   {
     foff = BMGetCurrFilePos();
     switch(reloc->reflen)
     {
       default:
       case 4: field_val = BMReadDWordEx(reloc->offset,BM_SEEK_SET);
               break;
       case 2: field_val = BMReadWordEx(reloc->offset,BM_SEEK_SET);
               break;
       case 1: field_val = BMReadByteEx(reloc->offset,BM_SEEK_SET);
               break;
     }
     BMSeek(foff,BM_SEEK_SET);
     if(reloc->segto < 3)
     {
       if(base_seg_off < ULONG_MAX)
       {
         base_seg_off += field_val;
         if(FindPubName(buff,sizeof(buff),base_seg_off))
         {
           strcat(str,buff);
         }
         else goto unnamed;
       }
       else
       {
         unnamed:
         strcat(str,ptr_type);
         strcat(str,Get8Digit(field_val));
         if(!EditMode && !DumpMode && (flags & APREF_TRY_LABEL))
           GidAddGoAddress(str,reloc->segto ? ds_start + field_val : cs_start + field_val);
       }
     }
     if(reloc->is_rel)
     {
       unsigned long seg_off;
       long fv;
       seg_off = ulShift < ds_start ? cs_start : ulShift < ds_start+ds_len ? ds_start : ULONG_MAX;
       fv = field_val;
       if(!(ulShift + fv + reloc->reflen == seg_off && (flags & APREF_TRY_LABEL)))
       {
         strcat(str,"(");
         strcat(str, Get8SignDig(field_val));
         strcat(str,")");
       }
     }
   }
   return 1;
}

int rdoff_AppendRef(char *str,unsigned long ulShift,int flags,int codelen,unsigned long r_sh)
{
  RDOFF_RELOC *rnlm,key;
  int ret;
  char buff[400];
  if(!rdoffReloc) BuildRelocRDOFF();
  if(!rdoffImpNames) ReadImpNameList(BMbioHandle(),1);
  if(!PubNames) rdoff_ReadPubNameList(BMbioHandle(),1);
  key.offset = ulShift;
  __codelen = codelen;
  rnlm = la_Find(rdoffReloc,&key,rdoff_compare_reloc);
  if(!rnlm) ret = 0;
  else      ret = rdoffBuildReferStr(str,rnlm,ulShift,flags);
  if(!ret && (flags & APREF_TRY_LABEL))
  {
     if(FindPubName(buff,sizeof(buff),r_sh))
     {
       strcat(str,buff);
       if(!DumpMode && !EditMode) GidAddGoAddress(str,r_sh);
       ret = 1;
     }
  }
  return ret;
}

int rdoff_check_fmt( void )
{
  char rbuff[6];
  BMReadBufferEx(rbuff,sizeof(rbuff),0L,BM_SEEK_SET);
  return memcmp(rbuff,"RDOFF1",sizeof(rbuff)) == 0 ||
         memcmp(rbuff,"RDOFF\x01",sizeof(rbuff) == 0);
}

void rdoff_init_fmt( void )
{
  unsigned long cs_len;
  PMRegLowMemCallBack(rdoffLowMemFunc);
  rdoff_hdrlen = BMReadDWordEx(6,BM_SEEK_SET);
  BMSeek(rdoff_hdrlen,BM_SEEK_CUR);
  cs_len = BMReadDWord();
  BMSeek(cs_len,BM_SEEK_CUR);
  ds_len = BMReadDWord();
  cs_start = rdoff_hdrlen + 14;
  ds_start = cs_start + cs_len + 4;
}

void rdoff_destroy_fmt()
{
  if(rdoffReloc) { la_Destroy(rdoffReloc); rdoffReloc = NULL; }
  if(rdoffImpNames) { la_Destroy(rdoffImpNames); rdoffImpNames = NULL; }
  if(PubNames) { la_Destroy(PubNames); PubNames = NULL; }
  PMUnregLowMemCallBack(rdoffLowMemFunc);
}

void rdoff_ReadPubName(BGLOBAL b_cache,unsigned long nameoff,char *buff,unsigned cb_buff)
{
    unsigned char ch;
    unsigned i;
    bioSeek(b_cache,nameoff,SEEK_SET);
    for(i = 0;i < cb_buff;i++)
    {
      ch = BMReadByte();
      buff[i] = ch;
      if(!ch) break;
    }
}

static int __NEAR__ FindPubName(char *buff,unsigned cb_buff,unsigned long pa)
{
  BGLOBAL b_cache;
  b_cache = BMbioHandle();
  return fmtFindPubName(b_cache,buff,cb_buff,pa,
                        rdoff_ReadPubNameList,
                        rdoff_ReadPubName);
}

static void rdoff_ReadPubNameList(BGLOBAL handle,int interact)
{
 unsigned char segno,rec;
 unsigned long segoff,abs_off;
 unsigned i;
 struct PubName rdf_pn;
 if(!PubNames)
   if(!(PubNames = la_Build(0,sizeof(struct PubName),interact))) return;
 BMSeek(10,BM_SEEK_SET);
 while(BMGetCurrFilePos() < rdoff_hdrlen + 5)
 {
    rec = BMReadByte();
    if(rec == 3)
    {
      char ch;
      segno = BMReadByte();
      segoff = BMReadDWord();
      rdf_pn.nameoff = BMGetCurrFilePos();
      for(i = 0;i < 32;i++)
      {
        ch = BMReadByte();
        if(!ch) break;
      }
      abs_off = segno == 0 ? cs_start : segno == 1 ? ds_start : ULONG_MAX;
      if(abs_off < ULONG_MAX) abs_off += segoff;
      rdf_pn.pa = abs_off;
      if(!la_AddData(PubNames,&rdf_pn,1)) break;
    }
    else
      if(!rdoff_skiprec(rec)) break;
 }
 if(PubNames->nItems) la_Sort(PubNames,fmtComparePubNames);
}

static int compare_impnames(const void __HUGE__ *v1,const void __HUGE__ *v2)
{
  struct rdoff_ImpName __HUGE__ *pnam1,__HUGE__ *pnam2;
  pnam1 = (struct rdoff_ImpName __HUGE__ *)v1;
  pnam2 = (struct rdoff_ImpName __HUGE__ *)v2;
  if(pnam1->lsegno < pnam2->lsegno) return -1;
  else
    if(pnam1->lsegno > pnam2->lsegno) return 1;
    else                              return 0;
}

static void __NEAR__ ReadImpNameList(BGLOBAL handle,int interact)
{
 unsigned char rec;
 unsigned i;
 struct rdoff_ImpName rdf_in;
 if(!rdoffImpNames)
   if(!(rdoffImpNames = la_Build(0,sizeof(struct rdoff_ImpName),interact))) return;
 BMSeek(10,BM_SEEK_SET);
 while(BMGetCurrFilePos() < rdoff_hdrlen + 5)
 {
    rec = BMReadByte();
    if(rec == 2)
    {
      char ch;
      rdf_in.lsegno = BMReadWord();
      rdf_in.nameoff = BMGetCurrFilePos();
      for(i = 0;i < 32;i++)
      {
        ch = BMReadByte();
        if(!ch) break;
      }
      if(!la_AddData(rdoffImpNames,&rdf_in,1)) break;
    }
    else
      if(!rdoff_skiprec(rec)) break;
 }
 if(rdoffImpNames->nItems) la_Sort(rdoffImpNames,compare_impnames);
}

/* bitness not declared, but we assign 32-bit as default */
int rdoff_bitness(unsigned long pa) { return DAB_USE32; }

unsigned long rdoffGetPubSym(char *str,unsigned cb_str,unsigned *func_class,
                           unsigned long pa,int as_prev)
{
  BGLOBAL b_cache;
  b_cache = BMbioHandle();
  return fmtGetPubSym(b_cache,str,cb_str,func_class,pa,as_prev,
                      rdoff_ReadPubNameList,
                      rdoff_ReadPubName);
}

unsigned rdoffGetObjAttr(unsigned long pa,char *name,unsigned cb_name,
                         unsigned long *start,unsigned long *end,int *class,int *bitness)
{
  unsigned ret;
  *start = 0;
  *end = BMGetFLength();
  *class = OC_NOOBJECT;
  *bitness = rdoff_bitness(pa);
  name[0] = 0;
  if(pa < cs_start)
  {
    *end = cs_start;
    ret = 0;
  }
  else
    if(pa >= cs_start && pa < ds_start - 4)
    {
      *start = cs_start;
      *class = OC_CODE;
      *end = ds_start-4;
      ret = 1;
    }
    else
    {
      if(pa >= ds_start - 4 && pa < ds_start)
      {
        *start = ds_start-4;
        *end = ds_start;
        *class = OC_NOOBJECT;
        ret = 2;
      }
      else
      {
        if(pa >= ds_start && pa < ds_start + ds_len)
        {
          *class = OC_DATA;
          *start = ds_start;
          *end = ds_start + ds_len;
          ret = 3;
        }
        else
        {
          *start = ds_start+ds_len;
          *class = OC_NOOBJECT;
          ret = 4;
        }
      }
    }
  return ret;
}

unsigned long rdoffVA2PA(unsigned long va)
{
  return va + cs_start;
}

unsigned long rdoffPA2VA(unsigned long pa)
{
  return pa > cs_start ? pa - cs_start : 0L;
}

void rdoff_AddressResolv(char *addr,unsigned long cfpos)
{
 if(cfpos < cs_start)
 {
    sprintf(addr,"RDFH:%s",Get4Digit(cfpos));
 }
}

REGISTRY_BIN rdoffTable =
{
  "RDOFF (Relocatable Dynamic Object File Format)",
  { "RdHelp", "ModRef", "Export", NULL, "Import", NULL, NULL, NULL, NULL, NULL },
  { rdoff_Help, rdoff_ShowModRef, rdoff_ShowExport, NULL, rdoff_ShowImport, NULL, NULL, NULL, NULL, NULL },
  rdoff_check_fmt,
  rdoff_init_fmt,
  rdoff_destroy_fmt,
  rdoff_ShowHeader,
  rdoff_AppendRef,
  fmtSetState,
  rdoff_bitness,
  rdoff_AddressResolv,
  rdoffVA2PA,
  rdoffPA2VA,
  rdoffGetPubSym,
  rdoffGetObjAttr,
  NULL,
  NULL
};
  