/****************************************************************************\
*   IX86_FUN - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/disasm/ix86/ix86_fun.c                     *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation common function and utility for Intel    *
* x86 disassembler                                                           *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include "plugins/disasm/ix86/ix86.h"
#include "biewutil.h"
#include "bmfile.h"
#include "codeguid.h"
#include "disasm.h"
#include "reg_form.h"
#include <string.h>
#include <stdio.h>

extern int AppendAsmRef(char *str,unsigned long ulShift,int mode,char codelen);
char *ix86_appstr;
char *ix86_dtile;
char *ix86_appbuffer;
char *ix86_apistr;
char *ix86_modrm_ret;

void   ix86_setModifier(char *str,const char *modf)
{
  disSetModifier(str,modf,TAB_POS);
}

void ix86_ArgES(char *str,ix86Param *param)
{
  param->codelen += 0;
  strcat(str,"es");
}

void ix86_ArgDS(char *str,ix86Param *param)
{
  param->codelen += 0;
  strcat(str,"ds");
}

void ix86_ArgSS(char *str,ix86Param *param)
{
  param->codelen += 0;
  strcat(str,"ss");
}

void ix86_ArgCS(char *str,ix86Param *param)
{
  param->codelen += 0;
  strcat(str,"cs");
}

void ix86_ArgFS(char *str,ix86Param *param)
{
  param->codelen += 0;
  strcat(str,"fs");
}

void ix86_ArgGS(char *str,ix86Param *param)
{
  param->codelen += 0;
  strcat(str,"gs");
}

void ix86_ArgByte(char *str,ix86Param *DisP)
{
  unsigned char *tile;
  tile = &DisP->RealCmd[DisP->codelen];
  disAppendDigits(str,DisP->DisasmPrefAddr + ((char *)tile - (char *)DisP->CodeBuffer),
               APREF_USE_TYPE,DisP->codelen,&DisP->RealCmd[DisP->codelen],DISARG_BYTE);
  DisP->codelen++;
}

void ix86_ArgWord(char *str,ix86Param *DisP)
{
  unsigned char *tile;
  tile = &DisP->RealCmd[DisP->codelen];
  disAppendDigits(str,DisP->DisasmPrefAddr + ((char *)tile - (char *)DisP->CodeBuffer),
               APREF_USE_TYPE,DisP->codelen,&DisP->RealCmd[DisP->codelen],DISARG_WORD);
  DisP->codelen+=2;
}

void ix86_ArgDWord(char *str,ix86Param *DisP)
{
  unsigned char *tile;
  tile = &DisP->RealCmd[DisP->codelen];
  disAppendDigits(str,DisP->DisasmPrefAddr + ((char *)tile - (char *)DisP->CodeBuffer),
               APREF_USE_TYPE,DisP->codelen,&DisP->RealCmd[DisP->codelen],DISARG_DWORD);
  DisP->codelen+=4;
}

char * ix86_GetDigitTile(char w,MBuffer tile,ix86Param* DisP)
{
  int cl,type;
  cl = w ? ( Use32Data ? 4 : 2 ) : 1;
  type = w ? ( Use32Data ? DISARG_DWORD : DISARG_WORD ) : DISARG_BYTE;
  ix86_dtile[0] = 0;
  disAppendDigits(ix86_dtile,
               DisP->DisasmPrefAddr + ((char *)tile - (char *)DisP->CodeBuffer),
               APREF_USE_TYPE,DisP->codelen,tile,type);
  DisP->codelen += cl;
  return ix86_dtile;
}

void ix86_ArgFar(char * str,ix86Param *DisP)
{
  unsigned long newpos;
  long off;
  unsigned short seg;
  seg = Use32Data ? *(short  *)(&DisP->RealCmd[5]) : *(short  *)(&DisP->RealCmd[3]);
  off = Use32Data ? *(long  *)(&DisP->RealCmd[1]) : (long)(*(short  *)(&DisP->RealCmd[1]));
  newpos = ((long)(seg) << 4) + off;
  if(FType == OLD_EXE) newpos += (((unsigned long)oldh.mz.mzHeaderSize) << 4);
  if(Use32Data) DisP->codelen += 6;
  else          DisP->codelen += 4;
  disAppendFAddr(str,DisP->CodeAddress + 1,off,newpos,
                 Use32Data ? DISADR_FAR32 : DISADR_FAR16,seg,DisP->codelen - 1);
}

void ix86_ArgNear(char * str,ix86Param *DisP)
{
  long lshift;
  unsigned long newpos;
  lshift = Use32Data ? *(long  *)(&DisP->RealCmd[1]) : (long)(*(short  *)(&DisP->RealCmd[1]));
  if(!Use32Data) DisP->codelen += 2;
  else           DisP->codelen += 4;
  newpos = DisP->CodeAddress + lshift + DisP->codelen;
  disAppendFAddr(str,DisP->CodeAddress + 1,lshift,newpos,
                 Use32Data ? DISADR_NEAR32 : DISADR_NEAR16,0,DisP->codelen - 1);
}

void ix86_ArgShort(char * str,ix86Param *DisP)
{
  unsigned long r_sh;
  signed char distin;
  distin = DisP->RealCmd[DisP->codelen++];
  r_sh = DisP->CodeAddress + distin + 2; // 2 byte it self code jmp<c> addr
  disAppendFAddr(str,DisP->CodeAddress + 1,(long)distin,r_sh,
                 DISADR_SHORT,0,1);
}

static char * __NEAR__ Get2DigitApp(unsigned char *val,ix86Param *DisP)
{
  ix86_appbuffer[0] = 0;
  disAppendDigits(ix86_appbuffer,DisP->DisasmPrefAddr + (val - DisP->CodeBuffer),
               APREF_USE_TYPE,1,val,DISARG_BYTE);
  return ix86_appbuffer;
}

static char * __NEAR__ Get2SignDigApp(signed char *val,ix86Param *DisP)
{
  ix86_appbuffer[0] = 0;
  disAppendDigits(ix86_appbuffer,DisP->DisasmPrefAddr + (val - (signed char *)DisP->CodeBuffer),
               APREF_USE_TYPE,1,val,DISARG_CHAR);
  return ix86_appbuffer;
}

static char * __NEAR__ Get4DigitApp(unsigned char *val,ix86Param *DisP)
{
  ix86_appbuffer[0] = 0;
  disAppendDigits(ix86_appbuffer,DisP->DisasmPrefAddr + (val - DisP->CodeBuffer),
               APREF_USE_TYPE,2,val,DISARG_WORD);
  return ix86_appbuffer;
}

static char * __NEAR__ Get4SignDigApp(unsigned char *val,ix86Param *DisP)
{
  ix86_appbuffer[0] = 0;
  disAppendDigits(ix86_appbuffer,DisP->DisasmPrefAddr + (val - DisP->CodeBuffer),
               APREF_USE_TYPE,2,val,DISARG_SHORT);
  return ix86_appbuffer;
}

static char * __NEAR__ Get8DigitApp(unsigned char *val,ix86Param *DisP)
{
  ix86_appbuffer[0] = 0;
  disAppendDigits(ix86_appbuffer,DisP->DisasmPrefAddr + (val - DisP->CodeBuffer),
               APREF_USE_TYPE,4,val,DISARG_DWORD);
  return ix86_appbuffer;
}

static char * __NEAR__ Get8SignDigApp(unsigned char *val,ix86Param *DisP)
{
  ix86_appbuffer[0] = 0;
  disAppendDigits(ix86_appbuffer,DisP->DisasmPrefAddr + (val - DisP->CodeBuffer),
               APREF_USE_TYPE,4,val,DISARG_LONG);
  return ix86_appbuffer;
}

static char * __NEAR__ Get2SquareDig(unsigned char *val,ix86Param *DisP,int is_stack)
{
  char *ptr = ix86_apistr;
  char *rets;
  rets = is_stack ? Get2SignDigApp((signed char *)val,DisP) : Get2DigitApp(val,DisP);
  if(!(rets[0] == '+' || rets[0] == '-')) *ptr++ = '+';
  strcpy(ptr,rets);
  return ix86_apistr;
}

static char * __NEAR__ Get4SquareDig(unsigned char * val,ix86Param *DisP,int is_stack)
{
  char *ptr = ix86_apistr;
  char *rets;
  rets = is_stack ? Get4SignDigApp(val,DisP) : Get4DigitApp(val,DisP);
  if(!(rets[0] == '+' || rets[0] == '-')) *ptr++ = '+';
  strcpy(ptr,rets);
  return ix86_apistr;
}

static char * __NEAR__ Get8SquareDig(unsigned char * val,ix86Param *DisP,int is_stack)
{
  char *ptr = ix86_apistr;
  char *rets;
  rets = is_stack ? Get8SignDigApp(val,DisP) : Get8DigitApp(val,DisP);
  if(!(rets[0] == '+' || rets[0] == '-')) *ptr++ = '+';
  strcpy(ptr,rets);
  return ix86_apistr;
}

const char * ix86_getREG(unsigned char reg,unsigned char w)
{
 if(UseXMMXSet) return ix86_XMMXRegs[reg];
 else
   if(UseMMXSet)  return ix86_MMXRegs[reg];
   else
   {
     if(!Use32Data) return w ? ix86_WordRegs[reg] : ix86_ByteRegs[reg];
     else           return w ? ix86_DWordRegs[reg] : ix86_ByteRegs[reg];
   }
}

static const char * __NEAR__ getSREG(unsigned char sreg)
{
  return ix86_SegRegs[sreg];
}

static int __NEAR__ getSIBRegs(char * base,char * scale,char * index,char *mod,char code)
{
  char scl = (code & 0xC0) >> 6;
  char bas = code & 0x07;
  char ind = (code & 0x38) >> 3;
  char use32data = Use32Data;
  char useMMX = UseMMXSet;
  Use32Data = 0xFF;
  UseMMXSet = 0;
  if(scl)
  {
    scale[0] = '*';
    scale[1] = ( 1 << scl ) + 0x30;
    scale[2] = 0;
  }
  else scale[0] = 0;
  if(ind == 4) index[0] = 0;
  else         strcpy(index,ix86_getREG(ind,1));
  if(bas == 5 && *mod == 0) { base[0] = 0; *mod = 2; }
  else         strcpy(base,ix86_getREG(bas,1));
  Use32Data = use32data;
  UseMMXSet = useMMX;
  return ind == 4 || ind == 5 || bas == 4 || bas == 5; // return is_stack
}


static char * __NEAR__ ConstrSibMod(char * store,char * scale,char * index,char * base,char code,char *mod,int *is_stack)
{
   *is_stack = getSIBRegs(base,scale,index,mod,code);
   store[0] = 0;
   if(index[0])
   {
     strcat(store,index);
     strcat(store,scale);
   }
   if(base[0])
   {
     if(index[0]) strcat(store,"+");
     strcat(store,base);
   }
   return store;
}

char * ix86_getModRM(unsigned char w,unsigned char mod,unsigned char rm,ix86Param *DisP)
{
 char square[50];
 MBuffer Disp = DisP->RealCmd;
 char ret1[50];
 char base[8];
 char index[8];
 char scale[4];
 char new_mode = mod;
 unsigned char clen;
 int is_stack;
 clen = 2;
 if(rm == 4 && Use32Addr) strcpy(square,ConstrSibMod(ret1,base,index,scale,DisP->RealCmd[2],&new_mode,&is_stack));
 else
 {
   strcpy(square,Use32Addr ? ix86_DWordRegs[rm] : ix86_A16[rm]);
   is_stack = Use32Addr ? rm == 4 || rm == 5 : rm == 2 || rm == 3 || rm == 6;
 }
 mod = new_mode;
 ix86_modrm_ret[0] = 0;
 if(mod != 3)
 {
   strcpy(ix86_modrm_ret,"[");
   strcat(ix86_modrm_ret,ix86_segpref);
   ix86_segpref[0] = 0;
 }
 switch(mod)
 {
       case 0:
             {
               if((rm != 6 && !Use32Addr) || (rm != 5 && Use32Addr)) // i.e. combine address
               {
                 clen = 0;
                 if(Use32Addr && rm == 4) clen = 1;
                 strcat(ix86_modrm_ret,square);
               }
               else /* fake mod = 2 */
               {
                  square[0] = 0;
                  goto disp_long;
               }
             }
             break;
       case 1:
             {
               if(!Use32Addr)
               {
                   strcat(ix86_modrm_ret,ix86_segpref);
                   ix86_segpref[0] = 0;
                   strcat(ix86_modrm_ret,square);
                   strcat(ix86_modrm_ret,Get2SquareDig((unsigned char *)&Disp[2],DisP,is_stack));
                   clen = 1;
               }
               else
               {
                 char tmp;
                 if(rm != 4) { tmp = Disp[2]; clen = 1; }
                 else        { tmp = Disp[3]; clen = 2; }
                 strcat(ix86_modrm_ret,square);
                 strcat(ix86_modrm_ret,Get2SquareDig((unsigned char *)&tmp,DisP,is_stack));
               }
             }
             break;
       case 2:
            {
              disp_long:
              if(!Use32Addr)
              {
                strcat(ix86_modrm_ret,ix86_segpref);
                ix86_segpref[0] = 0;
                strcat(ix86_modrm_ret,square);
                strcat(ix86_modrm_ret,Get4SquareDig((unsigned char *)&Disp[2],DisP,is_stack));
                clen = 2;
              }
              else
              {
                if(rm != 4) { Disp = &Disp[2]; clen = 4; }
                else        { Disp = &Disp[3]; clen = 5; }
                strcat(ix86_modrm_ret,square);
                strcat(ix86_modrm_ret,Get8SquareDig(Disp,DisP,is_stack));
              }
            }
            break;
      default:
            {
              clen = 0;
              strcat(ix86_modrm_ret,ix86_getREG(rm,w));
            }
            break;
   }
   if(mod != 3)
   {
     DisP->codelen += clen;
     strcat(ix86_modrm_ret,"]");
   }
   return ix86_modrm_ret;
}

char * ix86_CStile(char *str,const char *arg2)
{
  strcat(str,",");
  strcat(str,arg2);
  return str;
}

static char * __NEAR__ __getTile(char w,char d,ix86Param *DisP)
{
 char reg = (DisP->RealCmd[1] & 0x38) >> 3;
 char mod = (DisP->RealCmd[1] & 0xC0) >> 6;
 char rm  = (DisP->RealCmd[1] & 0x07);
 const char *regs,* modrm;
 regs = ix86_getREG(reg,w);
 if(mod == 3) modrm = ix86_getREG(rm,w);
 else         modrm = ix86_getModRM(w,mod,rm,DisP);
 ix86_dtile[0] = 0;
 strcat(ix86_dtile,d ? regs : modrm);
 ix86_CStile(ix86_dtile,d ? modrm : regs);
 return ix86_dtile;
}


static char * __NEAR__ getTile( ix86Param *DisP )
{
  return __getTile(DisP->RealCmd[0] & 0x01,DisP->RealCmd[0] & 0x02,DisP);
}

char * ix86_getTileWD( ix86Param *DisP )
{
  return __getTile(1,1,DisP);
}

char * ix86_getTileWnD( ix86Param *DisP )
{
  return __getTile(1,0,DisP);
}

char * ix86_getTilenWD( ix86Param *DisP )
{
  return __getTile(0,1,DisP);
}

char * ix86_getTilenWnD( ix86Param *DisP )
{
  return __getTile(0,0,DisP);
}

void ix86_ArgModRM(char * str,ix86Param *DisP)
{
 DisP->codelen++;
 strcat(str,getTile(DisP));
}

void          ix86_ArgModRMDW(char *str,ix86Param *DisP)
{
 DisP->codelen++;
 strcat(str,ix86_getTileWD(DisP));
}

void          ix86_ArgModRMnDW(char *str,ix86Param *DisP)
{
 DisP->codelen++;
 strcat(str,ix86_getTileWnD(DisP));
}

void          ix86_ArgModRMDnW(char *str,ix86Param *DisP)
{
 DisP->codelen++;
 strcat(str,ix86_getTilenWD(DisP));
}

void          ix86_ArgModRMnDnW(char *str,ix86Param *DisP)
{
 DisP->codelen++;
 strcat(str,ix86_getTilenWnD(DisP));
}

void          ix86_ArgMod(char *str,ix86Param *DisP)
{
  char mod = (DisP->RealCmd[1] >> 6) & 0x03;
  char rm = (DisP->RealCmd[1] & 0x07);
  int sizptr;
  sizptr = Use32Data ? DWORD_PTR : WORD_PTR;
  ix86_setModifier(str,ix86_sizes[sizptr]);
  DisP->codelen++;
  strcat(str,ix86_getModRM(1,mod,rm,DisP));
}

void          ix86_ArgModB(char *str,ix86Param *DisP)
{
  char mod = (DisP->RealCmd[1] >> 6) & 0x03;
  char rm = (DisP->RealCmd[1] & 0x07);
  ix86_setModifier(str,ix86_sizes[BYTE_PTR]);
  DisP->codelen++;
  strcat(str,ix86_getModRM(0,mod,rm,DisP));
}

void ix86_SMov(char * str,ix86Param *DisP)
{
  char direct = ( DisP->RealCmd[0] & 0x02 ) >> 1;
  char sreg  = (DisP->RealCmd[1] & 0x38) >> 3;
  char reg   = DisP->RealCmd[1] & 0x07;
  char mod   = (DisP->RealCmd[1] & 0xC0) >> 6;
  const char *tileptr,*sregptr;
  char use32data = Use32Data;
  Use32Data = 0;
  sregptr = getSREG(sreg);
  DisP->codelen = 2;
  tileptr = ix86_getModRM(1,mod,reg,DisP);
  if(sreg > 3) if((DisP->pro_clone & IX86_CPUMASK) < 3) DisP->pro_clone = IX86_CPU386;
  Use32Data = use32data;
  strcat(str,direct ? sregptr : tileptr);
  ix86_CStile(str,direct ? tileptr : sregptr);
}

void ix86_ArgSegRM(char * str,ix86Param *DisP)
{
  char sreg  = (DisP->RealCmd[1] & 0x38) >> 3;
  char reg   = DisP->RealCmd[1] & 0x07;
  char mod   = (DisP->RealCmd[1] & 0xC0) >> 6;
  const char *tileptr,*sregptr;
  char use32data = Use32Data;
  Use32Data = 0;
  sregptr = getSREG(sreg);
  strcat(str,sregptr);
  DisP->codelen++;
  tileptr = ix86_getModRM(1,mod,reg,DisP);
  if(sreg > 3) if((DisP->pro_clone & IX86_CPUMASK) < 3) DisP->pro_clone = IX86_CPU386;
  Use32Data = use32data;
  ix86_CStile(str,tileptr);
}

void          ix86_ArgAXDigit(char * str,ix86Param *DisP)
{
  char w = DisP->RealCmd[0] & 0x01;
  strcat(str,ix86_getREG(0,w));
  ix86_CStile(str,ix86_GetDigitTile(w,&DisP->RealCmd[1],DisP));
}

void          ix86_ArgRMDigit(char * str,ix86Param *DisP)
{
  char w = DisP->RealCmd[0] & 0x01;
  char mod = (DisP->RealCmd[1] & 0xC0) >> 6;
  char rm  = (DisP->RealCmd[1] & 0x07);
  char *a1,*a2;
  DisP->codelen++;
  a1 = ix86_getModRM(w,mod,rm,DisP);
  strcat(str,a1);
  a2 = ix86_GetDigitTile(w,&DisP->RealCmd[DisP->codelen],DisP);
  ix86_CStile(str,a2);
}

void          ix86_ArgIRegDigit(char * str,ix86Param *DisP)
{
  char w = (DisP->RealCmd[0] >> 3 ) & 0x01;
  char reg = (DisP->RealCmd[0]) & 0x07;
  strcat(str,ix86_getREG(reg,w));
  ix86_CStile(str,ix86_GetDigitTile(w,&DisP->RealCmd[1],DisP));
}

void          ix86_ArgIReg(char *str,ix86Param *DisP)
{
  strcat(str,ix86_getREG(DisP->RealCmd[0] & 0x07,1));
}

void          ix86_ArgAXIReg(char *str,ix86Param *DisP)
{
  strcat(str,ix86_getREG(0,1));
  ix86_CStile(str,ix86_getREG(DisP->RealCmd[0] & 0x07,1));
}

void          ix86_ArgSInt(char *str,ix86Param *DisP)
{
  ix86_setModifier(str,Use32Data ? ix86_sizes[DWORD_PTR] : ix86_sizes[WORD_PTR]);
  ix86_ArgByte(str,DisP);
}

void          ix86_ArgInt(char *str,ix86Param *DisP)
{
  Use32Data ? ix86_ArgDWord(str,DisP) : ix86_ArgWord(str,DisP);
}

void          ix86_ArgRegRMDigit(char *str,ix86Param *DisP)
{
   char * a1,*a2;
   char nw = DisP->RealCmd[0] & 0x02;
   DisP->codelen++;
   a1 = ix86_getTileWD(DisP);
   strcat(str,a1);
   a2 = ix86_GetDigitTile(!nw,&DisP->RealCmd[DisP->codelen],DisP);
   ix86_CStile(str,a2);
}

void          ix86_ArgRmDigit(char *str,ix86Param *DisP,char w)
{
   char mod = (DisP->RealCmd[1] >> 6) & 0x03;
   char rm = (DisP->RealCmd[1] & 0x07);
   DisP->codelen++;
   strcat(str,ix86_getModRM(w == 2 ? 1 : w,mod,rm,DisP));
   strcat(str,",");
   strcat(str,ix86_GetDigitTile(w == 2 ? 0 : w,&DisP->RealCmd[DisP->codelen],DisP));
}

void          ix86_ArgOp1(char *str,ix86Param *DisP)
{
   unsigned char w;
   unsigned char idx;
   w = DisP->RealCmd[0] & 0x01;
   idx = (DisP->RealCmd[1] & 0x38) >> 3;
   strcpy(str,ix86_Op1Names[idx]);
   TabSpace(str,TAB_POS);
   ix86_ArgRmDigit(str,DisP,w);
}

void          ix86_ArgOp2(char *str,ix86Param *DisP)
{
   unsigned char w;
   unsigned char idx;
   w = DisP->RealCmd[0] & 0x01;
   idx = (DisP->RealCmd[1] & 0x38) >> 3;
   strcpy(str,ix86_Op1Names[idx]);
   TabSpace(str,TAB_POS);
   if(w) ix86_setModifier(str,Use32Data ? ix86_sizes[DWORD_PTR] : ix86_sizes[WORD_PTR]);
   ix86_ArgRmDigit(str,DisP,w ? 2 : w);
}

void          ix86_ShOp2(char *str,ix86Param *DisP)
{
   char *a1,*a2;
   unsigned char code2 = ( DisP->RealCmd[1] & 0x38 ) >> 3;
   unsigned char mod = (DisP->RealCmd[1] >> 6) & 0x03;
   unsigned char rm = (DisP->RealCmd[1] & 0x07);
   unsigned char w = DisP->RealCmd[0] & 0x01;
   DisP->codelen++;
   strcpy(str,ix86_ShNames[code2]);
   TabSpace(str,TAB_POS);
   a1 = ix86_getModRM(w,mod,rm,DisP);
   strcat(str,a1);
   a2 = ix86_GetDigitTile(0,&DisP->RealCmd[DisP->codelen],DisP);
   ix86_setModifier(str,w ? Use32Data ? ix86_sizes[DWORD_PTR] : ix86_sizes[WORD_PTR] : ix86_sizes[BYTE_PTR]);
   ix86_CStile(str,a2);
}

void          ix86_ShOp1(char *str,ix86Param *DisP)
{
   char *a1;
   unsigned char code2 = ( DisP->RealCmd[1] & 0x38 ) >> 3;
   unsigned char mod = (DisP->RealCmd[1] >> 6) & 0x03;
   unsigned char rm = (DisP->RealCmd[1] & 0x07);
   unsigned char w = DisP->RealCmd[0] & 0x01;
   DisP->codelen++;
   a1 = ix86_getModRM(w,mod,rm,DisP);
   strcpy(str,ix86_ShNames[code2]);
   TabSpace(str,TAB_POS);
   strcat(str,a1);
   ix86_setModifier(str,w ? Use32Data ? ix86_sizes[DWORD_PTR] : ix86_sizes[WORD_PTR] : ix86_sizes[BYTE_PTR]);
   ix86_CStile(str,"1");
}

void          ix86_ShOpCL(char *str,ix86Param *DisP)
{
   char *a1;
   unsigned char code2 = ( DisP->RealCmd[1] & 0x38 ) >> 3;
   unsigned char mod = (DisP->RealCmd[1] >> 6) & 0x03;
   unsigned char rm = (DisP->RealCmd[1] & 0x07);
   unsigned char w = DisP->RealCmd[0] & 0x01;
   DisP->codelen++;
   a1 = ix86_getModRM(w,mod,rm,DisP);
   strcpy(str,ix86_ShNames[code2]);
   TabSpace(str,TAB_POS);
   strcat(str,a1);
   ix86_setModifier(str,w ? Use32Data ? ix86_sizes[DWORD_PTR] : ix86_sizes[WORD_PTR] : ix86_sizes[BYTE_PTR]);
   ix86_CStile(str,"cl");
}

void          ix86_ArgGrp1(char *str,ix86Param *DisP)
{
   unsigned char code2 = ( DisP->RealCmd[1] & 0x38 ) >> 3;
   unsigned char mod = (DisP->RealCmd[1] >> 6) & 0x03;
   unsigned char rm = (DisP->RealCmd[1] & 0x07);
   unsigned char w = DisP->RealCmd[0] & 0x01;
   strcpy(str,ix86_Gr1Names[code2]);
   TabSpace(str,TAB_POS);
   DisP->codelen++;
   ix86_setModifier(str,w ? Use32Data ? ix86_sizes[DWORD_PTR] : ix86_sizes[WORD_PTR] : ix86_sizes[BYTE_PTR]);
   if(!code2) // test only
   {
     strcat(str,ix86_getModRM(w,mod,rm,DisP));
     ix86_CStile(str,ix86_GetDigitTile(w == 2 ? 0 : w,&DisP->RealCmd[DisP->codelen],DisP));
   }
   else
   strcat(str,ix86_getModRM(w,mod,rm,DisP));
}

void          ix86_ArgGrp2(char *str,ix86Param *DisP)
{
   unsigned char code2 = ( DisP->RealCmd[1] & 0x38 ) >> 3;
   unsigned char mod = (DisP->RealCmd[1] >> 6) & 0x03;
   unsigned char rm = (DisP->RealCmd[1] & 0x07);
   unsigned char w = DisP->RealCmd[0] & 0x01;
   unsigned char wrd = code2 & 0x01;
   unsigned char sizptr;
   strcpy(str,ix86_Gr1Names[code2]);
   TabSpace(str,TAB_POS);
   DisP->codelen++;
   ix86_setModifier(str,w ? Use32Data ? ix86_sizes[DWORD_PTR] : ix86_sizes[WORD_PTR] : ix86_sizes[BYTE_PTR]);
   strcpy(str,ix86_Gr2Names[code2]);
   TabSpace(str,TAB_POS);
   if(code2 == 0 || code2 == 1) { strcat(str,ix86_getModRM(w,mod,rm,DisP)); } // inc dec
   else
   {
     wrd = code2 == 0x06 || code2 == 0x07 ? 0 : wrd; //push / pop;
     if(!wrd) sizptr = Use32Data ? DWORD_PTR : WORD_PTR;
     else     sizptr = Use32Data ? PWORD_PTR : DWORD_PTR;
     ix86_setModifier(str,ix86_sizes[sizptr]);
     strcat(str,ix86_getModRM(1,mod,rm,DisP));
   }
}

void          ix86_ArgAXMem(char *str,ix86Param *DisP)
{
  unsigned char d = DisP->RealCmd[0] & 0x02;
  const char *mem,*reg;
  mem = Use32Addr ? Get8SquareDig((unsigned char *)&DisP->RealCmd[1],DisP,0) : Get4SquareDig((unsigned char *)&DisP->RealCmd[1],DisP,0);
  strcpy(ix86_appstr,ix86_segpref);
  ix86_segpref[0] = 0;
  strcat(ix86_appstr,"[");
  strcat(ix86_appstr,mem);
  strcat(ix86_appstr,"]");
  reg = ix86_getREG(0,DisP->RealCmd[0] & 0x01);
  DisP->codelen = Use32Addr ? 5 : 3;
  strcat(str,d ? ix86_appstr : reg);
  ix86_CStile(str,d ? reg : ix86_appstr);
}

void          ix86_InOut(char * str,ix86Param *DisP)
{
  const char * reg1,*reg2,*regptr,*dig;
  char i;
  regptr = ix86_getREG(0,DisP->RealCmd[0] & 0x01);
  dig = Get2Digit(DisP->RealCmd[1]);
  i = DisP->RealCmd[0] & 0x08;
  if(!i) DisP->codelen++;
  if(DisP->RealCmd[0] & 0x02)
  {
      reg1  = i ? "dx" : dig;
      reg2  = regptr;
  }
  else
  {
      reg1  = regptr;
      reg2  = i ? "dx" : dig;
  }
  strcat(str,reg1);
  ix86_CStile(str,reg2);
}

  