/****************************************************************************\
*   NULL_DA - this source file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/disasm/null_da.c                           *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of Null disassembler.                    *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "pmalloc.h"
#include "reg_form.h"
#include "disasm.h"
#include "bconsole.h"
#include "biewhelp.h"
#include "biewutil.h"
#include "file_ini.h"
#include "reg_form.h"

static int nulWidth = 1;

static char *width_names[] =
{
   "~Byte",
   "~Word",
   "~Double word"
};

int nulSelect_width( void )
{
  unsigned nModes,i;
  nModes = sizeof(width_names)/sizeof(char *);
  i = SelBoxA(width_names,nModes," Select bitness mode: ",nulWidth);
  if(i != -1)
  {
    nulWidth = i;
    return 1;
  }
  return 0;
}

static char *outstr;

static DisasmRet nulDisassembler(unsigned long ulShift,MBuffer buffer)
{
  DisasmRet ret;
  int type,cl;
  char *preface;
  memset(&ret,0,sizeof(ret));
  ret.str = outstr;
  switch(nulWidth)
  {
    case 0: preface = "db ";
            type = DISARG_BYTE;
            cl = 1;
            break;
    default:
    case 1: preface = "dw ";
            type = DISARG_WORD;
            cl = 2;
            break;
    case 2: preface = "dd ";
            type = DISARG_DWORD;
            cl = 4;
            break;
  }
  ret.codelen = cl;
  strcpy(outstr,preface);
  disAppendDigits(outstr,ulShift,APREF_USE_TYPE,cl,buffer,type);
  return ret;
}

static void  nulHelpAsm( void )
{
  hlpDisplay(20001);
}

static int   nulMaxInsnLen( void ) { return 4; }
static Color nulGetAsmColor( unsigned long clone ) { return LightGray; }
static int   nulGetBitness( void ) { return DAB_USE16; }
static char  nulGetClone( unsigned long clone ) { return ' '; }
static void  nulInit( void )
{
  outstr = PMalloc(1000);
  if(!outstr)
  {
    MemOutBox("Data disassembler initialization");
    exit(EXIT_FAILURE);
  }
}

static void  nulTerm( void )
{
   PFREE(outstr);
}

static void nulReadIni( void )
{
  char tmps[10];
  if(isValidIniArgs())
  {
    biewReadProfileString(ini_name,"Biew","Browser","SubSubMode3","1",tmps,sizeof(tmps));
    nulWidth = (int)strtoul(tmps,NULL,10);
    if(nulWidth > 2) nulWidth = 0;
  }
}

static void nulWriteIni( void )
{
  char tmps[10];
  sprintf(tmps,"%i",nulWidth);
  biewWriteProfileString(ini_name,"Biew","Browser","SubSubMode3",tmps);
}

REGISTRY_DISASM Null_Disasm =
{
  "~Data",
  { NULL, "Width ", NULL, NULL, NULL },
  { NULL, nulSelect_width, NULL, NULL, NULL },
  nulDisassembler,
  NULL,
  nulHelpAsm,
  nulMaxInsnLen,
  nulGetAsmColor,
  nulGetBitness,
  nulGetClone,
  nulInit,
  nulTerm,
  nulReadIni,
  nulWriteIni
};
  