/****************************************************************************\
*   DISASM - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/mod/disasm.c                               *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of Intel dis disassembler (main module). *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "pmalloc.h"
#include "kbd_code.h"
#include "bmfile.h"
#include "disasm.h"
#include "biewutil.h"
#include "file_ini.h"
#include "reg_form.h"
#include "bconsole.h"
#include "editor.h"
#include "codeguid.h"
#include "__os_dep.h"

/* references resolution */
#define NEEDREF_ALL 2
#define NEEDREF_JMP 1
#define NEEDREF_NONE 0

extern REGISTRY_DISASM ix86_Disasm;
extern REGISTRY_DISASM Null_Disasm;

static REGISTRY_DISASM *mainDisasmTable[] =
{
  &ix86_Disasm,
  &Null_Disasm
};

static int DefDisasmSel = __DEFAULT_DISASM;
REGISTRY_DISASM *activeDisasm = NULL;

extern int hexAddressResolv;
static unsigned long PrevPageSize,CurrPageSize,PrevStrLen,CurrStrLen;
static long LastCFPos;
int disNeedRef = 0;
int DisasmCurrLine;
int disPanelMode = 0;
static int           HiLight = 1;
static char          *CurrStrLenBuff = NULL;
static unsigned long *PrevStrLenAddr = NULL;
static char          LastPrevLen;
static char          PrevStrCount = 0;
static char *        disCodeBuffer = NULL;
static int           disMaxCodeLen;

char DisasmPrepareMode = 0;

static void __NEAR__ disAcceptActions( void );

int disSelect_Disasm( void )
{
  char *modeName[sizeof(mainDisasmTable)/sizeof(REGISTRY_DISASM *)];
  unsigned nModes;
  int i;

  nModes = sizeof(mainDisasmTable)/sizeof(REGISTRY_DISASM *);
  for(i = 0;i < nModes;i++) modeName[i] = mainDisasmTable[i]->name;
  i = SelBoxA(modeName,nModes," Select disassembler: ",DefDisasmSel);
  if(i != -1)
  {
    if(activeDisasm->term) activeDisasm->term();
    activeDisasm = mainDisasmTable[i];
    DefDisasmSel = i;
    disAcceptActions();
    return 1;
  }
  return 0;
}

static void FillPrevAsmPage(unsigned long bound,unsigned predist)
{
  unsigned long distin,addr;
  unsigned j;
  unsigned totallen;
  char showref,addrdet;
  if(!predist) predist = Height*disMaxCodeLen;
  predist = (predist/16)*16+16;   /* align on 16-byte boundary */
  distin = bound >= predist ? bound - predist : 0;
  memset(PrevStrLenAddr,0,tvioHeight*sizeof(long));
  PrevStrCount = 0;
  totallen = 0;
  showref = disNeedRef;
  addrdet = hexAddressResolv;
  hexAddressResolv = 0;
  disNeedRef = 0;
  DisasmPrepareMode = 0xFF;
  for(j = 0;;j++)
  {
    DisasmRet dret;
       addr = distin + totallen;
       BMReadBufferEx(disCodeBuffer,disMaxCodeLen,addr,BM_SEEK_SET);
       dret = Disassembler(distin,(void *)disCodeBuffer);
       if(addr >= bound) break;
       totallen += dret.codelen;
       if(j < Height) PrevStrLenAddr[j] = addr;
       else
       {
          memmove(PrevStrLenAddr,&PrevStrLenAddr[1],(Height - 1)*sizeof(long));
          PrevStrLenAddr[Height - 1] = addr;
       }
  }
  PrevStrCount = j < Height ? j : Height;
  LastPrevLen = bound - PrevStrLenAddr[PrevStrCount - 1];
  disNeedRef = showref;
  DisasmPrepareMode = 0;
  hexAddressResolv = addrdet;
}

static void PrepareAsmLines(int keycode,unsigned long cfpos)
{
  switch(keycode)
  {
    case KE_DOWNARROW:
                    PrevStrLen = CurrStrLenBuff[0];
                    CurrStrLen = CurrStrLenBuff[1];
                    if(PrevStrCount < Height) PrevStrCount++;
                    else                      memmove(&PrevStrLenAddr[0],&PrevStrLenAddr[1],sizeof(long)*(Height - 1));
                    PrevStrLenAddr[PrevStrCount - 1] = cfpos - CurrStrLenBuff[0];
                    PrevPageSize = PrevStrLenAddr[PrevStrCount - 1] - PrevStrLenAddr[0] + PrevStrLen;
                    break;
    case KE_PGDN:
                    {
                      int i;
                      unsigned size;
                      unsigned long prevpos;
                      size = 0;
                      for(i = 0;i < Height;i++)
                      {
                         size += CurrStrLenBuff[i];
                      }
                      prevpos = cfpos - size;
                      size = 0;
                      for(i = 0;i < Height;i++)
                      {
                          size += CurrStrLenBuff[i];
                          PrevStrLenAddr[i] = prevpos + size;
                      }
                      PrevStrLen = CurrStrLenBuff[Height - 1];
                      PrevPageSize = size;
                      PrevStrCount = Height;
                    }
                    break;
   case KE_UPARROW:
                    FillPrevAsmPage(cfpos,(unsigned)(PrevPageSize + 15));
                    goto Calc;
    default:
                    FillPrevAsmPage(cfpos,0);
    Calc:
                    PrevPageSize = 0;
                    if(PrevStrCount)
                    {
                      PrevStrLen = LastPrevLen;
                      PrevPageSize = PrevStrLenAddr[PrevStrCount - 1] + LastPrevLen - PrevStrLenAddr[0];
                    }
                    else             PrevStrLen = 0;
                    break;
  }
}

unsigned drawAsm( unsigned keycode, unsigned textshift )
{
 int i,j,len,I,Limit,dir;
 unsigned long cfpos,flen;
 static unsigned long amocpos = 0L;
 char outstr[__TVIO_MAXSCREENWIDTH];
 cfpos = BMGetCurrFilePos();
 if(keycode == KE_UPARROW)
 {
   char showref,addrdet;
   DisasmRet dret;
   showref = disNeedRef;
   addrdet = hexAddressResolv;
   disNeedRef = hexAddressResolv = 0;
   BMReadBufferEx(disCodeBuffer,disMaxCodeLen,cfpos,BM_SEEK_SET);
   DisasmPrepareMode = 0xFF;
   dret = Disassembler(cfpos,(void *)disCodeBuffer);
   if(cfpos + dret.codelen != amocpos) keycode = KE_SUPERKEY;
   DisasmPrepareMode = 0;
   disNeedRef = showref;
   hexAddressResolv = addrdet;
 }
 PrepareAsmLines(keycode,cfpos);
 if(keycode == KE_UPARROW) { I = Height - 1; dir = -1; Limit = -1; }
 else                   { I = 0; dir = 1; Limit = Height; }
 if(amocpos != cfpos || keycode == KE_SUPERKEY)
 {
   GidResetGoAddress(keycode);
   amocpos = cfpos;
   flen = BMGetFLength();
   twUseWin(MainWnd);
   for(i = I;i != Limit;i+=1*dir)
   {
     DisasmRet dret;
     memset(outstr,' ',__TVIO_MAXSCREENWIDTH);
     DisasmCurrLine = i;
     if(cfpos < BMGetFLength())
     {
       if(keycode == KE_DOWNARROW && i < Height - 1)
       {
         char tbuff[__TVIO_MAXSCREENWIDTH*2];
         CurrStrLenBuff[i] = CurrStrLenBuff[i + 1];
         twReadBuffer(MainWnd,1,i + 2,tbuff,Width);
         twWriteBuffer(MainWnd,1,i + 1,tbuff,Width);
         cfpos += CurrStrLenBuff[i];
       }
       else
       {
         if(!(keycode == KE_UPARROW && i > 0))
         {
           len = cfpos + disMaxCodeLen < flen ? disMaxCodeLen : (int)(flen - cfpos);
           memset(disCodeBuffer,0,disMaxCodeLen);
           BMReadBufferEx((void *)disCodeBuffer,len,cfpos,BM_SEEK_SET);
           dret = Disassembler(cfpos,(void *)disCodeBuffer);
           if(i == 0) CurrStrLen = dret.codelen;
           CurrStrLenBuff[i] = dret.codelen;
           twTextColor(LightCyan);
           memcpy(outstr,GidEncodeAddress(cfpos,hexAddressResolv),10);
           len = 0;
           if(disPanelMode < PANMOD_FULL)
           {
             static char _clone;
             twDirectWrite(1,i + 1,outstr,10);
             len = 10;
             twTextColor(LightBlue);
             _clone = activeDisasm->CloneShortName(dret.pro_clone);
             twDirectWrite(10,i + 1,&_clone,1);
             twTextColor(LightCyan);
           }
           if(disPanelMode < PANMOD_MEDIUM)
           {
             int full_off,med_off;
             full_off = disMaxCodeLen*2+11;
             med_off = disMaxCodeLen*2+1;
             for(j = 0;j < dret.codelen;j++) { memcpy(&outstr[len],Get2Digit(disCodeBuffer[j]),2); len += 2; }
             if(len < (disPanelMode < PANMOD_FULL) ? full_off : med_off) len = disPanelMode < 2 ? full_off : med_off;
             twDirectWrite(disPanelMode < PANMOD_FULL ? 11 : 1,
                           i + 1,
                           &outstr[10],
                           disPanelMode < PANMOD_FULL ? len - 11 : len - 1);
             if(isHOnLine(cfpos,dret.codelen))  HiLightSearch(MainWnd,cfpos,10,dret.codelen,i,&outstr[10],1,0);
           }
           twDirectWrite(len,i + 1," ",1);
           len++;
           twTextColor(HiLight ? activeDisasm->GetInsnColor(dret.pro_clone) : LightCyan);
           j = strlen(dret.str);
           twDirectWrite(len,i + 1,dret.str,j); len += j;
           if(len < Width)
           {
             twGotoXY(len,i + 1);
             twClrEOL();
           }
           cfpos += dret.codelen;
           BMSeek(cfpos,BM_SEEK_SET);
           if(keycode == KE_UPARROW) LastCFPos += dret.codelen;
           else
             if(i == Height - 1)  LastCFPos = cfpos;
         }
         else
         {
           char tbuff[__TVIO_MAXSCREENWIDTH];
           if(i == Height - 1) LastCFPos = cfpos;
           CurrStrLenBuff[i] = CurrStrLenBuff[i - 1];
           LastCFPos += CurrStrLenBuff[i];
           twReadBuffer(MainWnd,1,i,tbuff,Width);
           twWriteBuffer(MainWnd,1,i + 1,tbuff,Width);
         }
       }
     }
     else twDirectWrite(1,i + 1,outstr,Width);
   }
   twSetColor(LightCyan,Blue);
   lastbyte = cfpos;
   CurrPageSize = LastCFPos - amocpos;
 }
 return textshift;
}

unsigned long disPrevPageSize( void ) { return PrevPageSize; }
unsigned long disCurrPageSize( void ) { return CurrPageSize; }
unsigned long disPrevLineWidth( void ) { return PrevStrLen; }
unsigned long disCurrLineWidth( void ) { return CurrStrLen; }
char *        disMiscKeyName( void ) { return "Modify"; }

static char *refsdepth_names[] =
{
   "~None",
   "~Calls/jmps only (navigation)",
   "~All"
};

int disReferenceResolving( void )
{
  unsigned nModes,i;
  int ret;
  nModes = sizeof(refsdepth_names)/sizeof(char *);
  i = SelBoxA(refsdepth_names,nModes," References resolving depth: ",disNeedRef);
  if(i != -1)
  {
    disNeedRef = i;
    ret = 1;
  }
  else ret = 0;
  if(detectedFormat->set_state)
    detectedFormat->set_state(disNeedRef ? PS_ACTIVE : PS_INACTIVE);
  return ret;
}

static char *panmod_names[] =
{
   "~Full",
   "~Medium",
   "~Wide"
};

int disSelectPanelMode( void )
{
  unsigned nModes,i;
  nModes = sizeof(panmod_names)/sizeof(char *);
  i = SelBoxA(panmod_names,nModes," Select panel mode: ",disPanelMode);
  if(i != -1)
  {
    disPanelMode = i;
    return 1;
  }
  else return 0;
}

static char *hilight_names[] =
{
   "~Mono",
   "~Highlight"
};

int disSelectHiLight( void )
{
  unsigned nModes,i;
  nModes = sizeof(hilight_names)/sizeof(char *);
  i = SelBoxA(hilight_names,nModes," Select highlight mode: ",HiLight);
  if(i != -1)
  {
    HiLight = i;
    return 1;
  }
  else return 0;
}


int disDetect( void ) { return 1; }

static int DefAsmAction(int lastbyte,int start)
{
 int redraw,index;
 char xx,dox;
  redraw = 1;
  xx = edit_x / 2;
  dox = 1;
  index = start + xx;
   switch(lastbyte)
   {
     case KE_F(4)   : EditorMem.arr[index] = ~EditorMem.arr[index]; break;
     case KE_F(5)   : EditorMem.arr[index] |= edit_XX; break;
     case KE_F(6)   : EditorMem.arr[index] &= edit_XX; break;
     case KE_F(7)   : EditorMem.arr[index] ^= edit_XX; break;
     case KE_F(8)   : EditorMem.arr[index]  = edit_XX; break;
     case KE_F(9)   : EditorMem.arr[index] = EditorSaveMem.arr[index]; break;
     default        : redraw = edit_defaction(lastbyte); dox = 0; break;
   }
  if(dox) { xx++; edit_x = xx*2; }
  return redraw;
}

static void DisasmScreen(TWindow* ewnd,unsigned long cp,unsigned long flen,int st,int stop,int start,char Fore,char UndoFore)
{
 int i,j,len,lim;
 char outstr[81],outstr1[81];
 DisasmRet dret;
 twFreezeWin(ewnd);
 for(i = st;i < stop;i++)
 {
  if(start + cp < flen)
  {
   memcpy(outstr,GidEncodeAddress(cp + start,hexAddressResolv),10);
   twUseWin(MainWnd);
   twTextColor(LightCyan);
   twDirectWrite(1,i + 1,outstr,9);
   dret = Disassembler(cp + start,&EditorMem.arr[start]);
   alen[i] = dret.codelen;
   aolen[i] = alen[i];
   memset(outstr,0,Width);
   memset(outstr1,0,Width);
   len = 0; for(j = 0;j < alen[i];j++) { memcpy(&outstr1[len],Get2Digit(EditorSaveMem.arr[start + j]),2); len += 2; }
   len = 0; for(j = 0;j < alen[i];j++) { memcpy(&outstr[len],Get2Digit(EditorMem.arr[start + j]),2); len += 2; }
   twUseWin(ewnd);
   len = disMaxCodeLen*2;
   for(j = 0;j < len;j++)
   {
     twTextColor(outstr[j] == outstr1[j] ? Fore : UndoFore);
     twDirectWrite(j + 1,i + 1,&outstr[j],1);
   }
   len = strlen(dret.str);
   memset(outstr,0,Width);
   memcpy(outstr,dret.str,len);
   twUseWin(MainWnd);
   twTextColor(LightCyan);
   lim = disMaxCodeLen*2+11;
   twDirectWrite(lim+1,i + 1,outstr,Width-lim);
   start += alen[i];
  }
  else
  {
    twUseWin(MainWnd);
    twGotoXY(1,i + 1);
    twClrEOL();
    twUseWin(ewnd);
    twGotoXY(1,i + 1);
    twClrEOL();
    alen[i] = 0;
  }
 }
 twRefreshWin(ewnd);
}

static int FullAsmEdit(TWindow * ewnd,char Fore,char UndoFore)
{
 int j,len,rlen,lastbyte,start;
 unsigned long flen;
 unsigned max_buff_size = disMaxCodeLen*tvioHeight;
 char redraw = 0;
 char outstr[__TVIO_MAXSCREENWIDTH],owork[__TVIO_MAXSCREENWIDTH];

 flen = BMGetFLength();
 edit_cp = BMGetCurrFilePos();
 start = 0;

 rlen = edit_cp + max_buff_size < flen ? max_buff_size : (int)(flen - edit_cp);
 BMReadBufferEx((void *)EditorMem.arr,rlen,edit_cp,BM_SEEK_SET);
 memcpy(EditorSaveMem.arr,EditorMem.arr,max_buff_size);
 memset(alen,0,tvioHeight);

 DisasmScreen(ewnd,edit_cp,flen,0,Height,start,Fore,UndoFore);
 PaintETitle(0,1);
 start = 0;
 while(1)
 {
   twUseWin(ewnd);

   len = 0; for(j = 0;j < alen[edit_y];j++) { memcpy(&owork[len],Get2Digit(EditorSaveMem.arr[start + j]),2); len += 2; }
   len = 0; for(j = 0;j < alen[edit_y];j++) { memcpy(&outstr[len],Get2Digit(EditorMem.arr[start + j]),2); len += 2; }

   lastbyte = eeditstring(outstr,&legalchars[2],&len,edit_y + 1,&edit_x,
                          __ESS_FILLER_7BIT | __ESS_WANTRETURN | __ESS_HARDEDIT,
                          owork,Fore,UndoFore);
   CompressHex(&EditorMem.arr[start],outstr,alen[edit_y],0);
   switch(lastbyte)
   {
     case KE_TAB     :
                      {
                       AsmRet aret;
                       if(activeDisasm->asm_f)
                       {
                         char code[81];
                         if(GetStringDlg(code,activeDisasm->name,
                                         NULL,"Input assembelr instruction:"))
                         {
                           aret = (*activeDisasm->asm_f)(code);
                           if(aret.err_code)
                           {
                              ErrMessageBox("Syntax error",NULL);
                              continue;
                           }
                           else
                           {
                              memcpy(outstr,aret.insn,aret.insn_len);
                           }
                         }
                         break;
                       }
                       else continue;
                      }
     case KE_F(1)    : ExtHelp(); continue;
     case KE_CTL_F(1): activeDisasm->action[0](); continue;
     case KE_CTL_F(2): SelectSysInfo(); continue;
     case KE_CTL_F(3): SelectTool(); continue;
     case KE_F(2)    : BMWriteBuffEx(edit_cp,BM_SEEK_SET,(void *)EditorMem.arr,rlen);
     case KE_F(10)   :
     case KE_ESCAPE: return lastbyte;
     default: redraw = DefAsmAction(lastbyte,start); break;
   }
   CheckYBounds();
   start = edit_y ? Summ(alen,edit_y) : 0;
   if(redraw)
   {
     DisasmRet dret;
     dret = Disassembler(edit_cp + start,&EditorMem.arr[start]);
     alen[edit_y] = dret.codelen;
     DisasmScreen(ewnd,edit_cp,flen,0,Height,0,Fore,UndoFore);
     redraw = 0;
   }
   PaintETitle(start + edit_x/2,1);
   CheckXYBounds();
   start = edit_y ? Summ(alen,edit_y) : 0;
 }
}

void disEdit( void )
{
 TWindow * ewnd;
 int err_state;
 ewnd = WindowOpen(11,2,disMaxCodeLen*2+11,tvioHeight-1,TWS_CURSORABLE);
 twSetColor(White,Blue); twClearWin();
 twShowWin(ewnd);
 edit_x = edit_y = 0;
 EditMode = ~EditMode;
 editInitBuffs(disMaxCodeLen,&err_state);
 if(!err_state)
 {
   FullAsmEdit(ewnd,White,LightRed);
   editDestroyBuffs();
 }
 EditMode = ~EditMode;
 CloseWnd(ewnd);
 PaintTitle();
}

extern int hexAddressResolution( void );

static void HelpAsm( void )
{
   if( activeDisasm->ShowShortHelp ) activeDisasm->ShowShortHelp();
}

extern void ReadIniAResolv( void );
extern void WriteIniAResolv( void );

static void disReadIni( void )
{
  char tmps[10];
  if(isValidIniArgs())
  {
    biewReadProfileString(ini_name,"Biew","Browser","LastSubMode","0",tmps,sizeof(tmps));
    DefDisasmSel = (int)strtoul(tmps,NULL,10);
    if(DefDisasmSel >= sizeof(mainDisasmTable)/sizeof(REGISTRY_DISASM *)) DefDisasmSel = 0;
    ReadIniAResolv();
    biewReadProfileString(ini_name,"Biew","Browser","SubSubMode7","0",tmps,sizeof(tmps));
    disPanelMode = (int)strtoul(tmps,NULL,10);
    if(disPanelMode > PANMOD_FULL) disPanelMode = 0;
    biewReadProfileString(ini_name,"Biew","Browser","SubSubMode8","0",tmps,sizeof(tmps));
    disNeedRef = (int)strtoul(tmps,NULL,10);
    if(disNeedRef > NEEDREF_ALL) disNeedRef = 0;
    biewReadProfileString(ini_name,"Biew","Browser","SubSubMode9","0",tmps,sizeof(tmps));
    HiLight = (int)strtoul(tmps,NULL,10);
    if(HiLight > 1) HiLight = 1;
    activeDisasm = mainDisasmTable[DefDisasmSel];
    disAcceptActions();
    if(activeDisasm->read_ini) activeDisasm->read_ini();
  }
}

static void disInit( void )
{
  CurrStrLenBuff = PMalloc(tvioHeight);
  PrevStrLenAddr = PMalloc(tvioHeight*sizeof(long));
  if((!CurrStrLenBuff) || (!PrevStrLenAddr))
  {
    MemOutBox("Disassembler initialization");
    exit(EXIT_FAILURE);
  }
  activeDisasm = mainDisasmTable[DefDisasmSel];
  disAcceptActions();
}

static void disTerm( void )
{
  if(activeDisasm->term) activeDisasm->term();
  PFREE(CurrStrLenBuff);
  PFREE(PrevStrLenAddr);
  PFREE(disCodeBuffer);
}

static void disSaveIni( void )
{
  char tmps[10];
  sprintf(tmps,"%i",DefDisasmSel);
  biewWriteProfileString(ini_name,"Biew","Browser","LastSubMode",tmps);
  WriteIniAResolv();
  sprintf(tmps,"%i",disPanelMode);
  biewWriteProfileString(ini_name,"Biew","Browser","SubSubMode7",tmps);
  sprintf(tmps,"%i",disNeedRef);
  biewWriteProfileString(ini_name,"Biew","Browser","SubSubMode8",tmps);
  sprintf(tmps,"%i",HiLight);
  biewWriteProfileString(ini_name,"Biew","Browser","SubSubMode9",tmps);
  if(activeDisasm->save_ini) activeDisasm->save_ini();
}

DisasmRet Disassembler(unsigned long ulShift,MBuffer buffer)
{
  return activeDisasm->disasm(ulShift,buffer);
}

REGISTRY_MODE disMode =
{
  "~Dissasembler",
  { NULL, "Disasm", NULL, NULL, NULL, "AResol", "PanMod", "ResRef", "HiLght", NULL },
  { NULL, disSelect_Disasm, NULL, NULL, NULL, hexAddressResolution, disSelectPanelMode, disReferenceResolving, disSelectHiLight, NULL },
  disDetect,
  0,
  drawAsm,
  NULL,
  disMiscKeyName,
  disEdit,
  disPrevPageSize,
  disCurrPageSize,
  disPrevLineWidth,
  disCurrLineWidth,
  HelpAsm,
  disReadIni,
  disSaveIni,
  disInit,
  disTerm
};

static void __NEAR__ disAcceptActions( void )
{
  if(activeDisasm->init) activeDisasm->init();
  disMaxCodeLen = activeDisasm->max_insn_len();
  disCodeBuffer = PMalloc(disMaxCodeLen);
  if(!disCodeBuffer)
  {
    MemOutBox("Disassembler initialization");
    exit(EXIT_FAILURE);
  }
  disMode.prompt[0] = activeDisasm->prompt[0];
  disMode.action[0] = activeDisasm->action[0];
  disMode.prompt[2] = activeDisasm->prompt[1];
  disMode.action[2] = activeDisasm->action[1];
  disMode.prompt[3] = activeDisasm->prompt[2];
  disMode.action[3] = activeDisasm->action[2];
  disMode.prompt[4] = activeDisasm->prompt[3];
  disMode.action[4] = activeDisasm->action[3];
  disMode.prompt[9] = activeDisasm->prompt[4];
  disMode.action[9] = activeDisasm->action[4];
}

/* Common disassembler utility */

extern int AppendAsmRef(char *str,unsigned long ulShift,int mode,char codelen,unsigned long r_shift);

char * TabSpace(char * str,unsigned nSpace)
{
  int i,len;
  len = strlen(str);
  for(i = len;i < max(len,nSpace);i++) str[i] = ' ';
  if(len >= nSpace) str[i++] = ' ';
  str[i] = 0;
  return str;
}

void   disSetModifier(char *str,const char *modf,unsigned nSpace)
{
  int len;
  len = strlen(str) - 1;
  while(str[len] == ' ') len--;
  len+=2;
  strcpy(&str[len],modf);
  TabSpace(str,nSpace);
}

void disAppendDigits(char *str,unsigned long ulShift,int flags,
                     char codelen,void *defval,int type)
{
 int app;
 char *appstr;
  if(disNeedRef == NEEDREF_ALL)
    app = AppendAsmRef(str,ulShift,flags,codelen,0L);
  else
    app = 0;
  if(!app)
  {
    switch(type)
    {
      case DISARG_LONG:  appstr = Get8SignDig(*(long *)defval); break;
      case DISARG_SHORT: appstr = Get4SignDig(*(short *)defval); break;
      case DISARG_CHAR:  appstr = Get2SignDig(*(char *)defval); break;
      case DISARG_BYTE:  appstr = Get2Digit(*(unsigned char *)defval); break;
      case DISARG_WORD:  appstr = Get4Digit(*(unsigned short *)defval); break;
      default:
      case DISARG_DWORD: appstr = Get8Digit(*(unsigned long *)defval); break;
    }
    strcat(str,appstr);
  }
}

void disAppendFAddr(char * str,long ulShift,long distin,unsigned long r_sh,char type,unsigned seg,char codelen)
{
 int appended,needref;
 appended = 0;
 if(disNeedRef > NEEDREF_NONE)
 {
   if(AppendAsmRef(str,ulShift,APREF_TRY_LABEL,codelen,r_sh)) appended = 1;
 }
 if(!appended)
 {
   needref = disNeedRef;
   disNeedRef = NEEDREF_NONE;
   if(r_sh <= BMGetFLength())
   {
     char * cptr;
     char lbuf[10];
     if(DumpMode) cptr = "L";
     else
     {
       cptr = "file:";
     }
     strcat(str,cptr);
     sprintf(lbuf,"%08lX",r_sh);
     strcat(str,lbuf);
     if(!DumpMode && !EditMode) GidAddGoAddress(str,r_sh);
   }
   else
   {
     char * pstr = "";
     if(type & 0x04)
     {
       strcat(str,Get4Digit(seg));
       strcat(str,":");
     }
     if(!type) pstr = Get2SignDig((char)distin);
     else
       if(type & 0x01) pstr = type & 0x04 ? Get4Digit((unsigned)distin) : Get4SignDig((unsigned)distin);
       else
        if(type & 0x02)   pstr = Get8SignDig(distin);
     strcat(str,pstr);
   }
   disNeedRef = needref;
 }
}
  