/****************************************************************************\
*   DISASM - this header file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/mod/disasm.h                               *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain function prototypes for dissasembler.                   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#ifndef __DISASM_H
#define __DISASM_H

#ifndef __TWIN_H
#include "twin.h"
#endif

typedef int (*DisasmAction)( void );
typedef unsigned char * MBuffer;

typedef struct tagDisasmRet
{
  char         *str;
  unsigned char codelen;
  unsigned long pro_clone;
}DisasmRet;

#define ASM_NOERR  0
#define ASM_SYNTAX 1

typedef struct tagAsmRet
{
  MBuffer       insn;
  unsigned char insn_len;
  int           err_code;
}AsmRet;

typedef DisasmRet (*DisasmFunc)(unsigned long shift,MBuffer insn_buff);
typedef AsmRet    (*AsmFunc)(const char *str);

typedef struct tag_REGISTRY_DISASM
{
  char *       name;         /* disassembler name */
  char *       prompt[5];    /* prompt on Ctrl-(F1,F3-F5,F10) */
  DisasmAction action[5];    /* actions on Ctrl-(F1,F3-F5,F10) */
  DisasmFunc   disasm;       /* main function of disasm */
  AsmFunc      asm_f;        /* assembler (vice versa of disasm) */
  void         (*ShowShortHelp)(void);
  int          (*max_insn_len)(void); /* Max length of 1 disasm instruction */
  Color        (*GetInsnColor)(unsigned long clone);
  int          (*GetDefBitness)(void);
  char         (*CloneShortName)(unsigned long clone);
  void         (*init)(void);
  void         (*term)(void);
  void         (*read_ini)(void);
  void         (*save_ini)(void);
}REGISTRY_DISASM;

extern REGISTRY_DISASM *activeDisasm;
#define PANMOD_FULL 2
#define PANMOD_MEDIUM 1
#define PANMOD_WIDE 0
extern int disPanelMode;

extern DisasmRet Disassembler(unsigned long ulShift,MBuffer buffer);

/* Common disassembler utility */

extern char * TabSpace(char * str,unsigned nSpace);
extern void   disSetModifier(char *str,const char *modf,unsigned nSpace);

#define DISARG_LONG  -4
#define DISARG_SHORT -2
#define DISARG_CHAR  -1
#define DISARG_BYTE  1
#define DISARG_WORD  2
#define DISARG_DWORD 4
/*
   str     - return string
   flags   - as described in reg_form.h
   ulShift - offset to field for binding
   codelen - length of field for binding
   defval  - default value if not binding
   type    - see above
   Examples:
   *** 1 ***
   00005678: B8 34 12       mov   ax, 1234
   after references resolving it can will:
                            mov   ax, off16 KERNEL.@4
   i.e. i must call this function as:
   strcpy(outstr,"mov ax,");
   disAppendDigits(outstr, 0x5679, 1, 2, 0x1234, DISARG_WORD);

   *** 2 ***
   00005678: 8B 80 34 12       mov   ax, [bx+si+1234]
   after references resolving it can will:
                               mov   ax, [bx+si+off16 KERNEL.@4]
   i.e. i must call this function as:
   strcpy(outstr,"mov ax,[bx+si+");
   disAppendDigits(outstr, 0x5680, 1, 2, 0x1234, DISARG_WORD);
   strcat(outstr,"]");
*/
extern void   disAppendDigits(char *str,unsigned long ulShift,int flags,
                              char codelen,void *defval,int type);

#define DISADR_SHORT   0x00
#define DISADR_NEAR16  0x01
#define DISADR_NEAR32  0x02
#define DISADR_USESEG  0x04
#define DISADR_FAR16   (DISADR_USESEG | DISADR_NEAR16)
#define DISADR_FAR32   (DISADR_USESEG | DISADR_NEAR32)
/*
   str     - return string
   ulShift - offset to field for binding
   codelen - length of field for binding
   distin  - original value of field (like defval)
   r_sh    - real shift to begin of insn (for pass to CodeGuider as return addr)
   type    - see above
   seg     - segment value (optional)
   Examples:
   *** 1 ***
   00001002: 9A 78 56 34 12       callf 1234:5678
   after references resolving it may:
                                  callf KERNEL.@90
   i.e. i must call this function as:
   strcpy(outstr,"callf ");
   disAppendFAddr(outstr, 0x1003, 0x5678, 0x1002, DISADR_FAR16, 0x1234, 4);
   *** 2 ***
   00001002: 66 E9 78 56 34 12       calln32 12345678
   after references resolving it may:
                                     calln32 KERNEL32.TerminateApp
   i.e. i must call this function as:
   strcpy(outstr,"calln32 ");
   disAppendFAddr(outstr, 0x1004, 0x12345678, 0x1002, DISADR_NEAR32, 0, 4);
*/
extern void   disAppendFAddr(char * str,long ulShift,long distin,
                             unsigned long r_sh,char type,
                             unsigned seg,char codelen);

#endif
 