/****************************************************************************\
*   HEXMODE - this source file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/mod/hexmode.c                              *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of bite-hexadecimal mode viewers.        *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "bconsole.h"
#include "biewutil.h"
#include "biewhelp.h"
#include "bmfile.h"
#include "file_ini.h"
#include "reg_form.h"
#include "kbd_code.h"
#include "codeguid.h"
#include "editor.h"
#include "__os_dep.h"

typedef char *(__NEAR__ *hexFunc)(unsigned long);

static char * __NEAR__ GetB(unsigned long val) { return GetBinary(BMReadByteEx(val,BM_SEEK_SET)); }
static char * __NEAR__ Get2D(unsigned long val) { return Get2Digit(BMReadByteEx(val,BM_SEEK_SET)); }
static char * __NEAR__ Get4D(unsigned long val) { return Get4Digit(BMReadWordEx(val,BM_SEEK_SET)); }
static char * __NEAR__ Get8D(unsigned long val) { return Get8Digit(BMReadDWordEx(val,BM_SEEK_SET)); }

static unsigned hmode = 1;

int hexAddressResolv = 0;

static char * hex_names[] =
{
  "B~it",
  "~Byte",
  "~Word",
  "~Double Word"
};

const hexFunc hf[4] = { GetB, Get2D, Get4D, Get8D };
const char hHWidths[4] = { 7, 16, 20, 20 };
const char h__incs[4] = { 1, 1, 2, 4 };
const char hdlens[4] = { 8, 2, 4, 8 };


unsigned drawHex( unsigned keycode,unsigned textshift )
{
 short i,j,I,Limit,dir;
 char outstr[__TVIO_MAXSCREENWIDTH+1];
 char tbuff[__TVIO_MAXSCREENWIDTH*2];
 char HWidth;
 unsigned long sindex,cpos,flen,lindex,SIndex;
 static unsigned long hmocpos = 0L;
 hexFunc hmethod;
 short __inc,dlen,rwidth,lw,len,xmin;
 cpos = BMGetCurrFilePos();
 if(hmocpos != cpos || keycode == KE_SUPERKEY)
 {
   HWidth = lw = hHWidths[hmode];
   if(hmocpos != cpos + HWidth || hmocpos != cpos - HWidth) keycode = KE_SUPERKEY;
   hmocpos = cpos;
   __inc = h__incs[hmode];
   dlen = hdlens[hmode];
   hmethod = hf[hmode];
   flen = BMGetFLength();
   if(flen < HWidth) HWidth = flen;
   if(keycode == KE_UPARROW) { I = Height - 1; dir = -1; SIndex = cpos + HWidth*(Height - 1); Limit = -1; }
   else { I = 0; dir = 1; SIndex = cpos; Limit = Height; }
   lindex = flen - SIndex;
   for(i = I,sindex = SIndex;i != Limit;i += 1*dir,sindex += HWidth*dir)
   {
     memset(outstr,' ',Width);
     if(sindex < flen)
     {
       int freq;
       if(keycode == KE_DOWNARROW && i < Height - 1)
       {
         twReadBuffer(MainWnd,1,i + 2,tbuff,Width);
         twWriteBuffer(MainWnd,1,i + 1,tbuff,Width);
       }
       else
       {
         if(!(keycode == KE_UPARROW && i > 0))
         {
            lindex = flen - sindex;
            rwidth = lindex > HWidth ? HWidth : lindex;
            memcpy(outstr,GidEncodeAddress(sindex,hexAddressResolv),10);
            len = 10;
            for(j = 0,freq = 0,lindex = sindex;j < rwidth;j += __inc,lindex += __inc,freq++)
            {
               memcpy(&outstr[len],(*hmethod)(lindex),dlen);
               len += dlen + 1;
               if(hmode == 1) if(freq == 3) { freq = -1; len++; }
            }
            BMReadBufferEx((void *)&outstr[Width - lw],rwidth,sindex,BM_SEEK_SET);
            xmin = hmode ? (hmode == 1 ? 64 : 60) : 73;
            twDirectWrite(1,i + 1,outstr,xmin);
            if(isHOnLine(sindex,HWidth))  HiLightSearch(MainWnd,sindex,xmin,Width,i,&outstr[xmin],0,0);
            else                          twDirectWrite(xmin + 1,i + 1,&outstr[xmin],Width - xmin);
         }
         else
         {
            twReadBuffer(MainWnd,1,i,tbuff,Width);
            twWriteBuffer(MainWnd,1,i + 1,tbuff,Width);
         }
       }
     }
     else twDirectWrite(1,i + 1,outstr,Width);
   }
   lastbyte = lindex + __inc;
 }
 return textshift;
}

void HelpHex( void )
{
   hlpDisplay(1002);
}

unsigned long hexPrevPageSize( void ) { return hHWidths[hmode]*Height; }
unsigned long hexCurrPageSize( void ) { return hHWidths[hmode]*Height; }
unsigned long hexPrevLineWidth( void ) { return hHWidths[hmode]; }
unsigned long hexCurrLineWidth( void ) { return hHWidths[hmode]; }
char *        hexMiscKeyName( void ) { return "Modify"; }

int hexSelectMode( void )
{
  unsigned nModes,i;
  nModes = sizeof(hex_names)/sizeof(char *);
  i = SelBoxA(hex_names,nModes," Select hexadecimal mode: ",hmode);
  if(i != -1)
  {
    hmode = i;
    return 1;
  }
  return 0;
}

static char *aresolv[] =
{
  "~Global (global file offset)",
  "~Local (block and local offset)"
};

int hexAddressResolution( void )
{
  unsigned nModes,i;
  nModes = sizeof(aresolv)/sizeof(char *);
  i = SelBoxA(aresolv,nModes," Select address resolving: ",hexAddressResolv);
  if(i != -1)
  {
    hexAddressResolv = i;
    return 1;
  }
  return 0;
}

int hexDetect( void ) { return 1; }

static int FullHexEdit( int width ,TWindow * txtwnd,char ncol,char hcol,char tcol)
{
 int i,j,mlen;
 unsigned int lastbyte;
 char work[__TVIO_MAXSCREENWIDTH],owork[__TVIO_MAXSCREENWIDTH],redraw;
 {
   TWindow * using = twUsedWin();
   twUseWin(txtwnd);
   twTextColor(tcol);
   twFreezeWin(txtwnd);
   for(i = 0;i < Height;i++)
   {
     twDirectWrite(Width - width + 1,i + 1,&EditorMem.arr[i*tvioWidth],alen[i]);
     if(alen[i] + 1 < width)
     {
        twGotoXY(Width - width + alen[i] + 2,i + 1); twClrEOL();
     }
   }
   twRefreshWin(txtwnd);
   twUseWin(using);
 }
 twTextColor(ncol);
 twFreezeWin(twUsedWin());
 for(i = 0;i < Height;i++)
 {
   ExpandHex(work,&EditorMem.arr[i*tvioWidth],alen[i],1);
   mlen = ExpandHex(owork,&EditorSaveMem.arr[i*tvioWidth],alen[i],1);
   for(j = 0;j < mlen;j++)
   {
     twTextColor(work[j] == owork[j] ? ncol : hcol);
     twDirectWrite(j + 1,i + 1,&work[j],1);
   }
   if(mlen + 1 < width)
   {
      twGotoXY(mlen + 1,i + 1); twClrEOL();
   }
 }
 twRefreshWin(twUsedWin());
 redraw = 1;
 PaintETitle(edit_y*width + edit_x,0);
 while(1)
 {
   mlen = alen[edit_y];
   ExpandHex(work,&EditorMem.arr[edit_y*tvioWidth],mlen,1);
   mlen = ExpandHex(owork,&EditorSaveMem.arr[edit_y*tvioWidth],mlen,1);
   edit_x*=3;
   lastbyte = eeditstring(work,&legalchars[2],&mlen,edit_y + 1,&edit_x,
                          __ESS_WANTRETURN | __ESS_HARDEDIT | __ESS_ASHEX,
                          owork,ncol,hcol);
   edit_x/=3;
   CompressHex(&EditorMem.arr[edit_y*tvioWidth],work,mlen/3,1);
   switch(lastbyte)
   {
     case KE_F(1)   : ExtHelp(); continue;
     case KE_F(2)   : editSaveContest();
     case KE_F(10)  :
     case KE_ESCAPE :
     case KE_TAB : return lastbyte;
     default     : redraw = editDefAction(lastbyte); break;
   }
   CheckBounds();
   if(redraw)
   {
      TWindow * using = twUsedWin();
      twUseWin(txtwnd);
      twDirectWrite(Width - width + 1,edit_y + 1,&EditorMem.arr[edit_y*tvioWidth],mlen/3);
      twUseWin(using);
      redraw = 0;
   }
   PaintETitle(edit_y*width + edit_x,0);
 }
 return lastbyte;
}

void EditHex( void )
{
 TWindow * ewnd[2];
 int active = 0,oactive = 1,err_state;
 if(hmode != 1) return;
 ewnd[0] = WindowOpen(11,2,64,tvioHeight-1,TWS_CURSORABLE);
 twSetColor(White,Blue); twClearWin();
 ewnd[1] = WindowOpen(65,2,Width,tvioHeight-1,TWS_CURSORABLE);
 twSetColor(White,Blue); twClearWin();
 drawEditPrompt();
 edit_cp = editInitBuffs(16,&err_state);
 if(!err_state)
 {
   edit_x = edit_y = 0;
   while(1)
   {
     int lastbyte;
     if(active != oactive)
     {
       twHideWin(ewnd[oactive]);
       twShowWin(ewnd[active]);
       oactive = active;
     }
     twUseWin(ewnd[active]);
     if(!active) lastbyte = FullHexEdit(16,MainWnd,White,LightRed,LightCyan);
     else        lastbyte = FullEdit(16,MainWnd,White,LightRed);
     if(lastbyte == KE_TAB) active = active ? 0 : 1;
     else break;
   }
   editDestroyBuffs();
 }
 twSetColor(LightCyan,Blue);
 CloseWnd(ewnd[0]);
 CloseWnd(ewnd[1]);
 PaintTitle();
}

void ReadIniAResolv( void )
{
  char tmps[10];
  biewReadProfileString(ini_name,"Biew","Browser","SubSubMode6","0",tmps,sizeof(tmps));
  hexAddressResolv = (unsigned)strtoul(tmps,NULL,10);
  if(hexAddressResolv > 1) hexAddressResolv = 0;
}

void WriteIniAResolv( void )
{
  char tmps[10];
  sprintf(tmps,"%i",hexAddressResolv);
  biewWriteProfileString(ini_name,"Biew","Browser","SubSubMode6",tmps);
}

void hexReadIni( void )
{
  char tmps[10];
  if(isValidIniArgs())
  {
    biewReadProfileString(ini_name,"Biew","Browser","LastSubMode","2",tmps,sizeof(tmps));
    hmode = (unsigned)strtoul(tmps,NULL,10);
    if(hmode > 3) hmode = 1;
    ReadIniAResolv();
  }
}

void hexSaveIni( void )
{
  char tmps[10];
  sprintf(tmps,"%i",hmode);
  biewWriteProfileString(ini_name,"Biew","Browser","LastSubMode",tmps);
  WriteIniAResolv();
}

REGISTRY_MODE hexMode =
{
  "~Hexadecimal mode",
  { NULL, "HexMod", NULL, NULL, NULL, "AResol", NULL, NULL, NULL, NULL },
  { NULL, hexSelectMode, NULL, NULL, NULL, hexAddressResolution, NULL, NULL, NULL, NULL },
  hexDetect,
  0,
  drawHex,
  NULL,
  hexMiscKeyName,
  EditHex,
  hexPrevPageSize,
  hexCurrPageSize,
  hexPrevLineWidth,
  hexCurrLineWidth,
  HelpHex,
  hexReadIni,
  hexSaveIni,
  NULL,
  NULL
};  