/****************************************************************************\
*   REFS - this source file is part of Binary vIEW project (BIEW)            *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   resf.c                                             *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain basic level routines for resolving refernces.           *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stdio.h>
#include <string.h>

#include "pmalloc.h"
#include "bmfile.h"
#include "reg_form.h"
#include "bconsole.h"
#include "biewutil.h"
#include "codeguid.h"
#include "biewlib.h"

/***************************************************************************/
/************************  FOR MZ  *****************************************/
/***************************************************************************/

long __HUGE__ * CurrMZChain = 0;
static unsigned long CurrMZCount;
static char __codelen;

static int compare_ptr(const void __HUGE__ *e1,const void __HUGE__ *e2)
{
  unsigned long v1,v2;
  v1 = *((unsigned long __HUGE__ *)e1);
  v2 = *((unsigned long __HUGE__ *)e2);
  if(v1 < v2) return -1;
  else
    if(v1 > v2) return 1;
    else        return 0;
}

static void __NEAR__ BuildMZChain( void )
{
  unsigned i;
  unsigned long fpos;
  TWindow * w,*usd;
  usd = twUsedWin();
  w = CrtHlpWnd(" System is busy ",49,1);
  twUseWin(w);
  twGotoXY(1,1);
  twPutS(" Building reference chains for MZ part");
  twUseWin(usd);
  CurrMZCount = 0;
  fpos = BMGetCurrFilePos();
  for(i = 0;i < oldh.mz.mzRelocationCount;i++)
  {
    unsigned off,seg,j;
    unsigned long ptr;
    void __HUGE__ * tptr;
    if(!CurrMZChain) tptr = PHMalloc(sizeof(void *));
    else             tptr = PHRealloc(CurrMZChain,(CurrMZCount + 1)*sizeof(void *));
    if(!tptr) break;
    CurrMZChain = tptr;
    j = oldh.mz.mzTableOffset + i*4;
    BMSeek(j,BM_SEEK_SET);
    off = BMReadWord();
    seg = BMReadWord();
    ptr = (((long)seg) << 4) + off + (((long)oldh.mz.mzHeaderSize) << 4);
    CurrMZChain[CurrMZCount++] = ptr;
  }
  HQSort(CurrMZChain,CurrMZCount,sizeof(void *),compare_ptr);
  BMSeek(fpos,BM_SEEK_SET);
  CloseWnd(w);
}

static int compare_mz(const void __HUGE__ *e1,const void __HUGE__ *e2)
{
  long l1,l2;
  int ret;
  l1 = *(long __HUGE__ *)e1;
  l2 = *(long __HUGE__ *)e2;
  if(l1 >= l2 && l1 < l2 + __codelen) return 0;
  else
    if(l1 < l2) ret = -1;
    else        ret = 1;
  return ret;
}

static int __NEAR__ isMZReferenced(unsigned long shift,char len)
{
  if(FType == OLD_EXE && oldh.mz.mzRelocationCount)
  {
     unsigned long mz_size;
     mz_size = (long)(oldh.mz.mzPageCount)*512 + oldh.mz.mzPartLastPage;
     if(shift <= mz_size && shift >= ((long)oldh.mz.mzHeaderSize) << 4)
     {
       if(!CurrMZChain) BuildMZChain();
       __codelen = len;
       return HLFind(&shift,CurrMZChain,CurrMZCount,sizeof(long),compare_mz) != 0;
     }
  }
  return 0;
}

static int __NEAR__ AppendMZRef(char *str,unsigned long ulShift,char codelen)
{
  if(isMZReferenced(ulShift,codelen))
  {
     unsigned wrd;
     wrd = BMReadWordEx(ulShift,BM_SEEK_SET);
     strcat(str,Get4Digit(wrd));
     strcat(str,"+PID");
     return 1;
  }
  else return 0;
}


extern REGISTRY_BIN binTable;
int AppendAsmRef(char *str,unsigned long ulShift,int mode,char codelen,unsigned long r_sh)
{
  static int warn_displayed = 0;
  if(detectedFormat->bind) return detectedFormat->bind(str,ulShift,mode,codelen,r_sh);
  else
  {
    if(detectedFormat != &binTable && !warn_displayed)
    {
      ErrMessageBox("Sorry! References resolving for this format still not supported"," Warning! ");
      warn_displayed = 1;
    }
    if(FType == OLD_EXE) return AppendMZRef(str,ulShift,codelen);
    else                     return 0;
  }
}
  