/****************************************************************************\
*   SEARCH - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   search.c                                           *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of file search interface.                *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <limits.h>
#include <stdlib.h>
#define  _CT_FTM
#include <ctype.h>

#include "kbd_code.h"
#include "bmfile.h"
#include "tstrings.h"
#include "bbio.h"
#include "biewutil.h"
#include "bconsole.h"
#include "biewlib.h"
#include "reg_form.h"

extern unsigned long FoundTextSt,FoundTextEnd;
static TWindow *prcntswnd;
extern void ReReadFile( int );
extern REGISTRY_MODE* activeMode;

unsigned char search_buff[21] = "";
unsigned char search_len = 0;
unsigned char search_Case = 0xFF,search_Word = 0,search_Backward = 0;
static char __found;

static unsigned long __NEAR__ __lfind(unsigned long start)
{
  unsigned long flen;
  unsigned long endscan;
  unsigned long tsize,cpos,findptr,retval;
  unsigned proc,pproc;
  int direct,cond;
  char fbuff[21];
  char cbuff[21];
  unsigned ch,ch1;
  unsigned bio_opt;
  int cache[UCHAR_MAX+1];
  /*
   * Cache initialization for adapted Boyer-Moore serach algorithm
  */
  memset(cache,0,sizeof(cache));
  for(ch = 0;ch < search_len;ch++)
  {
    cache[ search_Case ? search_buff[ch] : toupper(search_buff[ch]) ] = ch+1;
  }
  __found = 0;
  flen = BMGetFLength();
  endscan = search_Backward ? 0 : flen;
  direct  = search_Backward ? -1 : 1;
  tsize = labs(endscan-start);
  cond = 0;
  pproc = proc = 0;
  start = search_Backward ? start : start+search_len-1;
  bio_opt = bioGetOptimization(BMbioHandle());
  bioSetOptimization(BMbioHandle(),
                    (bio_opt & (~BIO_OPT_DIRMASK)) |
                    (search_Backward ? BIO_OPT_RBACKSCAN : BIO_OPT_RFORWARD));
  retval = 0;
  for(cpos = start;start != endscan;start+=search_len*direct,cpos=start)
  {
   if(direct == 1 && start > flen - search_len) break;
   if(direct == -1 && start < search_len) break;
   proc = (unsigned)((cpos*100)/tsize);
   if(proc != pproc)
   {
     if(!ShowPercentInWnd(prcntswnd,pproc=proc))  break;
   }
   ch = BMReadByteEx(start,BM_SEEK_SET);
   if(activeMode->is_text) activeMode->convert_cp((char *)&ch,1);
   if(!search_Case) ch = toupper(ch);
   if(cache[ch])
   {
    if(search_len > 1)
    {
      findptr = start-(cache[ch]-1);
      BMReadBufferEx((void *)fbuff,search_len,findptr,BM_SEEK_SET);
      memcpy(cbuff,search_buff,search_len);
      if(activeMode->is_text) activeMode->convert_cp((char *)fbuff,search_len);
      if(!search_Case)
      {
         memupr((void *)fbuff,search_len);
         memupr((void *)cbuff,search_len);
      }
      if(memcmp(fbuff,cbuff,search_len) == 0) cond = 1;
      else
      {
        start -= (search_len-1)*direct;
        continue;
      }
    }
    else { findptr = start; cond = 1; }
    if(search_Word && cond)
    {
       if(start)
       {
         ch = BMReadByteEx(findptr - 1,BM_SEEK_SET);
         if(activeMode->is_text) activeMode->convert_cp((char *)&ch,1);
       }
       else      ch = ' ';
       if(start + search_len < flen)
       {
         ch1 = BMReadByteEx(findptr + search_len,BM_SEEK_SET);
         if(activeMode->is_text) activeMode->convert_cp((char *)&ch,1);
       }
       else      ch1 = ' ';
       if(!(isseparate(ch) && isseparate(ch1))) cond = 0;
    }
   }
   if(cond) { __found = 1; retval = findptr; break; }
  }
  bioSetOptimization(BMbioHandle(),bio_opt);
  return retval;
}

extern char legalchars[];
extern char GetHexAnalog(char byte);
extern int isSpace(int);
extern void CompressHex(unsigned char * dest,const char * src,int size,int usespace);

int ExpandHex(char * dest,const unsigned char * src,int size,char hard)
{
  int i,k;
    dest[0] = '\0';
    k = 0;
    for(i = 0;i < size;i++)
    {
        char * cptr;
        cptr = Get2Digit(src[i]);
        strcat(dest,cptr); k += 2;
        if(hard == 1) { strcat(dest,!((i + 1)%4) ? ( !((i+1)%16) ? " " : "-" ) : " "); k++; }
        else
          if(hard == 2) { strcat(dest,!((i + 1)%4) ? "  " : " ");  k += !((i + 1)%4) ? 2 : 1; }
          else { strcat(dest," "); k++; }
    }
  return k;
}

static void __NEAR__ SearchPaint(TWindow *wdlg,char ** ebuff,unsigned char searchlen,unsigned char search_CaseSens,unsigned char WordOnly,unsigned char search_Backwarding)
{
 TWindow *using = twUsedWin();
 twUseWin(wdlg);
 twDirectWrite(10,3,ebuff[0],searchlen);
 twGotoXY(10+searchlen,3); { int i; for(i = searchlen;i < 20;i++) twPutChar(' '); }
 twDirectWrite(10,5,ebuff[1],searchlen*3);
 twGotoXY(10+searchlen*3,5); { int i; for(i = searchlen*3;i < 60;i++) twPutChar(''); }
 twTextBkGnd(Cyan);
 twGotoXY(40,2); twPutChar(GetBool(search_CaseSens));
 twGotoXY(40,3); twPutChar(GetBool(WordOnly));
 twGotoXY(45,4); twPutS(search_Backwarding ? BACKWARD : FORWARD);
 twTextBkGnd(LightGray);
 twUseWin(using);
}

int SearchDialog( char * searchbuff, unsigned char *searchlen,unsigned char *search_CaseSens,unsigned char *WordOnly,unsigned char *search_Backwarding)
{
  TWindow *hwnd,* ewnd[2];
  int X1,Y1,X2,Y2;
  int x[2] = { 0, 0 };
  char attr[2] = { __ESS_FILLER_7BIT | __ESS_WANTRETURN | __ESS_ENABLEINSERT,
                   __ESS_WANTRETURN | __ESS_ASHEX };
  char ebuff1[21],ebuff2[61];
  char *ebuff[2],*legal[2];
  int mlen[2];
  int ret,ch,i;
  unsigned char active,oactive;
  hwnd = CreateWindowDDnls(FIND_STR,70,5,Black,LightGray,Black,LightGray);
  twSetFooter(hwnd,FIND_PRMT,TMODE_CENTER,Black,LightGray);
  twGetWinPos(hwnd,&X1,&Y1,&X2,&Y2);
  X1 += 10;
  Y1 += 3;
  X2 = X1 + 19;
  Y2 = Y1;
  ewnd[0] = WindowOpen(X1,Y1,X2,Y2,TWS_CURSORABLE);
  twSetColor(White,Cyan); twClearWin();
  Y1 += 2;
  Y2 += 2;
  X2 = X1 + 59;
  ewnd[1] = WindowOpen(X1,Y1,X2,Y2,TWS_CURSORABLE | TWS_NLS866);
  twSetColor(White,Cyan); twClearWin();
  twUseWin(hwnd);
  twGotoXY(2,1);  twPutS(TYPE_STR);
  twGotoXY(2,3);  twPutS("ASCII :");
  twGotoXY(4,5);  twPutS("HEX :");
  twTextBkGnd(Cyan);
  for(i = 0;i < 3;i++) { twGotoXY(38,i + 2); twPutS(msgFindOpt[i]); }
  twTextBkGnd(LightGray);
  legal[0] = NULL;
  legal[1] = &legalchars[2];
  ebuff1[0] = ebuff2[0] = '\0';
  ebuff[0] = ebuff1;
  ebuff[1] = ebuff2;
  if(searchlen)
  {
    memcpy(ebuff[0],searchbuff,*searchlen);
    ExpandHex(ebuff[1],(unsigned char *)searchbuff,*searchlen,0);
  }
  active = 0;
  oactive = active ? 0 : 1;
  ret = 2;
  SearchPaint(hwnd,ebuff,*searchlen,*search_CaseSens,*WordOnly,*search_Backwarding);
  while(1)
  {
    if(oactive != active)
    {
      twHideWin(ewnd[oactive]);
      twShowWin(ewnd[active]);
      oactive = active;
    }
    mlen[0] = 20;
    mlen[1] = 60;
    twUseWin(ewnd[active]);
    ch = eeditstring(ebuff[active],legal[active],&mlen[active],1,
                     &x[active],attr[active],NULL,
                     active ? (*searchlen)*3 : *searchlen,0);
    switch(ch)
    {
       case KE_UPARROW  : active = 0; continue;
       case KE_DOWNARROW: active = 1; continue;
       case KE_TAB      : active = active ? 0 : 1; continue;
       case KE_ENTER    : if(searchlen) ret = 1; else ret = 0; break;
       case KE_ESCAPE   : ret = 0; break;
       case KE_F(2)     : *search_CaseSens = ~(*search_CaseSens); break;
       case KE_F(3)     : (*WordOnly) = ~(*WordOnly); break;
       case KE_F(4)     : (*search_Backwarding) = ~(*search_Backwarding); break;
       default : break;
    }
    if(ret != 2) break;
    twUseWin(hwnd);
    if(!active) { *searchlen = mlen[0]; memcpy(searchbuff,ebuff[0],mlen[0]); }
    else  { *searchlen = mlen[1] / 3; CompressHex((unsigned char *)searchbuff,ebuff[1],*searchlen,1); }
    if(searchlen) memcpy(ebuff[0],searchbuff,*searchlen);
    else     ebuff[0][0] = '\0';
    mlen[0] = *searchlen;
    ExpandHex(ebuff[1],(unsigned char *)searchbuff,*searchlen,0);
    mlen[1] = (*searchlen)*3;
    for(i = 0;i < 2;i++) if(x[i] > mlen[i]) x[i] = mlen[i];
    SearchPaint(hwnd,ebuff,*searchlen,*search_CaseSens,*WordOnly,*search_Backwarding);
  } // while
  CloseWnd(ewnd[0]);
  CloseWnd(ewnd[1]);
  CloseWnd(hwnd);
  return ret;
}

extern unsigned long OldCurrFilePos; // means previouse File position
extern TWindow * ErrorWnd;

unsigned long Search( int mod )
{
  unsigned long found;
  unsigned long fmem,lmem;
  int ret;
  fmem = BMGetCurrFilePos();
  if(!mod) ret = SearchDialog((char *)search_buff,(unsigned char *)&search_len,(unsigned char *)&search_Case,(unsigned char *)&search_Word,(unsigned char *)&search_Backward);
  else     ret = 1;
  if(ret && search_len)
  {
    prcntswnd = PercentWnd(PLEASE_WAIT,SEARCHING);
    lmem = fmem;
    if(FoundTextSt != FoundTextEnd) { if(search_Backward && lmem) lmem--; else if(lmem < BMGetFLength()) lmem++; }
    found = __lfind(lmem);
    CloseWnd(prcntswnd);
    if(__found)  { FoundTextSt = found; FoundTextEnd = found + search_len; return found; }
    else  ErrMessageBox(STR_NOT_FOUND,SEARCH_MSG);
  }
  BMSeek(fmem,BM_SEEK_SET);
  return fmem;
}
  